// convolution using FFT

/*
J.M.P. Franosch (Jan-Moritz.Franosch@physik.tu-muenchen.de)

This file is part of the GNUSSL software package.  This package is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This software is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this distribution; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef _convolution_h_
#define _convolution_h_


#include <fft.h>




template<class Real, class S1, class S2, class S3>
mVector<Real,S3>& 
convolution(const mVector<Real,S1>& function, const mVector<Real,S2>& kernel,
	    mVector<Real,S3>& conv) {
  // Convolves the function with the kernel, i.e.
  // conv(i)=sum( function(j)*kernel(i-j), j=-infinity...infinity )
  // where i=1..conv.size() and 
  // function, kernel=0 by definition outside the appropriate index ranges.
  // Runtime complexity is function.size()*lb(kernel.size()) as FFT is used.
  // Preconditions:
  //   'Real' is a real type.
  // Result:
  //   'conv' contains the convolution and is returned as a reference.
  //   conv.size() is not changed.
  conv=0;

  size_t n=0;
  for (size_t i=0; kernel.size()>(n=(1<<i)); i++) {
  };

  mVector<Real> kernel_part(2*n);
  mVector<Real> function_part(2*n);
  mVector<Real> conv_part(2*n);

  kernel_part=0;
  function_part=0;

  FFT< Real, complex<Real> > fft, fft_kernel;

  for (size_t i=1; i<=kernel.size(); i++) {
    kernel_part(i)=kernel(i);
  };
  fft_kernel.fft(kernel_part);

  for (size_t k=0; k*n<conv.size(); k++) {
    int max_i=( (k+1)*n<=function.size() ? n : function.size()-k*n);
    max_i=max(0,max_i);
    for (int i=1; i<=max_i; i++) {
      function_part(i)=function(k*n+i);
    };
    for (int i=max_i+1; i<=n; i++) {
      function_part(i)=0;
    };

    fft.fft(function_part);

    fft.vc(1)=complex<Real>(fft.vc(1).real()*fft_kernel.vc(1).real(),
			    fft.vc(1).imag()*fft_kernel.vc(1).imag());
    for (size_t i=2; i<=fft.vc.size(); i++) {
      fft.vc(i)*=fft_kernel.vc(i);
    };

    fft.fft(conv_part,-1);

    for (size_t i=1; 
	 i<= ( (k+2)*n<=conv.size() ? 2*n : conv.size()-k*n ); i++) {
      conv(k*n+i)+=sqrt(2*n)*conv_part(i);
    };
  };
  return conv;
};




#endif _convolution_h_
