// computation of eigenvectors


/*
J.M.P. Franosch (Jan-Moritz.Franosch@physik.tu-muenchen.de)

This file is part of the GNUSSL software package.  This package is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This software is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this distribution; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef _eigenvectors_h_
#define _eigenvectors_h_


#include <complex>
#include <limits>
#include <lin_alg.h>




template<class Real, class S1, class S2, class S3>
void 
eigenvectors(const mMatrix<complex<Real>,S1>& M, mMatrix<complex<Real>,S2>& EV,
	     mVector<complex<Real>,S3>& w, 
	     Real tol=0.0) {
// The eigenvalues and eigenvectors of the matrix M are computed.
// Preconditions:
//   'Real' is a real type.
//   M and EV are a n-dimensional square matrices.
//   w is a n-dimensional vector.
// Result:
//   The columns of EV contain the normalized eigenvectors of M.
//   w contains the eigenvalues of M.
  if (tol==0.0) {
    tol=numeric_limits<Real>::epsilon();
  }
  size_t n=M.rsize();
  mMatrix< complex<Real> > C(n,n);
  C=M;
  eigenvalues(C,tol);
  w=C.diagonal();
  mVector<Real> sv(n);
  mMatrix< complex<Real> > u(n,n), v(n,n);
  for (size_t i=1; i<=n; i++) {
    C=M;
    C.diagonal()-=w(i);
    u=0;
    u.diagonal()=complex<Real>(1.0,0);
    v=0;
    v.diagonal()=complex<Real>(1.0,0);
    svd(C,u,sv,v,tol);
    EV.col(i)=v.col(n)/complex<Real>(sqrt(norm(v.col(n))));
#ifdef GNUSSL_CHECK
    mVector< complex<Real> > z(n);
    if ( sup(dot(M,EV.col(i),z)-w(i)*EV.col(i))>1000*numeric_limits<Real>::epsilon() ) {
      gnussl_warning("eigenvectors","eigenvector incorrect");
    }
#endif
  }
}


template<class Real, class S1, class S2, class S3>
void 
eigenvectors_of_real_matrix(const mMatrix<Real,S1>& M, 
			    mMatrix<complex<Real>,S2>& EV,
			    mVector<complex<Real>,S3>& w,
			    Real tol=0.0) {
// The eigenvalues and eigenvectors of the matrix M are computed.
// Preconditions:
//   'Real' is a real type.
//   M and EV are a n-dimensional square matrices.
//   w is a n-dimensional vector.
// Result:
//   The columns of EV contain the normalized eigenvectors of M.
//   w contains the eigenvalues of M.
  int n=M.rsize();
  mMatrix< complex<Real> > C(n,n);
  C=M;
  eigenvectors(C,EV,w,tol);
}




#endif _eigenvectors_h_
