// This file implements numeric_limits
// for Alpha and Intel i486-i686 architectures

/*
J.M.P. Franosch (Jan-Moritz.Franosch@physik.tu-muenchen.de)

This file is part of the GNUSSL software package.  This package is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This software is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this distribution; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#if !defined(__limits_h)
#define __limits_h


#if ( defined(__GNUC__) && __GNUC__>=3 )

#include <limits>

#else



#if !defined(__i386__) && !defined(__alpha__)
#error sorry, numeric_limits has not been implemented for this target yet
#endif





template <class T> class numeric_limits;

enum float_round_style
{
	round_indeterminate		= -1,
	round_toward_zero		= 0,
	round_to_nearest		= 1,
	round_toward_infinity		= 2,
	round_toward_neg_infinity	= 3
};

template <class T> class numeric_limits
{
public:
	static const bool is_specialized;
	static T min();
	static T max();
	static const int digits;
	static const int digits10;
	static const bool is_signed;
	static const bool is_integer;
	static const bool is_exact;
	static const int radix;
	static T epsilon();
	static T round_error();

	static const int min_exponent;
	static const int min_exponent10;
	static const int max_exponent;
	static const int max_exponent10;

	static const bool has_infinity;
	static const bool has_quite_NaN;
	static const bool has_signaling_NaN;
	static const bool has_denorm;
	static T infinity();
	static T quiet_NaN();
	static T signaling_NaN();
	static T denorm_min();

	static const bool is_iec559;
	static const bool is_bounded;
	static const bool is_modulo;

	static const bool traps;
	static const bool tinyness_before;
	static const float_round_style round_style;
};

class numeric_limits<bool>
{
public:
	static const bool is_specialized = true;

	inline static bool min() { return false; }
	inline static bool max() { return true; }

	// these are wrong, but seem irrelevant
	static const int digits		= 1;
	static const int digits10	= 1;

	// this could be wrong, but seems irrelevant
	static const bool is_signed 	= false;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static bool epsilon() { return false; }
	inline static bool round_error() { return false; }

	static const int min_exponent	= false;
	static const int min_exponent10	= false;
	static const int max_exponent	= false;
	static const int max_expontent10 = false;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static bool infinity() { return false; } 
	inline static bool quiet_NaN() { return false; } 
	inline static bool signaling_NaN() { return false; } 
	inline static bool denorm_min() { return false; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= false;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_indeterminate;
};


#include <cfloat>
#include <climits>

//
// Character specializations.
//
class numeric_limits<char>
{
public:
	static const bool is_specialized = true;

	inline static char min() { return CHAR_MIN; }
	inline static char max() { return CHAR_MAX; }

#ifdef __CHAR_UNSIGNED__
	static const int digits		= sizeof(char)*CHAR_BIT;
#else
	static const int digits		= sizeof(char)*CHAR_BIT -1;
#endif
	static const int digits10	= 2;

#ifdef __CHAR_UNSIGNED__
	static const bool is_signed 	= false;
#else
	static const bool is_signed 	= true;
#endif
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static char epsilon() { return 0; }
	inline static char round_error() { return 0; }

	static const int min_exponent	= 0;
	static const int min_exponent10	= 0;
	static const int max_exponent	= 0;
	static const int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static char infinity() { return 0; } 
	inline static char quiet_NaN() { return 0; } 
	inline static char signaling_NaN() { return 0; } 
	inline static char denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};

class numeric_limits<signed char>
{
public:
	static const bool is_specialized = true;

	inline static signed char min() { return CHAR_MIN; }
	inline static signed char max() { return CHAR_MAX; }

	static const int digits		= sizeof(signed char)*CHAR_BIT -1;
	static const int digits10	= 2;

	static const bool is_signed 	= true;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static signed char epsilon() { return 0; }
	inline static signed char round_error() { return 0; }

	static const int min_exponent	= 0;
	static const int min_exponent10	= 0;
	static const int max_exponent	= 0;
	static const int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static signed char infinity() { return 0; } 
	inline static signed char quiet_NaN() { return 0; } 
	inline static signed char signaling_NaN() { return 0; } 
	inline static signed char denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};

class numeric_limits<unsigned char>
{
public:
	static const bool is_specialized = true;

	inline static unsigned char min() { return 0; }
	inline static unsigned char max() { return UCHAR_MAX; }

	static const int digits		= sizeof(unsigned char)*CHAR_BIT;
	static const int digits10	= 2;

	static const bool is_signed 	= false;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static signed char epsilon() { return 0; }
	inline static signed char round_error() { return 0; }

	static const int min_exponent	= 0;
	static const int min_exponent10	= 0;
	static const int max_exponent	= 0;
	static const int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static signed char infinity() { return 0; } 
	inline static signed char quiet_NaN() { return 0; } 
	inline static signed char signaling_NaN() { return 0; } 
	inline static signed char denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};

class numeric_limits<wchar_t>
{
public:
  static const bool is_specialized = true;

  inline static wchar_t min() { 
    return 0; // WCHAR_MIN
  }

  inline static wchar_t max() { 
    return INT_MAX;  
    // WCHAR_MAX is defined buggy as (~WCHAR_MIN) which is -1 as wchar_t is int.
  }

  static const int digits = sizeof(wchar_t)*CHAR_BIT;
  static const int digits10 = 9;

  static const bool is_signed = false;
  static const bool is_integer = true;
  static const bool is_exact = true;

  static const int radix 		= 2;
  inline static signed char epsilon() { return 0; }
  inline static signed char round_error() { return 0; }
  
  static const int min_exponent	= 0;
  static const int min_exponent10 = 0;
  static const int max_exponent	= 0;
  static const int max_expontent10 = 0;

  static const bool has_infinity = false;
  static const bool has_quiet_NAN = false;
  static const bool has_signaling_NAN = false;
  static const bool has_denorm	= false;

  inline static signed char infinity() { return 0; } 
  inline static signed char quiet_NaN() { return 0; } 
  inline static signed char signaling_NaN() { return 0; } 
  inline static signed char denorm_min() { return 0; } 
  
  static const bool is_iec559 = false;
  static const bool is_bounded = true;
  static const bool is_modulo = true;
  static const bool traps = false;
  static const bool tinyness_before = false;

  static const float_round_style round_style = round_toward_zero;
};

//
// Integer specializations.
//
class numeric_limits<short>
{
public:
	static const bool is_specialized = true;

	inline static short min() { return SHRT_MIN; }
	inline static short max() { return SHRT_MAX; }

	static const int digits		= sizeof(short)*CHAR_BIT -1;
	static const int digits10	= 4;

	static const bool is_signed 	= true;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static short epsilon() { return 0; }
	inline static short round_error() { return 0; }

	static const int min_exponent	= 0;
	static const int min_exponent10	= 0;
	static const int max_exponent	= 0;
	static const int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static short infinity() { return 0; } 
	inline static short quiet_NaN() { return 0; } 
	inline static short signaling_NaN() { return 0; } 
	inline static short denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};

class numeric_limits<unsigned short>
{
public:
	static const bool is_specialized = true;

	inline static unsigned short min() { return 0; }
	inline static unsigned short max() { return USHRT_MAX; }

	static const int digits		= sizeof(short)*CHAR_BIT;
	static const int digits10	= 4;

	static const bool is_signed 	= false;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static unsigned short epsilon() { return 0; }
	inline static unsigned short round_error() { return 0; }

	static const int min_exponent	= 0;
	static const int min_exponent10	= 0;
	static const int max_exponent	= 0;
	static const int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static unsigned short infinity() { return 0; }
	inline static unsigned short quiet_NaN() { return 0; } 
	inline static unsigned short signaling_NaN() { return 0; } 
	inline static unsigned short denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};

class numeric_limits<int>
{
public:
	static const bool is_specialized = true;

	inline static int min() { return INT_MIN; }
	inline static int max() { return INT_MAX; }

	static const int digits		= sizeof(int)*CHAR_BIT -1;
	static const int digits10	= 9;

	static const bool is_signed 	= true;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static int epsilon() { return 0; }
	inline static int round_error() { return 0; }

	static const int min_exponent	= 0;
	static const int min_exponent10	= 0;
	static const int max_exponent	= 0;
	static const int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static int infinity() { return 0; } 
	inline static int quiet_NaN() { return 0; } 
	inline static int signaling_NaN() { return 0; } 
	inline static int denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};

class numeric_limits<unsigned int>
{
public:
	static const bool is_specialized = true;

	inline static unsigned int min() { return 0; }
	inline static unsigned int max() { return UINT_MAX; }

	static const int digits		= sizeof(int)*CHAR_BIT;
	static const int digits10	= 9;

	static const bool is_signed 	= false;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static unsigned int epsilon() { return 0; }
	inline static unsigned int round_error() { return 0; }

	static const unsigned int min_exponent	= 0;
	static const unsigned int min_exponent10	= 0;
	static const unsigned int max_exponent	= 0;
	static const unsigned int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static unsigned int infinity() { return 0; } 
	inline static unsigned int quiet_NaN() { return 0; } 
	inline static unsigned int signaling_NaN() { return 0; } 
	inline static unsigned int denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};

class numeric_limits<long>
{
public:
	static const bool is_specialized = true;

	inline static long min() { return LONG_MIN; }
	inline static long max() { return LONG_MAX; }

	static const int digits		= sizeof(long)*CHAR_BIT -1;
#if defined (__alpha__) || (defined (__sparc_v9__) && defined (__arch64__))
	static const int digits10	= 18;
#else
	static const int digits10	= 9;
#endif

	static const bool is_signed 	= true;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static long epsilon() { return 0; }
	inline static long round_error() { return 0; }

	static const int min_exponent	= 0;
	static const int min_exponent10	= 0;
	static const int max_exponent	= 0;
	static const int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static long infinity() { return 0; } 
	inline static long quiet_NaN() { return 0; } 
	inline static long signaling_NaN() { return 0; } 
	inline static long denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};

class numeric_limits<unsigned long>
{
public:
	static const bool is_specialized = true;

	inline static unsigned long min() { return 0; }
	inline static unsigned long max() { return ULONG_MAX; }

	static const int digits		= sizeof(unsigned long)*CHAR_BIT;
#if defined (__alpha__) || (defined (__sparc_v9__) && defined (__arch64__))
	static const int digits10	= 19;
#else
	static const int digits10	= 9;
#endif

	static const bool is_signed 	= false;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static long epsilon() { return 0; }
	inline static long round_error() { return 0; }

	static const int min_exponent	= 0;
	static const int min_exponent10	= 0;
	static const int max_exponent	= 0;
	static const int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static unsigned long infinity() { return 0; } 
	inline static unsigned long quiet_NaN() { return 0; } 
	inline static unsigned long signaling_NaN() { return 0; } 
	inline static unsigned long denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};

#if ( defined (__GNU_LIBRARY__) ? defined (__USE_GNU) : !defined (__STRICT_ANSI__) ) \
    && !defined(__alpha__)
class numeric_limits<long long>
{
public:
	static const bool is_specialized = true;

	inline static long long min() { return LONG_LONG_MIN; }
	inline static long long max() { return LONG_LONG_MAX; }

	static const int digits		= sizeof(long long)*CHAR_BIT;
	static const int digits10	= 18;

	static const bool is_signed 	= true;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static long long epsilon() { return 0; }
	inline static long long round_error() { return 0; }

	static const int min_exponent	= 0;
	static const int min_exponent10	= 0;
	static const int max_exponent	= 0;
	static const int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static long long infinity() { return 0; } 
	inline static long long quiet_NaN() { return 0; } 
	inline static long long signaling_NaN() { return 0; } 
	inline static long long denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};

class numeric_limits<unsigned long long>
{
public:
	static const bool is_specialized = true;

	inline static unsigned long long min() { return 0; }
	inline static unsigned long long max() { return ULONG_LONG_MAX; }

	static const int digits		= sizeof(unsigned long long)*CHAR_BIT;
	static const int digits10	= 19;

	static const bool is_signed 	= false;
	static const bool is_integer 	= true;
	static const bool is_exact 	= true;

	static const int radix 		= 2;
	inline static unsigned long long epsilon() { return 0; }
	inline static unsigned long long round_error() { return 0; }

	static const int min_exponent	= 0;
	static const int min_exponent10	= 0;
	static const int max_exponent	= 0;
	static const int max_expontent10 = 0;

	static const bool has_infinity	= false;
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = false;
	static const bool has_denorm	= false;

	inline static unsigned long long infinity() { return 0; } 
	inline static unsigned long long quiet_NaN() { return 0; } 
	inline static unsigned long long signaling_NaN() { return 0; } 
	inline static unsigned long long denorm_min() { return 0; } 

	static const bool is_iec559	= false;
	static const bool is_bounded	= true;
	static const bool is_modulo	= true;
	static const bool traps		= false;
	static const bool tinyness_before = false;

	static const float_round_style round_style = round_toward_zero;
};
#endif

//
// Floating point specializations.
//
class numeric_limits<float>
{
public:
	static const bool is_specialized = true;

	inline static float min() { return FLT_MIN; }
	inline static float max() { return FLT_MAX; }

	static const int digits		= FLT_MANT_DIG;
	static const int digits10	= FLT_DIG;

	static const bool is_signed 	= true;
	static const bool is_integer 	= false;
	static const bool is_exact 	= false;

	static const int radix 		= FLT_RADIX;
	inline static float epsilon() { return FLT_EPSILON; }
	inline static float round_error() { return 0.5f; }

	static const int min_exponent	= FLT_MIN_EXP;
	static const int min_exponent10	= FLT_MIN_10_EXP;
	static const int max_exponent	= FLT_MAX_EXP;
	static const int max_expontent10 = FLT_MAX_10_EXP;

	static const bool has_infinity	= true;
#ifdef __i386__
	static const bool has_quiet_NAN	= true;
	static const bool has_signaling_NAN = true;
	static const bool has_denorm	= true;
#endif
#ifdef __alpha__
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = true;
	static const bool has_denorm	= false;
#endif

	inline static float infinity();
	inline static float quiet_NaN();
	inline static float signaling_NaN();
        inline static float denorm_min() { 
#ifdef __i386__
	  unsigned val=0x1;
	  return *(reinterpret_cast<float*>(&val)); 
#else
	  return min();
#endif
	}

	static const bool is_iec559	= true;
	static const bool is_bounded	= true;
	static const bool is_modulo	= false;
	static const bool traps		= true;
	static const bool tinyness_before = true;

	static const float_round_style round_style = round_to_nearest;
};

inline float numeric_limits<float>::infinity()
{
	unsigned val = 0x7f800000;
	return *reinterpret_cast<float*>(&val);
} 

inline float numeric_limits<float>::quiet_NaN()
{
	unsigned val = 0x7fffffff;
	return *reinterpret_cast<float*>(&val);
} 

inline float numeric_limits<float>::signaling_NaN()
{
        unsigned val = 0x7fa00000;
	return *reinterpret_cast<float*>(&val);
} 

class numeric_limits<double>
{
public:
	static const bool is_specialized = true;

	inline static double min() { return DBL_MIN; }
	inline static double max() { return DBL_MAX; }

	static const int digits		= DBL_MANT_DIG;
	static const int digits10	= DBL_DIG;

	static const bool is_signed 	= true;
	static const bool is_integer 	= false;
	static const bool is_exact 	= false;

	static const int radix 		= FLT_RADIX;
	inline static double epsilon() { return DBL_EPSILON; }
	inline static double round_error() { return 0.5F; }

	static const int min_exponent	= DBL_MIN_EXP;
	static const int min_exponent10	= DBL_MIN_10_EXP;
	static const int max_exponent	= DBL_MAX_EXP;
	static const int max_expontent10 = DBL_MAX_10_EXP;

	static const bool has_infinity	= true;
#ifdef __i386__
	static const bool has_quiet_NAN	= true;
	static const bool has_signaling_NAN = true;
	static const bool has_denorm	= true;
#endif
#ifdef __alpha__
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = true;
	static const bool has_denorm	= false;
#endif

	inline static double infinity();
	inline static double quiet_NaN();
	inline static double signaling_NaN();
	inline static double denorm_min() { 
#ifdef __i386__
	  unsigned char val[]={0x1,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
	  return *(reinterpret_cast<double*>(&val)); 
#else
	  return min();
#endif
	} 

	static const bool is_iec559	= true;
	static const bool is_bounded	= true;
	static const bool is_modulo	= false;
	static const bool traps		= true;
	static const bool tinyness_before = true;

	static const float_round_style round_style = round_to_nearest;
};

inline double numeric_limits<double>::infinity()
{
	return static_cast<double>(numeric_limits<float>::infinity());
} 

inline double numeric_limits<double>::quiet_NaN()
{
	return static_cast<double>(numeric_limits<float>::quiet_NaN());
} 

inline double numeric_limits<double>::signaling_NaN()
{
	return static_cast<double>(numeric_limits<float>::signaling_NaN());
} 

class numeric_limits<long double>
{
public:
	static const bool is_specialized = true;

	inline static long double min() { return LDBL_MIN; }
	inline static long double max() { return LDBL_MAX; }

	static const int digits		= LDBL_MANT_DIG;
	static const int digits10	= LDBL_DIG;

	static const bool is_signed 	= true;
	static const bool is_integer 	= false;
	static const bool is_exact 	= false;

	static const int radix 		= FLT_RADIX;
	inline static long double epsilon() { return LDBL_EPSILON; }
	inline static long double round_error() { return 0.5L; }

	static const int min_exponent	= LDBL_MIN_EXP;
	static const int min_exponent10	= LDBL_MIN_10_EXP;
	static const int max_exponent	= LDBL_MAX_EXP;
	static const int max_expontent10 = LDBL_MAX_10_EXP;

	static const bool has_infinity	= true;
#ifdef __i386__
	static const bool has_quiet_NAN	= true;
	static const bool has_signaling_NAN = true;
	static const bool has_denorm	= true;
#endif
#ifdef __alpha__
	static const bool has_quiet_NAN	= false;
	static const bool has_signaling_NAN = true;
	static const bool has_denorm	= false;
#endif

	static long double infinity();
	static long double quiet_NaN();
	static long double signaling_NaN();
	inline static long double denorm_min() { 
#ifdef __i386__
	  unsigned char val[]={0x1,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
	  return *(reinterpret_cast<long double*>(&val)); 
#else
	  return min();
#endif
	} 

	static const bool is_iec559	= true;
	static const bool is_bounded	= true;
	static const bool is_modulo	= false;
	static const bool traps		= true;
	static const bool tinyness_before = true;

	static const float_round_style round_style = round_to_nearest;
};

inline long double numeric_limits<long double>::infinity()
{
	return static_cast<long double>(numeric_limits<float>::infinity());
} 

inline long double numeric_limits<long double>::quiet_NaN()
{
	return static_cast<long double>(numeric_limits<float>::quiet_NaN());
} 

inline long double numeric_limits<long double>::signaling_NaN()
{
	return static_cast<long double>(numeric_limits<float>::signaling_NaN());
} 

#endif

#endif

// Local Variables:
// mode:C++
// End:
