// This file implements methods for numeric optimization


/*
J.M.P. Franosch (Jan-Moritz.Franosch@physik.tu-muenchen.de)

This file is part of the GNUSSL software package.  This package is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This software is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this distribution; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef _optimize_h_
#define _optimize_h_


#include <cmath>
#include <limits>
#include <mMatrix.h>

#include <least_squares.h>
#include <eigenvectors.h>
#include <solve_linear.h>
#include <mathext.h>

#include <vector>




template<class Real, class function_cla>
void optimize_downhill_simplex(mMatrix<Real>& p, 
			       mVector<Real>& y, 
			       const function_cla& f,
			       int& nfunc,
			       Real ftol=0.0) {
// The function f() is minimized using the downhill simplex method by Nelder and Mead.
// Preconditions:
//   Real is a real type.
//   f is a function with the declaration 'Real f(const mVector<Real>& x)'
//   or an object of a class in which operator() is overloaded appropriately.
//   The matrix p is a n+1 x n matrix which contains 
//   n+1 n-dimensional starting-points in its rows.
//   The vector y is a n+1 dimensional vector initialized with y(i)=f(p.row(i)).
//   The algorithm returns when max_i(f(p.row(i)))-min_i(f(p.row(i)))<=ftol.
//   If ftol==0.0, ftol is set to numeric_limits<Real>::epsilon(), the machine accuracy.
// Result:
//   The first row p.row(1) of p contains a point where f has a local minimum and
//   y(1)=f(p.row(1)) contains the value of the function at this point.
//   nfunc contains the number of function calls that have been necessary.
  if (ftol==0.0) {
    ftol=numeric_limits<Real>::epsilon();
  };
  int n=y.size()-1;
  int mpts=y.size();
  mVector<Real> psum(n);
  nfunc=0;
  
  for (int j=1; j<=n; j++) {
    psum(j)=sum(p.col(j));
  };

  int nfunc_next_out=0;

  cout << "starting optimize_downhill_simplex" << endl;

  for (;;) {

    if (nfunc >= nfunc_next_out) {

      cout.precision(10);
      cout << "nfunc=" << nfunc << endl;
      cout << "y(1)= " << y(1) << endl;
      //      cout << "y= " << y << endl;
      cout << "p(1,1)= " << p(1,1) << endl;

      nfunc_next_out+=1000;
    };

    int lowest=1;
    int next_highest;
    int highest= y(1)>y(2) ? (next_highest=2,1) : (next_highest=1,2);
    for (int i=1; i<=mpts; i++) {
      if ( y(i)<=y(lowest) ) {
	lowest=i;
      };
      if ( y(i)>y(highest) ) {
	next_highest=highest;
	highest=i;
      } else {
	if ( y(i)>y(next_highest) && i!=highest ) {
	  next_highest=i;
	};
      };
    };
    if ( 2*abs(y(highest)-y(lowest))/(abs(y(highest))+abs(y(lowest)))<ftol ) {
      swap(y,y,1,lowest);
      swap(p,p,1,lowest);
      return;
    };
    nfunc+=2;
    Real ytry=amotry(p,y,psum,n,f,highest,Real(-1.0));
    if ( ytry<=y(lowest) ) {
      ytry=amotry(p,y,psum,n,f,highest,Real(2.0));
    } else {
      if ( ytry>=y(next_highest) ) {
	Real ysave=y(highest);
	ytry=amotry(p,y,psum,n,f,highest,Real(0.5));
	if (ytry>=ysave) {
	  for (int i=1; i<=mpts; i++) {
	    if (i!=lowest) {
	      for (int j=1; j<=n; j++) {
		p(i,j)=psum(j)=0.5*(p(i,j)+p(lowest,j));
	      };
	      y(i)=f(psum);
	    };
	  };
	  nfunc+=n;

	  for (int j=1; j<=n; j++) {
	    psum(j)=sum(p.col(j));
	  };

	};
      } else {
	--nfunc;
      };
    };
  };
};


template<class Real, class function_cla>
Real amotry(mMatrix<Real>& p, mVector<Real>& y, 
	    mVector<Real>& psum, 
	    int n, const function_cla& f, const int highest, 
	    const Real fac) {
  mVector<Real> ptry(n);
  Real fac1=(1.0-fac)/n;
  Real fac2=fac1-fac;
  for (int j=1; j<=n; j++) {
    ptry(j)=psum(j)*fac1-p(highest,j)*fac2;
  };
  Real ytry=f(ptry);
  if ( ytry<y(highest) ) {
    y(highest)=ytry;
    for (int j=1; j<=n; j++) {
      psum(j)+=ptry(j)-p(highest,j);
      p(highest,j)=ptry(j);
    };
  };
  return ytry;
};


template<class Real>
inline Real sign(const Real& a, const Real& b) {
  return b>0 ? abs(a) : -abs(a);
};


template<class Real>
inline void shift(Real& a, Real& b, Real& c, const Real& d) {
  a=b;
  b=c;
  c=d;
};


template<class Real, class function_cla>
void mnbrak(Real& ax, Real& bx, Real& cx, Real& fa, Real& fb, Real& fc,
	    const function_cla& f, int& nfunc) {
  // Computes a bracketing triplet of the function f.
  // Preconditions:
  //   ax and bx contain initial points.
  // Result:
  //   ax, bx, cx are set to points that bracket a minimum of the function f, 
  //   i.e. bx is between ax and cx and f(bx)<f(ax),f(cx).
  //   The function values fa=f(ax), fb=f(bx), fc=f(cx) at these three points 
  //   are also returned.
  const Real gold=1.618034;
  const Real tiny=numeric_limits<Real>::min();
  const Real glimit=100;
  nfunc=0;
  Real ulim,u,r,q,fu,dum;
  fa=f(ax);
  nfunc++;
  fb=f(bx);
  nfunc++;
  if (fb > fa) {
    shift(dum,ax,bx,dum);
    shift(dum,fb,fa,dum);
  }
  cx=(bx)+gold*(bx-ax);
  fc=f(cx);
  nfunc++;
  while (fb > fc) {
    r=(bx-ax)*(fb-fc);
    q=(bx-cx)*(fb-fa);
    u=(bx)-((bx-cx)*q-(bx-ax)*r)/
      (2.0*sign(max(fabs(q-r),tiny),q-r));
    ulim=(bx)+glimit*(cx-bx);
    if ((bx-u)*(u-cx) > 0.0) {
      fu=f(u);
      nfunc++;
      if (fu < fc) {
	ax=(bx);
	bx=u;
	fa=(fb);
	fb=fu;
	return;
      } else if (fu > fb) {
	cx=u;
	fc=fu;
	return;
      }
      u=(cx)+gold*(cx-bx);
      fu=f(u);
      nfunc++;
    } else if ((cx-u)*(u-ulim) > 0.0) {
      fu=f(u);
      nfunc++;
      if (fu < fc) {
	shift(bx,cx,u,cx+gold*(cx-bx));
	shift(fb,fc,fu,f(u));
	nfunc++;
      }
    } else if ((u-ulim)*(ulim-cx) >= 0.0) {
      u=ulim;
      fu=f(u);
      nfunc++;
    } else {
      u=(cx)+gold*(cx-bx);
      fu=f(u);
      nfunc++;
    }
    shift(ax,bx,cx,u);
    shift(fa,fb,fc,fu);
  }
};


template<class Real, class function_cla>
void mnbrak_limits(Real& ax, Real& bx, Real& cx, Real& fa, Real& fb, Real& fc,
	    const function_cla& f, int& nfunc, Real a_min, Real c_max) {
  // Computes a bracketing triplet of the function f.
  // Preconditions:
  //   ax and bx contain initial points.
  // Result:
  //   ax, bx, cx are set to points that bracket a minimum of the function f, 
  //   i.e. bx is between ax and cx and f(bx)<f(ax),f(cx).
  //   The function values fa=f(ax), fb=f(bx), fc=f(cx) at these three points 
  //   are also returned.
  //////
};


template<class Real, class function_cla>
Real optimize_line_Brent(const Real& ax, const Real& bx, const Real& cx, 
			 const function_cla& f, Real& xmin,
			 int& nfunc, const Real& ftol=0.0) {
  // The function f() is minimized using Brent's
  // (a quadratic interpolation) method.  
  // Preconditions:
  //   Real is a real type.
  //   f is a function with the declaration 'Real f(const Real& x)'
  //   or an object of a class in which operator() is overloaded appropriately.
  //   The values ax, bx, cx contain a bracketing triplet of abscissas, 
  //   i.e. bx is between ax and cx and f(bx)<f(ax),f(cx).
  //   The algorithm returns when the minimum is determined 
  //   with a fractional precision of about ftol.
  //   If ftol==0.0, ftol is set to sqrt(numeric_limits<Real>::epsilon()),
  //   the square root of the machine accuracy. 
  //   Note that ftol should not be set to a value smaller than this.
  // Result:
  //   xmin is set to a value where f has a local minimum.
  //   nfunc contains the number of function calls that have been necessary.
  const Real cgold=0.3819660;
  nfunc=0;
  Real e=0;
  Real a=min(ax,cx);
  Real b=max(ax,cx);
  Real x,w,v;
  x=w=v=bx;
  Real fw,fv,fx;
  fw=fv=fx=f(x);
  nfunc++;
  for (size_t iter=1; ; iter++) {
    Real xm=0.5*(a+b);
    Real tol1=ftol*abs(x)+numeric_limits<Real>::min();
    Real tol2=2*tol1;
    if ( abs(x-xm) <= tol2-0.5*(b-a) ) {
      xmin=x;
      return fx;
    };
    Real d=0, u=0;
    if ( abs(e)>tol1 ) {
      Real r=(x-w)*(fx-fv);
      Real q=(x-v)*(fx-fw);
      Real p=(x-v)*q-(x-w)*r;
      q=2*(q-r);
      if (q>0) {
	p=-p;
      };
      q=abs(q);
      Real etemp=e;
      e=d;
      if ( abs(p)>=abs(0.5*q*etemp) || p<=q*(a-x) || p>=q*(b-x) ) {
	e = (x>=xm ? a-x : b-x);
	d=cgold*e;
      } else {
	d=p/q;
	u=x+d;
	if ( u-a<tol2 || b-u<tol2 ) {
	  d=sign(tol1,xm-x);
	};
      };
    } else {
      e= (x>=xm ? a-x : b-x);
      d=cgold*e;
    };
    u= (abs(d)>=tol1 ? x+d : x+sign(tol1,d) );
    Real fu=f(u);
    nfunc++;
    if (fu<=fx) {
      if (u>=x) {
	a=x;
      } else {
	b=x;
      };
      v=w; w=x; x=u;
      fv=fw; fw=fx; fx=fu;
    } else {
      if (u<x) {
	a=u;
      } else {
	b=u;
      };
      if ( fu<=fw || w==x ) {
	v=w;
	w=u;
	fv=fw;
	fw=fu;
      } else if (fu<=fv || v==x || v==w) {
	v=u;
	fv=fu;
      };
    };
  };
};


template<class Real, class function_cla>
class f1dim_cla {
private:
  const mVector<Real> p, d;
  const function_cla& f;
public:
  f1dim_cla(const mVector<Real>& p_ini, const mVector<Real>& d_ini, 
	    const function_cla& f_ini) : 
    p(p_ini), d(d_ini), f(f_ini) {
  };
  inline Real operator()(const Real& t) const {
    mVector<Real> x(p.size());
    x=p+t*d;
    return f(x);
  };
};


template<class Real, class function_cla>
void optimize_line_Brent(mVector<Real>& p,
			 mVector<Real>& d,
			 Real& f_min,
			 const function_cla& f,
			 int& nfunc,
			 const Real& ftol) {
  // The function f() is minimized in direction d using Brent's
  // (a quadratic interpolation) method.  
  // Preconditions:
  //   Real is a real type.
  //   f is a function with the declaration 'Real f(const mVector<Real>& x)'
  //   or an object of a class in which operator() is overloaded appropriately.
  //   The vector p contains the starting point, 
  //   the vector d the direction of the minimization.
  //   ftol is given to the routine 
  //     'Real optimize_line_Brent(const Real& ax, const Real& bx, const Real& cx, 
  //                               const function_cla& f, Real& xmin,
  //                               int& nfunc, const Real& ftol)'
  //   which does the one-dimensional line optimization.
  // Result:
  //   p is set to p+t*d where t is determined so that the function t->f(p+t*d) 
  //   has a local minimum.
  //   d is set to t*d.
  //   f_min is set to f(p+t*d), the minimal value of the function f.
  //   nfunc contains the number of function calls that have been necessary.
  f1dim_cla<Real,function_cla> f1dim(p,d,f);
  nfunc=0;
  Real ax=0;
  Real xx=1;
  Real bx, fa, fx, fb;
  int nfunc2=0;
  mnbrak(ax,xx,bx,fa,fx,fb,f1dim,nfunc2);
  nfunc+=nfunc2;
  Real xmin;
  f_min=optimize_line_Brent(ax,xx,bx,f1dim,xmin,nfunc2,ftol);
  nfunc+=nfunc2;
  d *= xmin;
  p += d;
};


template<class Real, class function_cla>
void optimize_Powell(mVector<Real>& p, 
		     const function_cla& f,
		     int& nfunc,
		     Real ftol=0.0) {
  // The function f() is minimized using Powell's method.
  // Preconditions:
  //   Real is a real type.
  //   f is a function with the declaration 'Real f(const mVector<Real>& x)'
  //   or an object of a class in which operator() is overloaded appropriately.
  //   The vector p is the starting point of the minimization.
  //   The algorithm returns when f could not be decreased by more than 
  //   the fractional tolerance ftol in one iteration.
  //   If ftol==0.0, ftol is set to numeric_limits<Real>::epsilon(), 
  //   the machine accuracy.
  // Result:
  //   p contains a point where f has a local minimum.
  //   nfunc contains the number of function calls that have been necessary.
  nfunc=0;
  if (ftol==0.0) {
    ftol=numeric_limits<Real>::epsilon();
  };
  size_t n=p.size();
  mMatrix<Real> xi(n,n);
  xi=0;
  xi.diagonal()=1;
  mVector<Real> pt(n), ptt(n), xit(n);
  Real fret=f(p);
  nfunc++;
  for (size_t j=1; j<=n; j++) {
    pt(j)=p(j);
  };

  int nfunc_next_out=0;

  cout << "starting optimize_Powell" << endl;

  for (size_t iter=1; ; iter++) {

    if (nfunc >= nfunc_next_out) {
      cout.precision(10);
      cout << "nfunc=" << nfunc << endl;
      cout << "fret= " << fret << endl;
      cout << "p(1)= " << p(1) << endl;
      cout << "p= " << p << endl;
      nfunc_next_out+=1000;
    };

    Real fp=fret;
    size_t ibig=0;
    Real del=0;
    for (size_t i=1; i<=n; i++) {
      for (size_t j=1; j<=n; j++) {
	xit(j)=xi(j,i);
      };
      Real fret_old=fret;
      int nfunc2=0;
      optimize_line_Brent(p,xit,fret,f,nfunc2,sqrt(ftol));
      nfunc+=nfunc2;
      if ( abs(fret_old-fret)>del ) {
	del=abs(fret_old-fret);
	ibig=i;
      };
    };
    if ( 2*abs(fp-fret) <= ftol*(abs(fp)+abs(fret)) ) {
      return;
    };
    for (size_t j=1; j<=n; j++) {
      ptt(j)=2*p(j)-pt(j);
      xit(j)=p(j)-pt(j);
      pt(j)=p(j);
    };
    Real fptt=f(ptt);
    nfunc++;
    if (fptt<fp) {
      Real t=2*(fp-2*fret+fptt)*sqr(fp-fret-del)-del*sqr(fp-fptt);
      if (t<0) {
	int nfunc2=0;
	optimize_line_Brent(p,xit,fret,f,nfunc2,sqrt(ftol));
	nfunc+=nfunc2;
	for (size_t j=1; j<=n; j++) {
	  xi(j,ibig)=xi(j,n);
	  xi(j,n)=xit(j);
	};
      };
    };
  };
};


template<class Real>
class multidimensional_quadratic_basisfunction_cla {
  // Provides a basisfunction for multidimensional quadratic approximation.
  // Preconditions:
  //   'Real' is a real type.
private:
  size_t n;
public:
  multidimensional_quadratic_basisfunction_cla(const size_t n_ini) {
    // Preconditions:
    //   n_ini=x.size()
    n=n_ini;
  };
  size_t size() const {
    // Returns the number of basisfunctions.
    return (n*(n+1))/2+n+1;
  };
  Real operator()(size_t i, const mVector<Real>& x) const {
    // Returns the value of the basisfunction i at the point x.
    // Basisfunctions are the constant 1, The linear terms x(1)...x(n) 
    // the quadratic terms and x(i)*x(j), where i,j=1...n and n=x.size().
    // Preconditions:
    //   1<=i<=size()
    //   x.size()=n_ini
    if (i==1) {
      return 1;
    };
    i--;
    if (i<=n) {
      return x(i);
    };
    i-=n;
    size_t k=1;
    for (size_t l=1; i>l; i-=l, l++, k++) {
    };
    return x(k)*x(i);
  };
  void initialize_multidimensional_quadratic_approximation(const mVector<Real>& q,
    mMatrix<Real>& Q, mVector<Real>& a, Real& b) {
    // Initializes Q, a, b such that
    // sum(q(i)*basisfunction(i,x),i=1...size())=1/2 * x^T.Q.x + a^T.x + b .
    // Preconditions:
    //   Q is a n x n matrix, where n=n_ini.
    //   a is a n-dimensional vector.
    // Result:
    //   Q is symmetric.
    for (size_t j=1; j<=q.size(); j++) {
      size_t i=j;
      if (i==1) {
	b=q(j);
	continue;
      };
      i--;
      if (i<=n) {
	a(i)=q(j);
	continue;
      };
      i-=n;
      size_t k=1;
      for (; i>k; i-=k, k++) {
      };
      if (i!=k) {
	Q(i,k)=Q(k,i)=q(j);
      } else {
	Q(i,k)=2.0*q(j);
      };
    };
  };
};


template<class Real>
class multidimensional_quadratic_approximation_cla {
  // Implements a multidimensional quadratic approximation
  // Preconditions:
  //   'Real' is a real type.
private:
  mMatrix<Real> Q;
  mVector<Real> a;
  Real b;
public:
  template<class iterator>
  multidimensional_quadratic_approximation_cla(
    const iterator& begin, const mVector<Real>& y) {
    // Calculates the symmetric n x n matrix Q=quadratic_part(), 
    // the n-dimensional vector a=linear_part() and the real number 
    // b=constant_part() such that
    // the function with values y(i), i=1...N, at the points in the list
    // starting at begin is in the least squares sense best approximated by
    // 1/2 * x^T.Q.x + a^T.x + b .
    // Preconditions:
    //   N >= (n*(n+1))/2+n+1, where n is the dimension of the points in the list, 
    //   i.e. n=(*begin).size()
    //   The points *begin ... in the list are of type mVector<Real>.
    int n=(*begin).size();
    multidimensional_quadratic_basisfunction_cla<Real> basisfunction(n);
    mVector<Real> q(basisfunction.size());
    least_squares(begin,y,basisfunction,q);
    a.reserve(n);
    Q.reserve(n,n);
    basisfunction.initialize_multidimensional_quadratic_approximation(q,Q,a,b);
  };
  inline const mMatrix<Real>& quadratic_part() const {
    return Q;
  };
  inline const mVector<Real>& linear_part() const {
    return a;
  };
  inline const Real& constant_part() const {
    return b;
  };
};


template<class Real>
class multidimensional_quadratic_approximation_with_principals_cla :
  public multidimensional_quadratic_approximation_cla<Real> {
  // Same as multidimensional_quadratic_approximation,
  // but also calculates the principal directions and principal values of
  // Q=quadratic_part().
private:
  mVector<Real> v;
  mMatrix<Real> EV;
  mVector<Real> m;
public:
  template<class iterator>
  multidimensional_quadratic_approximation_with_principals_cla(
    const iterator& begin, const mVector<Real>& y) :
    multidimensional_quadratic_approximation_cla(begin,y) {
    // The same as multidimensional_quadratic_approximation
    int n=linear_part().size();
    mMatrix< complex<Real> > EV_(n,n);
    mVector< complex<Real> > v_(n);
    eigenvectors_of_real_matrix(quadratic_part(),EV_,v_);
    v.reserve(n);
    EV.reserve(n,n);
    for (int i=1; i<=n; i++) {
      v(i)=v_(i).real();
      for (int j=1; j<=n; j++) {
	EV(i,j)=EV_(i,j).real();
      };
    };
    m.reserve(n);
    solve_linear(quadratic_part(),linear_part(),m);
    m*=Real(-1.0);
  };
  inline const mVector<Real>& principal_values() const {
    // Returns the principal values of Q=quadratic_part().
    return v;
  };
  inline const mMatrix<Real>& principal_directions() const {
    // Returns the principal directions of Q=quadratic_part().
    // Result:
    //   The columns of the returned matrix contain the normed 
    //   principal directions of Q=quadratic_part().
    return EV;
  };
  inline const mVector<Real>& midpoint() const {
    // Returns the midpoint of the paraboloid 1/2 x^T.Q.x + a^T.x + b,
    // i.e. the solution of grad(1/2 x^T.Q.x + a^T.x + b) = Q.x + a = 0 .
    return m;
  };
};


template<class Real, class function_cla>
int optimize_quadratic_approximation(const mVector<Real>& x0, const Real R_ini,
				     const function_cla& f,
				     mVector<Real>& x_min,
				     int& nfunc,
				     Real tol=0.0) {
  // The function f is minimized using iterated quadratic approximations of f.
  // Preconditions:
  //   'Real' is a real type.
  //   f is a function object that takes m mVector<Real> as argument 
  //   and returns a 'Real' type.
  //   x0 is the starting point.
  //   R is the initial radius in which values of f are calculated.
  //   The sphere ||x-x0||<=R should contain the minimum of f.
  // Result:
  //   x_min contains the minimum of f with an accuracy of tol.
  //   If tol=0.0 (the default value), tol is set to numeric_limits<Real>::epsilon(),
  //   i.e. the machine accuracy.
  if (tol==Real(0.0)) {
    tol=numeric_limits<Real>::epsilon();
  };
  Real R=R_ini;
  nfunc=0;
  int numpoints=( x0.size()*(x0.size()+1)/2+x0.size()+1 ); ////
  vector< mVector<Real> > points(numpoints);
  mVector<Real> y(numpoints);
  mVector<Real> m(x0.size());
  m=x0;
  Real accuracy=1;

  mVector<Real> principal_values(x0.size());
  principal_values=1;
  mMatrix<Real> principal_directions(x0.size(),x0.size());
  principal_directions=0;
  principal_directions.diagonal()=1;

  int step=0;

  while (true) {

    cout << "step " << step << endl;
    step++;

    for (int i=1; i<=numpoints; i++) {
      points[i-1].reserve(x0.size());
      points[i-1]=m;
      for (size_t j=1; j<=x0.size(); j++) {
	if ( principal_values(j)>0 && R/sqrt(abs(principal_values(j)))<R_ini ) {
	  points[i-1]+=R*frand(-1,1)/sqrt(abs(principal_values(j)))
	    *principal_directions.col(j);
	} else {
	  points[i-1]+=R_ini*frand(-1,1)
	    *principal_directions.col(j);
	};
      };
      y(i)=f(points[i-1]);
    };
    nfunc+=numpoints;
    accuracy=max(y)-min(y);

    cout << "accuracy=" << accuracy << endl;

    if (accuracy<=tol) {
      break;
    };

    multidimensional_quadratic_approximation_with_principals_cla<Real>
      qapp(points.begin(),y);

    mVector<Real> r(x0.size());
    dot(qapp.midpoint()-m,qapp.principal_directions(),r);

    //    cout << "r=" << r << endl;

    for (size_t i=1; i<=r.size(); i++) {
      r(i)*=sqrt(abs(qapp.principal_values()(i)));
    };

    R=2*sqrt(norm(r));

    cout << "R=" << R << endl;

    m=qapp.midpoint();

    
    cout << "m=" << m << endl;
    cout << "f(m)=" << f(m) << endl;
    
    //    cout << "Q=" << qapp.quadratic_part() << endl;
    //    cout << "principal_directions=" << qapp.principal_directions() << endl;
        cout << "principal_values=" << qapp.principal_values() << endl;
    //    cout << "a=" << qapp.linear_part() << endl;

    principal_directions=qapp.principal_directions();
    principal_values=qapp.principal_values();

  };
  x_min=m;
  return 1;
};




#endif _optimize_h_
