// linear least squares approximation


/*
J.M.P. Franosch (Jan-Moritz.Franosch@physik.tu-muenchen.de)

This file is part of the GNUSSL software package.  This package is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This software is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this distribution; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef _least_squares_h_
#define _least_squares_h_


#include <limits>
#include <lin_alg.h>




template<class iterator, class basisfunction_cla, class Real, class S1, class S2>
mVector<Real,S2>&
least_squares(const iterator& begin,
	      const mVector<Real,S1>& y,
	      const basisfunction_cla& basis_function,
	      mVector<Real,S2>& a,
	      Real tol=0.0) {
  // Computes the linear least squares approximation of the set of data points 
  // (x_1,y(1)) ... (x_N,y(N)) where N=y.size() and 
  // the list starting at begin contains the points x_i.
  // Preconditions:
  //   'Real' is a real type.
  //   The list starting at begin contains y.size() 'points'.
  //   The vector y contains the function values at these points.
  //   basis_function is a function object such that basis_function(m,*i), 
  //   where i is an iterator in the list and 1<=m<=a.size() returns a 'Real' type.
  // Result:
  //   a contains the coefficients of the linear least square approximation problem, i.e.
  //   a is computed such that 
  //   sum( [ sum(a(j)*basis_function(j,*(begin+i)),j=1...a.size()) - y(i) ]^2, i=1...N )
  //   is minimized.
  //   a is returned as a reference.
  if (tol==Real(0.0)) {
    tol=numeric_limits<Real>::epsilon();
  };
  int M=a.size();
  int N=y.size();
  mMatrix<Real> A(N,M);
  iterator i=begin;
  for (int n=1; n<=N; i++, n++) {
    for (int m=1; m<=M; m++) {
      A(n,m)=basis_function(m,*i);
    };
  };
  mMatrix<Real> u(N,M);
  u=0;
  u.diagonal()=1;
  mMatrix<Real> v(N,M);
  v=0;
  v.diagonal()=1;
  mVector<Real> d(M);
  pseudo_inverse(A,u,d,v,tol);
  dot(A,y,a);
  return a;
};




#endif _least_squares_h_
