// This file implements solve_linear,
// solving a system of linear equations

/*
J.M.P. Franosch (Jan-Moritz.Franosch@physik.tu-muenchen.de)

This file is part of the GNUSSL software package.  This package is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This software is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this distribution; if not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef _solve_linear_h_
#define _solve_linear_h_

#ifdef HAVE_LIMITS
#include <limits>
#else
#include "numeric_limits.h"
#endif

#include "lin_alg.h"




template<class T, class Real, class S1, class S2, class S3>
mVector<T,S3>& 
solve_linear(const mMatrix<T,S1>& A, const mVector<T,S2>& b,
	     mVector<T,S3>& x,
	     Real tol=0.0) {
// The Matrix equation A.x=b is solved in the least square sense 
// by singular value decomposition,
// i.e. if the equation has no solutions, the euclidian norm ||A.x-b|| is minimized,
// else x with the minimal ||x|| is chosen that solves the equation.
// Preconditions: 
//   'T' is a real type.
//   A is a n x m dimensional matrix (m>=n),
//   x is a n dimensional vector and b is a m dimensional vector.
// Return:
//   x is the solution and returned as a reference.
#ifdef GNUSSL_CHECK
  if (A.csize()!=x.size()) {
    gnussl_error("solve_linear","A.rsize()!=x.size()");
  };
  if (A.rsize()!=b.size()) {
    gnussl_error("solve_linear","A.csize()!=b.size()");
  };
#endif
  if (tol==T(0.0)) {
    tol=std::numeric_limits<Real>::epsilon();
  };
  int n=x.size();
  int m=b.size();
  mMatrix<T> u(m,n);
  u=0;
  u.diagonal()=T(1);  // if u is square, u is the identity matrix
  mMatrix<T> v(n,n);
  v=0;
  v.diagonal()=T(1);  // v is the identity matrix
  mVector<Real> d(n);
  mMatrix<T> opm(m,n);
  opm=A;
  pseudo_inverse(opm,u,d,v,tol);
  dot(opm,b,x);
  return x;
};




#endif // _solve_linear_h_
