/***************************************************************************
                          pointer.h  -  description
                             -------------------
    begin                : Tue Dec 17 2002
    author               : Moritz Franosch
    copyright            : (C) 2002 by ASKOS http://www.ASKOS.de
    email                : mail@Franosch.org
 ***************************************************************************/

/*************************************************************************** 
  Modifizierte Version von 

  shared_ptr.hpp
  (C) Copyright Greg Colvin and Beman Dawes 1998, 1999.
  Copyright (c) 2001, 2002 Peter Dimov

  detail/shared_count.hpp
  Copyright (c) 2001, 2002 Peter Dimov and Multi Media Ltd.

  Permission to copy, use, modify, sell and distribute this software
  is granted provided this copyright notice appears in all copies.
  This software is provided "as is" without express or implied
  warranty, and with no claim as to its suitability for any purpose.

  See http://www.boost.org/libs/smart_ptr/shared_ptr.htm for
  documentation.
****************************************************************************/

#ifndef POINTER_H
#define POINTER_H

#include "askosio/errorcheck.h"


class counted_base {

private:

  long use_count_; // Anzahl der Referenzen.

public:

  counted_base() :
    use_count_(0) { 
  }

  explicit counted_base(long initial_use_count):
    use_count_(initial_use_count) {
  }

  virtual ~counted_base() {
  }

  void add_ref() {
#ifdef ASKOS_ERRORCHECK
    if ( use_count_==0 ) {
      FATAL_ERROR_TEXT("Use count is 0.");
    }
#endif
    ++use_count_;
  }

  void release() {
    --use_count_;
    if (use_count_ == 0) {
      delete this;
    }
  }

  long use_count() const {
    return use_count_;
  }

private:

  counted_base(counted_base const &);

  counted_base& operator=(counted_base const &);

};


template<typename T>
class shared_count : public counted_base {

private:

  T* pointee;

public:

  shared_count() : counted_base(1) {
#ifdef ASKOS_ERRORCHECK
    pointee=0;
#endif
  }

  virtual ~shared_count() {
    delete pointee;
  }

  shared_count(T* p) : 
    counted_base(1),
    pointee(p) {
  }

  bool unique() const {
    return use_count() == 1;
  }

  friend inline bool operator==(shared_count const & a, shared_count const & b)
  {
    return a.pointee == b.pointee;
  }

  friend inline bool operator<(shared_count const & a, shared_count const & b)
  {
    return a.pointee<b.pointee;
  }

  const T& operator*() const {
    return *pointee;
  }

  T& operator*() {
    return *pointee;
  }

  T* get() const {
    return pointee;
  }

};




/** Pointer mit Referenzzaehlung.
 *
 *  Das referenzierte Objekt wird mit delete zerstoert, wenn der
 *  letzte Pointer zerstoert wird (bzw. beim letzten Pointer reset()
 *  aufgerufen wird).
 */
template<typename T>
class Pointer {

private:

  shared_count<T>* pn; // Referenzzaehler

public:

  typedef T element_type;
  typedef T value_type;

  Pointer() : pn(new shared_count<T>(0)) {
  }

  Pointer(const Pointer<T>& p) : pn(p.pn) {
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereference.");
    }
#endif
    pn->add_ref();
  }

  explicit Pointer(T* p) : pn(new shared_count<T>(p)) {
  }
 
  ~Pointer() {
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereference.");
    }
#endif
    pn->release();
  }

  Pointer& operator=(const Pointer& p) {
    if ( pn != p.pn ) {
#ifdef ASKOS_ERRORCHECK
      if (pn == 0) {
	FATAL_ERROR_TEXT("Nullpointer dereference.");
      }
#endif
      pn->release(); // ruft 'delete pn' auf, wenn der Referenzzaehler 0 wird
      pn=p.pn;
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereference.");
    }
#endif
      pn->add_ref();
    }
    return *this;
  }

  void reset() {
    pn->release();
#ifdef ASKOS_ERRORCHECK
    pn=0;
#endif
  }

  const T& operator*() const {
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereference.");
    }
#endif
    return **pn;
  }
  
  T& operator*() {
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereference.");
    }
#endif
    return **pn;
  }
  
  T* operator->() const {
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereferenced.");
    }
#endif
    return pn->get();
  }
    
  T* get() const {
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereferenced.");
    }
#endif
    return pn->get();
  }
  
  operator bool() const {
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereferenced.");
    }
#endif
    return pn->get() != 0;
  }

  bool operator!() const {
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereferenced.");
    }
#endif
    return pn->get() != 0;
  }

  bool unique() const {
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereferenced.");
    }
#endif
    return pn->unique();
  }

  long use_count() const {
#ifdef ASKOS_ERRORCHECK
    if (pn == 0) {
      FATAL_ERROR_TEXT("Nullpointer dereferenced.");
    }
#endif
    return pn->use_count();
  }

};


template<typename T, typename U> inline bool operator==(Pointer<T> const & a, Pointer<U> const & b)
{
    return a.get() == b.get();
}

template<typename T, typename U> inline bool operator!=(Pointer<T> const & a, Pointer<U> const & b)
{
    return a.get() != b.get();
}

/*
#if __GNUC__ == 2 && __GNUC_MINOR__ <= 96

// Resolve the ambiguity between our op!= and the one in rel_ops

template<typename T> inline bool operator!=(Pointer<T> const & a, Pointer<T> const & b)
{
    return a.get() != b.get();
}

#endif
*/

template<typename T> inline bool operator<(Pointer<T> const & a, Pointer<T> const & b)
{
    return a.get() < b.get();
}



#endif // POINTER_H
