/***************************************************************************
                          sbool.h  -  description
                             -------------------
    begin                : Thu Dec 12 2002
    author               : Christof Schatz, Moritz Franosch
    copyright            : (C) 2002 by ASKOS http://www.ASKOS.de
    email                : info@askos.de
 ***************************************************************************/

#ifndef SBOOL_H
#define SBOOL_H

#include <ostream>


/** Boolean fuer Statistiken.
 *
 *  Kann die Werte true, false, missing und undefined annehmen.
 *
 *  Implementiert eine vierwertige Logik:
 *
 *  Die Wertemenge {true, false} bildet die klassische zweiwertige Logik.
 *
 *  Die Wertemenge {true, false, missing} bildet die dreiwertige Logik:
 *
 *  Sei f eine n-stellige logische Funktion (die Operatoren &&, ||
 *  sind zweistellige logische Funktionen, der Operator ! ist eine
 *  einstellige logische Funktion). Seien x1, x2, ..., xn, xi aus
 *  {true, false, missing} die Parameter der Funktion und f(x1,...,xn)
 *  der Funktionswert. Die Funktion sei fuer die zweiwertige Logik
 *  definiert, also fuer alle Kombinationen der Parameter aus {true,
 *  false}. Die dreiwertige Erweiterung von f ergibt sich dann wie
 *  folgt: Haben keine Parameter den Wert missing, dann ist der Wert
 *  der dreiwertigen Erweiterung von f natuerlich gleich dem Wert des
 *  zweiwertigen f. Hat mindestens ein Parameter den Wert missing, so
 *  hat das dreiwertige f den Wert true/false, falls das zweiwertige f
 *  unabhaengig von der Belegung der "missing"-Parameter mit true oder
 *  false den Wert true/false hat. Gibt es dagegen mindestens eine
 *  Belegung der "missing"-Parameter, so dass das zweiwertige f den
 *  Wert true und mindestens eine andere Belegungen, so dass das
 *  zweiwertige f den Wert false hat, so hat das dreiwertige f den
 *  Wert missing.
 *
 *  Beispiel: true || missing == true, true && missing == missing,
 *  !missing == missing.
 *
 *  {true, false, missing, undefined} bilden eine vierwertige Logik:
 *
 *  Sei f eine n-stellige dreiwertige logische Funktion. Dann hat die
 *  vierwertige Erweiterung den gleichen Funktionswert wie f, falls
 *  alle Parameter aus {true, false, missing} sind. Falls mindestens
 *  ein Parameter undefined ist, hat das vierwertige f den Wert
 *  undefined.
 */
class Sbool {

private:

  enum value_t {true_v, false_v, missing_v, undefined_v};

  value_t value;

  Sbool(const value_t v);

public: 

  /** Ein Sbool, der den vierwertigen Wert "missing" hat.
   */
  static const Sbool missing;

  /** Ein Sbool, der den vierwertigen Wert "undefined" hat.
   */
  static const Sbool undefined;

  /** Konstruktor.
   *
   *  Setzt den Sbool auf den vierwertigen Wert "undefined".
   */
  Sbool();
  
  /** Konstruktor.
   *
   *  Setzt den Sbool auf den vierwertigen Wert "true", falls b==true,
   *  andernfalls "false".
   */
  Sbool(const bool b);
  
  /** Copy-Konstruktor.
   *
   *  Postcondition: *this==sbool.
   */
  Sbool(const Sbool& sbool);
  
  /** Liefert true, falls der vierwertige Wert "true" ist, andernfalls
   *  false.
   */
  bool is_true() const;

  /** Liefert false, falls der vierwertige Wert "false" ist, andernfalls
   *  false.
   */
  bool is_false() const;

  /** Liefert true, falls der vierwertige Wert "missing" ist, andernfalls
   *  false.
   */
  bool is_missing() const;

  /** Liefert true, falls der vierwertige Wert "undefined" ist,
   *  andernfalls false.
   */
  bool is_undefined() const;

  /** True, falls dieser Sbool weder den Wert missing noch den Wert
   *  undefined hat.
   */
  bool is_valid() const;

  /** Weist dem Sbool den Wert von sbool zu.
   *
   *  Liefert eine Referenz auf *this zurueck.
   *
   *  Postcondition: *this==sbool.
   */
  Sbool& operator=(const Sbool& sbool);

  /** Vierwertige Erweiterung des Operators ! (Negation).
   */
  const Sbool operator!() const;

  /** Automatische Typumwandlung zu bool.
   *
   *  Gibt is_true() zurueck.
   */
  operator bool() const;

};


/** Vierwertige Erweiterung des Operators && (und).
 */
const Sbool operator&&(const Sbool& sbool1, const Sbool& sbool2);


/** Vierwertige Erweiterung des Operators || (oder).
 */
const Sbool operator||(const Sbool& sbool1, const Sbool& sbool2);


/** Vierwertige Erweiterung des Operators == (Identitaet).
 */
const Sbool operator==(const Sbool& sbool1, const Sbool& sbool2);


/** Ausgabe von Sbool.
 */
std::ostream& operator<<(std::ostream& o, const Sbool& sbool);


#endif // SBOOL_H
