/***************************************************************************
                          sdouble.cpp  -  description
                             -------------------
    begin                : Thu Dec 12 2002
    author               : Christof Schatz, Moritz Franosch
    copyright            : (C) 2002 by ASKOS http://www.ASKOS.de
    email                : info@askos.de
 ***************************************************************************/

#include <limits>
#include <cmath>

#ifdef __BORLANDC__
  // Borland C++ kennt kein ISO C99 isnan (!),
  // allerdings _isnan.
#include <float.h>
inline int isnan(double x) {
  return _isnan(x);
}
#endif

#include "sdouble.h"

#include "askosio/errorcheck.h"


using namespace std;


// Borland C++ kennt kein NAN und numeric_limits<double>::quiet_NaN()
// ist bei GCC vor Version 3.3 fehlerhaft (gleich 0).
#ifndef __GNUC__
  #define NAN numeric_limits<double>::quiet_NaN()
#endif


const Sdouble Sdouble::missing(NAN,true);
// C99 (?) NAN statt numeric_limits<double>::quiet_NaN()
// wg. Bug in gcc 3.2.


const Sdouble Sdouble::undefined;


Sdouble::Sdouble(const double value_ini, const bool is_missing_b_ini) :
  value(value_ini),
  is_missing_b(is_missing_b_ini) {
}


Sdouble::Sdouble() :
  value( NAN ),
// C99 (?) NAN statt numeric_limits<double>::quiet_NaN()
// wg. Bug in gcc 3.2.
  is_missing_b(false) {
}


Sdouble::Sdouble(const double& r) :
  value(r),
  is_missing_b(false) {
#ifdef ASKOS_ERRORCHECK
  if ( isnan(r) && (!is_undefined()) ) {
    FATAL_ERROR;
  }
#endif
}


double Sdouble::to_double() const {
#ifdef ASKOS_ERRORCHECK
  if ( is_missing() && !isnan(value) ) {
    FATAL_ERROR;
  }
#endif
  return value;
}


Sdouble& Sdouble::operator=(const Sdouble& r) {
  value=r.value;
  is_missing_b=r.is_missing_b;
  return *this;
}


bool Sdouble::is_missing() const {
  return is_missing_b;
}


bool Sdouble::is_undefined() const {
  return (!is_missing_b) && isnan(value);
}


bool Sdouble::is_valid() const {
  return !is_missing() && !is_undefined();
}


const Sbool operator==(const Sdouble& d1, const Sdouble& d2) {
  if ( d1.is_undefined() || d2.is_undefined() ) {
    return Sbool::undefined;
  }
  if ( d1.is_missing() || d2.is_missing() ) {
    return Sbool::missing;
  }
  return d1.to_double()==d2.to_double();
}


const Sbool operator!=(const Sdouble& d1, const Sdouble& d2) {
  if ( d1.is_undefined() || d2.is_undefined() ) {
    return Sbool::undefined;
  }
  if ( d1.is_missing() || d2.is_missing() ) {
    return Sbool::missing;
  }
  return d1.to_double()!=d2.to_double();
}


const Sbool operator<(const Sdouble& d1, const Sdouble& d2) {
  if ( d1.is_undefined() || d2.is_undefined() ) {
    return Sbool::undefined;
  }
  if ( d1.is_missing() || d2.is_missing() ) {
    return Sbool::missing;
  }
  return d1.to_double()<d2.to_double();
}


const Sbool operator>(const Sdouble& d1, const Sdouble& d2) {
  if ( d1.is_undefined() || d2.is_undefined() ) {
    return Sbool::undefined;
  }
  if ( d1.is_missing() || d2.is_missing() ) {
    return Sbool::missing;
  }
  return d1.to_double()>d2.to_double();
}


const Sbool operator<=(const Sdouble& d1, const Sdouble& d2) {
  if ( d1.is_undefined() || d2.is_undefined() ) {
    return Sbool::undefined;
  }
  if ( d1.is_missing() || d2.is_missing() ) {
    return Sbool::missing;
  }
  return d1.to_double()<=d2.to_double();
}


const Sbool operator>=(const Sdouble& d1, const Sdouble& d2) {
  if ( d1.is_undefined() || d2.is_undefined() ) {
    return Sbool::undefined;
  }
  if ( d1.is_missing() || d2.is_missing() ) {
    return Sbool::missing;
  }
  return d1.to_double()>=d2.to_double();
}


const Sdouble operator+(const Sdouble& d1, const Sdouble& d2) {
  if ( d1.is_undefined() || d2.is_undefined() ) {
    return Sdouble::undefined;
  }
  if ( d1.is_missing() || d2.is_missing() ) {
    return Sdouble::missing;
  }
  return d1.to_double()+d2.to_double();
}


const Sdouble operator-(const Sdouble& d1, const Sdouble& d2) {
  if ( d1.is_undefined() || d2.is_undefined() ) {
    return Sdouble::undefined;
  }
  if ( d1.is_missing() || d2.is_missing() ) {
    return Sdouble::missing;
  }
  return d1.to_double()-d2.to_double();
}


const Sdouble operator*(const Sdouble& d1, const Sdouble& d2) {
  if ( d1.is_undefined() || d2.is_undefined() ) {
    return Sdouble::undefined;
  }
  if ( d1.is_missing() || d2.is_missing() ) {
    return Sdouble::missing;
  }
  return d1.to_double()*d2.to_double();
}


const Sdouble operator/(const Sdouble& d1, const Sdouble& d2) {
  if ( d1.is_undefined() || d2.is_undefined() ) {
    return Sdouble::undefined;
  }
  if ( ( d2==Sdouble(0.0) ).is_true() ) {
    return Sdouble::undefined;
  }
  if ( d1.is_missing() || d2.is_missing() ) {
    return Sdouble::missing;
  }
  return d1.to_double()/d2.to_double();
}


ostream& operator<<(ostream& o, const Sdouble& sdouble) {
  if ( sdouble.is_undefined() ) {
    o << "undefined";
    return o;
  }
  if ( sdouble.is_missing() ) {
    o << "missing";
    return o;
  }
  // Wir sind weder undefined noch missing, die Konvertierung muss
  // also ein non-NaN double zurueckliefern.
  o << sdouble.to_double();
  return o;
}
