/***************************************************************************
                          sdouble.h  -  description
                             -------------------
    begin                : Thu Dec 12 2002
    author               : Christof Schatz, Moritz Franosch
    copyright            : (C) 2002 by ASKOS http://www.ASKOS.de
    email                : info@askos.de
 ***************************************************************************/

#ifndef SDOUBLE_H
#define SDOUBLE_H


#include <ostream>

#include "sbool.h"




/** Double fuer Statistiken.
 *
 *  Werte koennen doubles sein oder "missing" oder "undefined". Der
 *  Sinn von "missing" ist "Der Wert existiert zwar, ist aber
 *  unbekannt, nicht gemessen, fehlerhaft gemessen o.ae.". Der Sinn
 *  von "undefined" ist "Der Wert existiert nicht." oder "Es ist nicht
 *  sinnvoll, von diesem Wert zu sprechen.". Beispielsweise ist die
 *  Koerpergroesse des Deutschen Kaisers "undefined", denn Deutschland
 *  hat keinen Kaiser. Die Koepergroesse des Ulmer Muenster ist
 *  ebenfalls "undefined", denn das Ulmer Muenster hat keine
 *  Koerpergroesse, allenfalls eine Kirchturmhoehe. Die Zahl aller
 *  lebenden Individuen auf der Erde ist gegenwaertig "missing".
 *
 *  Funktionen auf doubles werden folgendermassen erweitert:
 *
 *  Sind alle Parameter weder "missing" noch "undefined", ist der
 *  Funktionswert der gleiche wie derjenige der auf doubles
 *  definierten Funktion.
 *
 *  Ist mindestens ein Parameter "missing" und kein Parameter
 *  "undefined", ist der Funktionswert "missing".
 *
 *  Ist mindestens ein Parameter "undefined" ist der Funktionswert
 *  "undefined".
 *
 *  Durch Verwendung von Sdoubles soll eine groessere Robustheit der
 *  entsprechenden Routinen erreicht werden.
 */
class Sdouble {

private:

  double value; // Wert dieses Sdoubles.
  
  // True: Dieser Sdouble ist missing. Der Wert von value muss in
  // diesem Fall NaN sein.
  // False: Dieser Sdouble ist undefined, falls value den Wert NaN
  // hat, ansonsten hat dieser Sdouble den Wert value.
  //
  // Sinn: Sollte eine arithmetische Operation value auf NaN setzen,
  // bekommt der Sdouble automatisch den Wert undefined.
  bool is_missing_b;

  Sdouble(const double value_ini, const bool is_missing_b_ini);

public: 

  /** Ein Sdouble, das den vierwertigen Wert "missing" hat.
   */
  static const Sdouble missing;

  /** Ein Sdouble, das den vierwertigen Wert "undefined" hat.
   */
  static const Sdouble undefined;

  /** Konstruktor.
   *
   *  Initialisiert den Sdouble mit "undefined".
   */
  Sdouble();

  /** Konstruktor.
   *
   *  Initialisiert den Sdouble mit dem Wert von r.
   *
   *  Ist r NaN ( isnan(r)==true ), so hat der Sdouble den Wert
   *  "undefined".
   */
  Sdouble(const double& r);

  /** Typumandlung in double.
   *
   *  Ist NaN (not a number), falls das Sdouble den Wert "missing"
   *  oder "undefined" hat, ansonsten gleich dem Wert.
   */
  double to_double() const;

  /** Weist dem Sdouble den Wert von r zu.
   *
   *  Liefert eine Referenz auf *this zurueck.
   *
   *  Postcondition: *this==r.
   */
  Sdouble& operator=(const Sdouble& r);

  /** True, falls dieser Sdouble den Wert missing hat.
   */
  bool is_missing() const;

  /** True, falls dieser Sdouble den Wert undefined hat.
   */
  bool is_undefined() const;

  /** True, falls dieser Sdouble weder den Wert missing noch den Wert
   *  undefined hat.
   */
  bool is_valid() const;

};


/** "Undefined", falls d1 oder d2 "undefined" sind.
 *
 *  Ansonsten:
 *  "Missing", falls d1 oder d2 "missing" sind.
 *
 *  Ansonsten:
 *  d1.to_double()==d2.to_double().
 */
const Sbool operator==(const Sdouble& d1, const Sdouble& d2);


/** "Undefined", falls d1 oder d2 "undefined" sind.
 *
 *  Ansonsten:
 *  "Missing", falls d1 oder d2 "missing" sind.
 *
 *  Ansonsten:
 *  d1.to_double()<d2.to_double().
 */
const Sbool operator<(const Sdouble& d1, const Sdouble& d2);


/** "Undefined", falls d1 oder d2 "undefined" sind.
 *
 *  Ansonsten:
 *  "Missing", falls d1 oder d2 "missing" sind.
 *
 *  Ansonsten:
 *  d1.to_double()>d2.to_double().
 */
const Sbool operator>(const Sdouble& d1, const Sdouble& d2);


/** "Undefined", falls d1 oder d2 "undefined" sind.
 *
 *  Ansonsten:
 *  "Missing", falls d1 oder d2 "missing" sind.
 *
 *  Ansonsten:
 *  d1.to_double()<=d2.to_double().
 */
const Sbool operator<=(const Sdouble& d1, const Sdouble& d2);


/** "Undefined", falls d1 oder d2 "undefined" sind.
 *
 *  Ansonsten:
 *  "Missing", falls d1 oder d2 "missing" sind.
 *
 *  Ansonsten:
 *  d1.to_double()>=d2.to_double().
 */
const Sbool operator>=(const Sdouble& d1, const Sdouble& d2);


/** "Undefined", falls d1 oder d2 "undefined" sind.
 *
 *  Ansonsten:
 *  "Missing", falls d1 oder d2 "missing" sind.
 *
 *  Ansonsten:
 *  d1.to_double()!=d2.to_double().
 */
const Sbool operator!=(const Sdouble& d1, const Sdouble& d2);


/** "Undefined", falls d1 oder d2 "undefined" sind.
 *
 *  Ansonsten:
 *  "Missing", falls d1 oder d2 "missing" sind.
 *
 *  Ansonsten:
 *  d1.to_double()+d2.to_double().
 */
const Sdouble operator+(const Sdouble& d1, const Sdouble& d2);


/** Analog zu operator+().
 */
const Sdouble operator-(const Sdouble& d1, const Sdouble& d2);


/** Analog zu operator+().
 */
const Sdouble operator*(const Sdouble& d1, const Sdouble& d2);


/** "Undefined", falls d2==0, sonst analog zu operator+().
 */
const Sdouble operator/(const Sdouble& d1, const Sdouble& d2);


/** Ausgabe von Sdouble.
 */
std::ostream& operator<<(std::ostream& o, const Sdouble& sdouble);


#endif // SDOUBLE_H
