/***************************************************************************
                          tree.h  -  description
                             -------------------
    begin                : Tue Apr 9 2002
    author               : Moritz Franosch
    copyright            : (C) 2002 by ASKOS http://www.ASKOS.de
    email                : mail@Franosch.org
 ***************************************************************************/

#ifndef TREE_H
#define TREE_H

#include <list>
#include <iostream>

#include "askosio/errorcheck.h"


using namespace std;


template<class T>
class Tree_Iterator;


template<class T>
class Tree {

public:

  typedef Tree_Iterator<T> iterator;
  typedef T leaf_t;
  typedef Tree<T> Tree_t;

  Tree();

  iterator begin();

  iterator end();

  bool has_parent() const;

  bool is_leaf() const;

  const T& get_leaf() const;

  bool empty() const;

private:

  ////  Tree(const Tree&);

  Tree& operator=(const Tree&);

  void set_leaf(const T& leaf);

  T* leaf_ptr;
  list<Tree_t> subtrees;
  Tree_t* parent_ptr;
  typename list<Tree_t>::iterator parent_it;

  friend class Tree_Iterator<T>;

};


template<class T>
Tree<T>::Tree() : leaf_ptr(0), subtrees(), parent_ptr(0), parent_it() {
}


template<class T>
typename Tree<T>::iterator Tree<T>::begin() {
  return iterator(this);
}


template<class T>
typename Tree<T>::iterator Tree<T>::end() {
  return iterator(this,subtrees.end());
}


template<class T>
bool Tree<T>::has_parent() const {
  return parent_ptr!=0;
}


template<class T>
bool Tree<T>::is_leaf() const {
  return leaf_ptr!=0;
}


template<class T>
const T& Tree<T>::get_leaf() const {
  return *leaf_ptr;
}


template<class T>
bool Tree<T>::empty() const {
  return subtrees.empty();
}


template<class T>
void Tree<T>::set_leaf(const T& leaf) {
  leaf_ptr=new T(leaf);
}




template<class T>
class Tree_Iterator {

public:

  typedef T leaf_t;
  typedef Tree<T> Tree_t;

  void next();

  void step();

  void up();

  Tree_Iterator insert(const T& leaf);

  Tree_Iterator insert_tree();

  bool at_end() const;
  
  bool operator==(const Tree_Iterator& it2) const;

  bool operator!=(const Tree_Iterator& it2) const;

  Tree_t& operator*();

  Tree_t* operator->();

  const Tree_t& operator*() const;

private:

  Tree_Iterator<T>(Tree_t* tree_ptr_ini);

  Tree_Iterator<T>(Tree<T>* tree_ptr_ini,
		   const typename list<Tree_t>::iterator& it_ini);

  Tree_t* tree_ptr;
  typename list<Tree_t>::iterator it;

  friend class Tree<T>;

};


template<class T>
Tree_Iterator<T>::Tree_Iterator(Tree<T>* tree_ptr_ini) :
  tree_ptr(tree_ptr_ini), it(tree_ptr->subtrees.begin()) {
}


template<class T>
Tree_Iterator<T>::Tree_Iterator(Tree<T>* tree_ptr_ini,
			        const typename list<Tree_t>::iterator& it_ini) :
  tree_ptr(tree_ptr_ini), it(it_ini) {
}


template<class T>
void Tree_Iterator<T>::next() {
  if ( at_end() ) {
    if (tree_ptr->parent_ptr==0) {
      FATAL_ERROR_TEXT("At the end of subtree, but tree has no parent.");
    }
    it=tree_ptr->parent_it;
    tree_ptr=tree_ptr->parent_ptr;
  }
  it++;
}


template<class T>
void Tree_Iterator<T>::up() {
  if ( tree_ptr->parent_ptr==0 ) {
    FATAL_ERROR_TEXT("Tree has no parent.");
  }
  it=tree_ptr->parent_it;
  tree_ptr=tree_ptr->parent_ptr;
}


template<class T>
void Tree_Iterator<T>::step() {
  if ( at_end() || 
       ( it->is_leaf() && it->empty() ) ) {
    next();
    return;
  }
  tree_ptr=&(*it); // address of the tree I point to
  it=tree_ptr->subtrees.begin();
}


template<class T>
Tree_Iterator<T> Tree_Iterator<T>::insert(const T& leaf) {
  typename list<Tree_t>::iterator new_it=(tree_ptr->subtrees.insert(it,Tree_t()));
  new_it->parent_ptr=tree_ptr;
  new_it->parent_it=new_it;
  new_it->set_leaf(leaf);
  return Tree_Iterator<T>(tree_ptr,new_it);
}


template<class T>
Tree_Iterator<T> Tree_Iterator<T>::insert_tree() {
  typename list<Tree_t>::iterator new_it=(tree_ptr->subtrees.insert(it,Tree_t()));
  new_it->parent_ptr=tree_ptr;
  new_it->parent_it=new_it;
  return Tree_Iterator<T>(tree_ptr,new_it);
}


template<class T>
bool Tree_Iterator<T>::at_end() const {
  return it==tree_ptr->subtrees.end();
}


template<class T>
bool Tree_Iterator<T>::operator==(const Tree_Iterator& it2) const {
  return tree_ptr==it2.tree_ptr && it==it2.it;
}


template<class T>
bool Tree_Iterator<T>::operator!=(const Tree_Iterator& it2) const {
  return !operator==(it2);
}


template<class T>
Tree<T>& Tree_Iterator<T>::operator*() {
  if ( at_end() ) {
    FATAL_ERROR_TEXT("Tree iterator at end, dereferencing with * not possible.");
  }
  return *it;
}


template<class T>
Tree<T>* Tree_Iterator<T>::operator->() {
  if ( at_end() ) {
    FATAL_ERROR_TEXT("Tree iterator at end, dereferencing with -> not possible.");
  }
  return &(*it);
}


template<class T>
const Tree<T>& Tree_Iterator<T>::operator*() const {
  if ( at_end() ) {
    FATAL_ERROR_TEXT("Tree iterator at end, dereferencing with * not possible.");
  }
  return *it;
}


template<class T>
ostream& operator<<(ostream& out, Tree<T>& tree) {
  typename Tree<T>::iterator it=tree.begin();
  bool insert_space=false;
  while ( it!=tree.end() ) {
    if ( it.at_end() ) {
      out << ')';
      insert_space=true;
    } else if ( it->is_leaf() ) {
      if (insert_space) {
	out << " ";
      }
      out << it->get_leaf();
      insert_space=true;
    } else {
      if (insert_space) {
	out << " ";
      }
      out << '(';
      insert_space=false;
    }
    it.step();
  }
  return out;
}


#endif // TREE_H
