/***************************************************************************
                          casc.cpp  -  description
                             -------------------
    begin                : Mon Dec 24 2001
    copyright            : (C) 2001 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <fstream>
#include <string>

#include "casc.h"
#include "string_convert.h"




void Casc::error(const string& message) const {
  cerr << "Casc: " << message << endl;
  exit(EXIT_FAILURE);
}


Casc::Casc(istream& in_ini) : 
  in(in_ini), 
  strings(), doubles(), ints(),
  string_lists(), double_lists(), int_lists(),
  parsed(false), type(undefined_type) {
}


Casc::~Casc() {
  if (!parsed) {
    return;
  }
  switch (type) {
  case double_type:
  case int_type: {
    if ( !in.eof() ) {
      error("Not read to end of file.");
      return;
    }
    break;
  }
  case ascii_type: {
    in >> ws;
    if ( !in.eof() ) {
      error("Not read to end of file.");
    }
    return;
  }
  default: {
    error("type not set");
  }
  }
}


string Casc::format_directory="";


void Casc::set_format_directory(const string& directory) {
  format_directory=directory;
}


void Casc::parse(const string& format) {
  char c;
  if ( !in.get(c) ) {
    error("Could not read character.");
  }
  if (c!='#') {
    error("First character must be '#' (was '"+string_convert::to_string(c)+"').");
  }
  string casc;
  if ( !(in >> casc) ) {
    error("Read '#'. Could not read string.");
  }
  if ( casc!="casc" ) {
    error("Read '# "+casc+"'. File must begin with '# casc'.");
  }
  string definition_filename;
  if ( !(in >> definition_filename) ) {
    error("Read '# casc'. Could not read string.");
  }
  if (format!="" && definition_filename!=format) {
    error("Expected '"+format+"' as format of casc file, but got '"+definition_filename+
	  "'.");
  }
  definition_filename+=".cascformat";
  ifstream definition;
  string definition_filename2 = format_directory+"/"+definition_filename;
  definition.open(definition_filename2.c_str());
  if ( !definition ) {
    // klappt bei gcc 3.2 nicht mehr
    definition.open(definition_filename.c_str());
    if ( !definition ) {
      error("Could not open definition file '"+definition_filename+"' nor '"+
	    definition_filename2+"' for reading.");
    }
  }
  definition >> ws;
  in >> ws;
  while ( definition.get(c) ) {
    if ( c!='#' ) {
      error("Expected '#' in definition file '"+definition_filename+"', got '"+c+"'.");
    }
    string parameter;
    if ( !(definition >> parameter) ) {
      error("Read '#'. Could not read parameter from definition file '"+
	    definition_filename+"'.");
    }
    string parameter_type;
    if ( !(definition >> parameter_type) ) {
      error("Read '# "+parameter+
	    "'. Could not read parameter type from definition file '"+
	    definition_filename+"'.");
    }
    if ( !in.get(c) ) {
      error("Could not read character.");
    }
    if ( c!='#' ) {
      error("Expected '#', got '"+string_convert::to_string(c)+"'.");
    }
    string parameter2;
    if ( !(in >> parameter2) ) {
      error("Read '#'. Could not read parameter name.");
    }
    if (parameter!=parameter2) {
      error("Read '"+parameter2+"' as parameter, but got '"+parameter+
	    "' from definition file '"+definition_filename+"'.");
    }
    string line;
    if ( !getline(in,line) ) {
      error("Read '# "+parameter2+"'. Could not read until end of line.");
    }
    if (parameter_type=="string") {
      read(parameter,line,strings);
    } else if (parameter_type=="double") {
      read(parameter,line,doubles);
    } else if (parameter_type=="int") {
      read(parameter,line,ints);
    } else if (parameter_type=="string_list") {
      read_list(parameter,line,string_lists);
    } else if (parameter_type=="double_list") {
      read_list(parameter,line,double_lists);
    } else if (parameter_type=="int_list") {
      read_list(parameter,line,int_lists);
    } else {
      error("Read '# "+parameter2+" "+parameter_type+"' from definition file '"+
	    definition_filename+"'. No such parameter type '"+parameter_type+"'.");
    }
    definition >> ws;
    in >> ws;
  }
  if ( !definition.eof() ) {
    error("Read error occured when reading definition file '"+definition_filename+"'.");
  }
  if ( !in.get(c) ) {
    error("Could not read character.");
  }
  if ( c!='#' ) {
    error("Expected '#', got '"+string_convert::to_string(c)+"'.");
  }
  string data_type;
  if ( !(in >> data_type) ) {
    error("Read '#'. Could not read data type.");
  }
  if (data_type=="double") {
    type=double_type;
  } else if (data_type=="ascii") {
    type=ascii_type;
  } else {
    error("Read # '"+data_type+"'. No such data type '"+data_type+"'.");
  }
  in >> ws;
  if ( !in ) {
    error("Read error occured.");
  }
  parsed=true;
}


const string& Casc::get_string(const string& name) const {
  return get(strings,name);
}


const double& Casc::get_double(const string& name) const {
  return get(doubles,name);
}


const int& Casc::get_int(const string& name) const {
  return get(ints,name);
}


const list<double>& Casc::get_double_list(const string& name) const {
  return get(double_lists,name);
}


const list<int>& Casc::get_int_list(const string& name) const {
  return get(int_lists,name);
}


double Casc::read() {
  if (!parsed) {
    error("read() called before parse().");
  }
  switch (type) {
  case double_type: {
    double r;
    if ( !in.read(reinterpret_cast<char*>(&r),sizeof(double)) ) {
      error("read() of double failed.");
    }
    return r;
    break;
  }
  case ascii_type: {
    double r;
    if ( !(in >> r) ) {
      error("read() of ascii failed.");
    }
    return r;
  }
  default: {
    error("type not set");
  }
  }
  return 0;
}


int Casc::read_int() {
  if (!parsed) {
    error("read_int() called before parse().");
  }
  switch (type) {
  case int_type: {
    int32_t i;
    if ( !in.read(reinterpret_cast<char*>(&i),sizeof(int32_t)) ) {
      error("read_int() of int failed.");
    }
    return i;
    break;
  }
  case ascii_type: {
    int i;
    if ( !(in >> i) ) {
      error("read_int() of ascii failed.");
    }
    return i;
  }
  default: {
    error("type not set");
  }
  }
  return 0;
}
