/***************************************************************************
                          commandline.cpp  -  description
                             -------------------
    begin                : Sun Mar 18 2001
    copyright            : (C) 2001 by Moritz Franosch
    email                : mail@franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <strstream>
#include <cerrno>
#include <iostream>

#include "commandline.h"
#include "string_convert.h"




void Commandline::err_argument_not_specified(const string& s) const {
  cerr << get_myname() << ": Argument '" 
       << s << "' not specified." << endl
       << "Use '" << get_myname() << " --help' for help." << endl;
  exit(EXIT_FAILURE);
}


void Commandline::err_argument_specified(const string& s, argp_state* state) {
  cerr << get_myname() << ": Argument '" 
       << s << "' already specified (value '" << get_string(s) << "')." << endl;
  argp_usage(state);
}


void Commandline::set_string(const char* s, const char* p, argp_state* state) {
  check_not_specified(s,strings,state);
  strings[s]=p;
}


void Commandline::set_int(const char* s, const char* p, argp_state* state) {
  check_not_specified(s,ints,state);
  char* tailptr=0;
  errno=0;
  ints[s]=strtol(p,&tailptr,0);
  if (tailptr==p) {
    cerr << get_myname() << ": Argument of option '--" << s 
	 << "' must be an integer, was '" << p << "'." << endl;
    argp_usage(state);
  }
  if (errno) {
    cerr << get_myname() << ": Argument of option '--" << s 
	 << "' out of range for an integer, was '" << p << "'." << endl;
    argp_usage(state);
  }
}


void Commandline::set_double(const char* s, const char* p, argp_state* state) {
  check_not_specified(s,doubles,state);
  char* tailptr=0;
  errno=0;
  doubles[s]=strtod(p,&tailptr);
  if (tailptr==p) {
    cerr << get_myname() << ": Argument of option '--" << s 
	 << "' must be a float, was '" << p << "'." << endl;
    argp_usage(state);
  }
  if (errno) {
    cerr << get_myname() << ": Argument of option '--" << s 
	 << "' out of range for a double, was '" << p << "'." << endl;
    argp_usage(state);
  }
}


void Commandline::set_switch(const char* s, argp_state* state) {
  check_not_specified(s,switches,state);
  switches[s]=true;
}


void Commandline::set_list_double(const char* s, const char* p, argp_state* state) {
  check_not_specified(s,lists_double,state);
  istrstream argstream(p);
  double r;
  while (argstream >> r) {
    lists_double[s].push_back(r);
    argstream >> ws;
    char c=0;
    argstream.get(c);
    if (c==',') {
      argstream >> ws;
    } else {
      argstream.unget();
    }
  }
  if ( !argstream.eof() ) {
    cerr << get_myname() << ": Error while reading list of doubles (list was '"
	 << p << "')." << endl;
    argp_usage(state);
  }
}


void Commandline::set_list_string(const char* s, const char* p, argp_state* state) {
  check_not_specified(s,lists_double,state);
  istrstream argstream(p);
  string r;
  while (argstream >> r) {
    lists_string[s].push_back(r);
    argstream >> ws;
  }
  if ( !argstream.eof() ) {
    cerr << get_myname() << ": Error while reading list of strings (list was '"
	 << p << "')." << endl;
    argp_usage(state);
  }
  clog << "Setting list of strings '" << s << "' to '";
  for (list<string>::const_iterator i=lists_string[s].begin(); 
       i!=lists_string[s].end(); i++) {
    if ( i!=lists_string[s].begin() ) { 
      clog << " ";
    }
    clog << *i;
  }
  clog << "'." << endl;
}


bool Commandline::specified_string(const string& s) {
  return specified(s,strings);
}


bool Commandline::specified_int(const string& s) {
  return specified(s,ints);
}


bool Commandline::specified_double(const string& s) {
  return specified(s,doubles);
}


bool Commandline::specified_list_double(const string& s) {
  return specified(s,lists_double);
}


bool Commandline::specified_list_string(const string& s) {
  return specified(s,lists_string);
}


const string& Commandline::get_string(const string& s) {
  check_specified(s,strings);
  return strings[s];
}


const string& Commandline::get_string(const string& s, const string& d) {
  if ( specified_string(s) ) {
    return get_string(s);
  }
  return d;
}


int Commandline::get_int(const string& s) {
  check_specified(s,ints);
  return ints[s];
}


int Commandline::get_int(const string& s, const int d) {
  if ( specified_int(s) ) {
    return get_int(s);
  }
  return d;
}


double Commandline::get_double(const string& s) {
  check_specified(s,doubles);
  return doubles[s];
}


double Commandline::get_double(const string& s, const double d) {
  if ( specified_double(s) ) {
    return get_double(s);
  }
  return d;
}


bool Commandline::get_switch(const string& s) {
  if ( !specified(s,switches) ) {
    return false;
  }
  return switches[s];
}


const list<double>& Commandline::get_list_double(const string& s) {
  check_specified(s,lists_double);
  return lists_double[s];
}


const list<string>& Commandline::get_list_string(const string& s) {
  check_specified(s,lists_string);
  return lists_string[s];
}


size_t Commandline::get_nfilenames() const {
  return filenames.size();
}


const string& Commandline::get_filename(const size_t i) const {
  if ( i>=filenames.size() ) {
    cerr << get_myname() << ": Program bug. Filename Nr. " << i << " out of range (" 
	 << filenames.size()-1 << " max)." << endl;
  }
  return filenames[i];
}


Commandline::Commandline(const int argc, char *argv[],
			 const char* version_ini, const char* bug_address,
			 const char* doc, const char* args_doc,
			 const struct argp_option options_ini[]) :
  myname(string_convert::basename(argv[0])),
  mydir(string_convert::dirname(argv[0])),
  version(version_ini),
  strings(),
  ints(),
  doubles(),
  switches(),
  lists_double(), lists_string(),
  filenames(),
  options(options_ini) {
  argp_program_version=version_ini;
  argp_program_bug_address=bug_address;
  struct argp argp_parser={options,parse_opt,args_doc,doc,0,0,0};
  argp_parse(&argp_parser,argc,argv,0,0,this);
}


const string& Commandline::get_myname() const {
  return myname;
}


const string& Commandline::get_mydir() const {
  return mydir;
}


const string& Commandline::get_version() const {
  return version;
}


error_t Commandline::parse(int key, char* argument, argp_state* state) {
  if (key==ARGP_KEY_ARG) {
    filenames.push_back(string(argument));
    return 0;
  }
  if (key==ARGP_KEY_END) {
    return 0;
  }
  int i=0;
  while ( options[i].key!=0 ) {
    if (options[i].key==key) {
      if ( options[i].name==0 ) {
	cerr << get_myname() << ": Program bug. Long option not provided." << endl;
	exit(EXIT_FAILURE);
      }
      if (options[i].arg==0) {
	set_switch(options[i].name,state);
	return 0;
      }
      string arg(options[i].arg);
      if ( arg.find("(string)")<=arg.size() ) {
	set_string(options[i].name,argument,state);
	return 0;
      }
      if ( arg.find("(integer)")<=arg.size() ) {
	set_int(options[i].name,argument,state);
	return 0;
      }
      if ( arg.find("(float)")<=arg.size() ) {
	set_double(options[i].name,argument,state);
	return 0;
      }
      if ( arg.find("(list of floats)")<=arg.size() ) {
	set_list_double(options[i].name,argument,state);
	return 0;
      }
      if ( arg.find("(list of strings)")<=arg.size() ) {
	set_list_string(options[i].name,argument,state);
	return 0;
      }
      cerr << get_myname() 
	   << ": Program bug. Option type not specified or unknown, arg was '" 
	   << arg << "'" << endl;
      exit(EXIT_FAILURE);
    }
    i++;
  }
  return ARGP_ERR_UNKNOWN;
}


error_t parse_opt(int key, char* argument, argp_state* state) {
  return static_cast<Commandline*>(state->input)->parse(key,argument,state);
}
