/***************************************************************************
                          log.h  -  description
                             -------------------
    begin                : Tue Apr 9 2002
    author               : Moritz Franosch
    copyright            : (C) 2002 by ASKOS http://www.ASKOS.de
    email                : mail@Franosch.org
 ***************************************************************************/

#ifndef LOG_H
#define LOG_H

#include <string>

#include "logdevice.h"
#include "nullostream.h"


using namespace std;


/** Globales Logging.
 *
 *  Beispiel:
 *  Log log;
 *  log.error("Text");
 */
class Log {

private:

  static Logdevice default_logdevice; // globales Logdevice

  // Ausgabestream, der Ausgaben ignoriert.
  static Nullostream nullostream;

  /** Gibt alles vor "::" zurueck bis zum ersten Whitespace.
   */
  static string get_classname(const string& function_name);

  /** Gibt alles nach "::" zurueck.
   */
  static string get_methodname(const string& function_name);

private:

  Logdevice& logdevice; // verwendet dieses Logdevice

  string classname; // Name der Klasse, in der Log() aufegrufen wurde.

  string methodname; // Name der Methode, in der Log() aufegrufen wurde.

  string class_method; // "Klasse::Methode"

  // Die verschiedenen ostreams.
  //  ostream out_info, out_warn, out_error, out_debug, out_internal;

  // Ausgabestream
  ostream* output_ptr;

  int depth; // aktuelle Verschachtelungstiefe von Log()

  string indentation; // wird zur Einrueckung ausgegeben

  bool log_debug; 
  // Debugging-Modus. Enter- und Exit-Meldungen werden ausgegeben.
  
  ostream& header(ostream& out, const string& text);

public: 

  /** Gibt das Default-Logdevice zurueck.
   */
  static Logdevice& get_default_logdevice();

  /** Loggt in das Logdevice logdevice.
   *
   *  Beispiel:
   *  C::method() {
   *      Log log(&logdevice,"C","method");
   *  }
   */
  Log(Logdevice* logdevice, const string& classname_ini, const string& methodname_ini);

  /** Wie Log(Logdevice* logdevice, const string& classname_ini, const string& methodname_ini),
   *  nutzt nur globales Logdevice.
   */
  Log(const string& classname_ini, const string& methodname_ini);

  /** Wie Log(const string& classname_ini, const string&
   *  methodname_ini), nur hat function_name den Aufbau
   *  Klasse::Memberfunktion.
   */
  Log(const string& function_name);

  /** Wie Log(const string& classname_ini, const string& methodname_ini),
   *  nur mit unbekannten Klassen- und Methodennamen.
   */
  Log();

  /** Destruktor.
   */
  virtual ~Log();

  /** Interner Fehler.
   *
   *  Es liegt ein Programmierfehler vor.
   */
  ostream& internal_error();

  /** Fehler.
   *
   *  Fehlerhafte Benutzereingabe.
   */
  ostream& error();

  /** Warnung an den Benutzer.
   */
  ostream& warn();

  /** Information fuer den Benutzer.
   */
  ostream& info();

  /** Debugging-Information fuer den Programmierer.
   */
  ostream& debug();

  /** Gibt eine Warnung in warn() aus und beendet dann das Programm
   *  mit EXIT_FAILURE.
   */
  void error_exit();

  /** Wirft die Exception Askos_Fatal_Exception(text).
   */
  virtual void fatal_error(const string& message="");

};


/** Macro: Definiert eine Instanz log der Klasse Log mit dem
 *  default-Logdevice. Klasse und Methode werden durch das Macro
 *  automatisch eingetragen.
 */
#ifdef __GNUC__
#define LOG(log) \
  Log log(__PRETTY_FUNCTION__);
#else
#define LOG(log) \
  Log log(__FUNC__);
#endif

#endif // LOG_H
