/***************************************************************************
                          statistics.cpp  -  description
                             -------------------
    begin                : Wed Nov 27 2002
    author               : Moritz Franosch
    copyright            : (C) 2002 by ASKOS http://www.ASKOS.de
    email                : mail@Franosch.org
 ***************************************************************************/


#include <cmath>
#include <limits>

#include "askosio/errorcheck.h"
#include "askosmath.h"
#include "statistics.h"


using namespace std;




double cumulative_binomial(const int N, const int n, const double p) {
  if (n<=0) {
    return 1;
  }
  if (n>N) {
    return 0;
  }
  return incomplete_beta(n,N-n+1,p);
}


double incomplete_beta(const double a, const double b, double x) {
#ifdef ASKOS_ERRORCHECK
  if (a<0 || b<0) {
    FATAL_ERROR_TEXT("a<0 or b<0");
  }
#endif
  if (a==0 || b==0) {
    return 1;
  }
  if (x<0) {
    x=0;
  }
  if (x>1) {
    x=1;
  }
  double bt;
  if (x==0 || x==1) {
    bt=0;
  } else {
    bt=exp(lgamma(a+b)-lgamma(a)-lgamma(b)+a*log(x)+b*log(1-x));
  }
  if ( x < (a+1)/(a+b+2) ) {
    return bt*beta_continued_fraction(a,b,x)/a;
  }
  return 1-bt*beta_continued_fraction(b,a,1-x)/b;
}


double beta_continued_fraction(const double a, const double b, const double x) {
  double qab=a+b;
  double qap=a+1;
  double qam=a-1;
  double c=1;
  double d=1-qab*x/qap;
  if ( abs(d) < numeric_limits<double>::min() ) {
    d=numeric_limits<double>::min();
  }
  d=1.0/d;
  double h=d;
  int maxit=100;
  for (int m=1; m<=maxit; ++m) {
    int m2=2*m;
    double aa=m*(b-m)*x/((qam+m2)*(a+m2));
    d=1.0+aa*d;
    if ( abs(d) < numeric_limits<double>::min() ) {
      d=numeric_limits<double>::min();
    }
    c=1.0+aa/c;
    if ( abs(c) < numeric_limits<double>::min() ) {
      c=numeric_limits<double>::min();
    }
    d=1.0/d;
    h *= d*c;
    aa = -(a+m)*(qab+m)*x/((a+m2)*(qap+m2));
    d=1.0+aa*d;
    if ( abs(d) < numeric_limits<double>::min() ) {
      d=numeric_limits<double>::min();
    }
    c=1.0+aa/c;
    if ( abs(c) < numeric_limits<double>::min() ) {
      c=numeric_limits<double>::min();
    }
    d=1.0/d;
    double del=d*c;
    h *= del;
    if ( abs(del-1.0) < numeric_limits<double>::epsilon() ) {
      return h;
    }
  }
  //// fixme
  FATAL_ERROR_TEXT("Parameter a or b too big, or maxit too small.");
  return h;
}


double normal(const double z) {
  return 1.0/sqrt(2*pi) * exp(-0.5*sqr(z));
}


double cumulative_normal(const double z) {
  return 0.5*erf(z/sqrt(2.0))+0.5;
}


void incomplete_gamma_series(const double a, const double x, double& gamser, double& gln) {
  /*
  cerr << "a=" << a << endl;
  cerr << "x=" << x << endl;
  */
  int itmax=100;
  gln=lgamma(a);
  //  cerr << "gln=" << gln << endl;
  if (x <= 0.0) {
    if (x < 0.0) {
      FATAL_ERROR_TEXT("x less than 0.");
    }
    gamser=0.0;
    return;
  } else {
    double ap=a;
    double sum=1.0/a;
    double del=sum;
    for (int n=1;n<=itmax;n++) {
      ap += 1.0;
      del *= x/ap;
      sum += del;
      /*
      cerr << "del=" << del << endl;
      cerr << "sum=" << sum << endl;
      */
      if ( abs(del) < abs(sum)*numeric_limits<double>::epsilon() ) {
	gamser=sum*exp(-x+a*log(x)-gln);
	return;
      }
    }
    //// fixme
    FATAL_ERROR_TEXT("a too large, or itmax too small.");
    gamser=numeric_limits<double>::quiet_NaN(); // fixme
    return;
  }
}


void incomplete_gamma_continued_fraction(const double a, const double x, double& gamcf, double& gln) {
  int itmax=100;
  double gold=0.0;
  double fac=1.0;
  double b1=1.0;
  double b0=0.0;
  double a0=1.0;
  double a1=x;
  gln=lgamma(a);
  if ( x >= numeric_limits<double>::max() ) {
    gamcf=0;
    return;
  }
  for (int n=1; n<=itmax; ++n) {
    double an=double(n);
    double ana=an-a;
    a0=(a1+a0*ana)*fac;
    b0=(b1+b0*ana)*fac;
    double anf=an*fac;
    a1=x*a0+anf*a1;
    b1=x*b0+anf*b1;
    if (a1) {
      fac=1.0/a1;
      double g=b1*fac;
      if ( abs((g-gold)/g) < numeric_limits<double>::epsilon() ) {
	gamcf=exp(-x+a*log(x)-gln)*g;
	return;
      }
      gold=g;
    }
  }
  //// fixme
  FATAL_ERROR_TEXT("a too large, itmax too small.");
  gamcf=numeric_limits<double>::quiet_NaN(); // fixme
}


double incomplete_gamma_P(const double a, const double x) {
  if (x < 0.0 || a <= 0.0) {
    FATAL_ERROR_TEXT("Invalid arguments.");
  }
  if (x < (a+1.0)) {
    double gamser, gln;
    incomplete_gamma_series(a,x,gamser,gln);
    return gamser;
  } else {
    double gamcf, gln;
    incomplete_gamma_continued_fraction(a,x,gamcf,gln);
    return 1.0-gamcf;
  }
}


double incomplete_gamma_Q(const double a, const double x) {
  if (x < 0.0 || a <= 0.0) {
    FATAL_ERROR_TEXT("Invalid arguments.");
  }
  if (x < (a+1.0)) {
    double gamser, gln;
    incomplete_gamma_series(a,x,gamser,gln);
    return 1.0-gamser;
  } else {
    double gamcf, gln;
    incomplete_gamma_continued_fraction(a,x,gamcf,gln);
    return gamcf;
  }
}


double prob_chi_square(const double chi_sqr, const double n) {
  return incomplete_gamma_P(n/2.0,chi_sqr/2.0);
}


double Quantile_chi_square::operator()(const double p, const int n) const {
  pair<double,int> value_pair(p,n);
  map< pair<double,int>, double>::iterator i=cache.find(value_pair);
  if ( i!=cache.end() ) {
    // Wert ist schon im Cache.
    return i->second;
  }
  // Wert muss berechnet werden.
  //
  // Die Wahrscheinlichkeit prob_chi_square(chi_sqr,n) ist streng
  // monoton in chi_sqr, eine Intervallteilung genuegt also.
  double chi_sqr=1;
  double chi_sqr_min=0;
  double chi_sqr_max=0;
  if ( prob_chi_square(chi_sqr,n)<p ) {
    // Startwert chi^2=1 ist zu klein.
    chi_sqr_min=chi_sqr;
    do {
      chi_sqr *= 2;
    } while ( prob_chi_square(chi_sqr,n)<p );
    chi_sqr_max=chi_sqr;
  } else {
    // Startwert chi^2=1 ist zu gross.
    chi_sqr_max=chi_sqr;
    do {
      chi_sqr /= 2;
    } while ( prob_chi_square(chi_sqr,n)>p );
    chi_sqr_min=chi_sqr;
  }
  // Gesuchter Wert liegt in [chi_sqr_min,chi_sqr_max].
  double chi_sqr_mid=(chi_sqr_max+chi_sqr_min)/2;
  while ( abs(chi_sqr_max-chi_sqr_min) >
	  (chi_sqr_min+chi_sqr_max)*numeric_limits<double>::epsilon() ) {
    chi_sqr_mid = (chi_sqr_min+chi_sqr_max)/2;
    if ( prob_chi_square(chi_sqr_mid,n)<p ) {
      // chi_sqr_mid ist zu klein.
      chi_sqr_min=chi_sqr_mid;
    } else {
      // chi_sqr_mid ist zu gross.
      chi_sqr_max=chi_sqr_mid;
    }
  }
  chi_sqr=chi_sqr_mid;
  cache[value_pair]=chi_sqr;
#ifdef ASKOS_ERRORCHECK
  // Haben wir richtig gerechnet?
  if ( abs(chi_sqr_max-chi_sqr_min) > 
       10*(chi_sqr_min+chi_sqr_max)*numeric_limits<double>::epsilon() ) {
    FATAL_ERROR;
  }
  if ( ! (chi_sqr_min<=chi_sqr && chi_sqr<=chi_sqr_max) ) {
    FATAL_ERROR;
  }
  if ( prob_chi_square(chi_sqr_min,n)>p+10*numeric_limits<double>::epsilon() ||
       prob_chi_square(chi_sqr_max,n)<p-10*numeric_limits<double>::epsilon() ) {
    /*
    cerr << "min=" << setprecision(30) << prob_chi_square(chi_sqr_min,n) << endl;
    cerr << "max=" << setprecision(30) << prob_chi_square(chi_sqr_max,n) << endl;
    */
    FATAL_ERROR;
  }
#endif
  return chi_sqr;
}


#ifndef __GNUC__ // Borland C++ hat keine erf.
double erf(const double x) {
  return x<0 ? -incomplete_gamma_P(0.5,sqr(x)) : incomplete_gamma_P(0.5,sqr(x));
}
#endif
