#! /usr/bin/perl -w


use Getopt::Long;
use Cwd;
use File::Basename;


if ( !GetOptions("sql",\$sql,"table=s",\$table,"help",\$help,
		 "maxchar=i",\$max_char_table_name)
     || defined($help) ) {
  $myname=basename($0);
  print "usage: $myname [--sql] file1 file2 ...\n";
  print "MkDb version 1.0. Creates a database from parameter files.\n";
  print "\n";
  print "'file1', 'file2', ... are the names of the parameter files.\n";
  print "\n";
  print "--sql  creates SQL output generating the database.\n";
  print "--table=TABLE  The name of the default table in the database.\n";
  print "--maxchar=MAXCHAR  Maximal number of charactes for table names (default 31).\n";
  print "--help displays this help message.\n";
  print "\n";
  print "If the parameter files contain a directory part, the contents of the file 'state' are\n";
  print "inserted into the tables in column 'state' as well.\n";
  exit 1;
}

if ( defined($sql) ) {
  if ( !defined($table) ) {
    $table="default";
  }
}

if ( !defined($max_char_table_name) ) {
  $max_char_table_name=31;
}


if ( $#ARGV!=-1 ) {
  # files are given on command line
  @files = @ARGV;
} else {
  # files are given on stdin
  $i=0;
  while ( defined($line=<>) ) {
    chomp($line);
    $files[$i]=$line;
    ++$i;
  }
}


foreach $i (@files) {
  $dir=dirname($i);
  open(PARAMETERS,"<$i") || die "Could not open file '$i' for reading.\n";
  while ( defined($line=<PARAMETERS>) ) {
    $discard="";
    ($parameter,$discard,$value) = $line =~ /^\s*((\w|\-)+)\s+(.*)$/;
    if ( !defined($parameter) || !defined($value) ) {
      chomp($line);
      die "Syntax error in line '$line' in file '$i'.";
    }
    $parameter =~ s/\-/_/g;
    if ($parameter eq "script") {
      $value =~ s/\-/_/g;
      $value =~ s/\.[^.]*$//;
      while ( length($value)>$max_char_table_name ) {
	$value =~ s/(^|_)([^_]{3})([^_]+)(_|$)/$1$2$4/;
      }
      $tables{$value}=1;
    }
    $db{$dir}{$parameter}=$value;
  }
  close(PARAMETERS);
  $db{$dir}{"state"} = "NA";
  if ( open(STATE,"<$dir/state") ) {
    if ( defined($line = <STATE>) ) {
      chomp($line);
      $db{$dir}{"state"} = $line;
    }
  }
  if ( !defined($db{$dir}) ) {
    warn "No parameter in file '$i'.";
    next;
  }
  if ( !defined($sql) ) {
    $all_parameters{$parameter}=1;
  } else {
    $current_table=$table;
    if ( defined($db{$dir}{"script"}) ) {
      $current_table=$db{$dir}{"script"};
    }
    foreach $parameter ( keys %{$db{$dir}} ) {
      if ($parameter eq "script") {
	next;
      }
      if ( !defined($parameters{$current_table}{$parameter}) ) {
	$value=$db{$dir}{$parameter};
	if ( $value =~ /^\s*-?((\d+\.?\d*)|(\.\d+))([eE]-?\d+)?\s*$/ ) {
	  warn "Parameter '$parameter' with value '$value' identified as a number.\n";
	  $parameters{$current_table}{$parameter}="REAL";
	} elsif ( $value =~ /^\s*\d+-\d+-\d+\s+\d+:\d+:\d+\s*$/ ) {
	  warn "Parameter '$parameter' with value '$value' identified as a timestamp.\n";
	  $parameters{$current_table}{$parameter}="TIMESTAMP";
	} elsif ( $value =~ /^\s*\d+-\d+-\d+\s*$/ ) {
	  warn "Parameter '$parameter' with value '$value' identified as a date.\n";
	  $parameters{$current_table}{$parameter}="DATE";
	} else {
	  warn "Parameter '$parameter' with value '$value' identified as a string.\n";
	  $parameters{$current_table}{$parameter}="VARCHAR(255)";
	}
      }
    }
  }
}

if ( !defined($sql) ) {
  print "directory";
  foreach $parameter (keys %all_parameters) {
    print " ".$parameter;
  }
  print "\n";
  foreach $dir (keys %db) {
    print "$dir";
    foreach $parameter (keys %all_parameters) {
      if ( defined($db{$dir}{$parameter}) ) {
	print "\t".$db{$dir}{$parameter};
      } else {
	print "\t\\N";
      }
    }
    print "\n";
  }
} else {
  foreach $current_table ($table, keys %tables) {
    if ( defined($parameters{$current_table}) ) {
      print "DROP TABLE $current_table;\n";
      print "CREATE TABLE $current_table (directory VARCHAR(255)";
      foreach $parameter (keys %{$parameters{$current_table}} ) {
        $colname=$parameter;
        if ( $colname eq "end" ) {
	  # strange: a column name can not be 'end' with Postgresql.
	  $colname = $colname."x";
	}
	print ", $colname $parameters{$current_table}{$parameter}";
      }
      print  ");\n";
    }
  }
  foreach $dir (keys %db) {
    $current_table=$table;
    if ( defined($db{$dir}{"script"}) ) {
      $current_table=$db{$dir}{"script"};
    }
    print "INSERT INTO $current_table VALUES ('$dir'";
    foreach $parameter (keys %{$parameters{$current_table}}) {
      if ( defined($db{$dir}{$parameter}) ) {
	print ",'$db{$dir}{$parameter}'";
      } else {
	print ",NULL";
      }
    }
    print ");\n";
  }
}
