/***************************************************************************
                          kernel.h  -  description
                             -------------------
    begin                : Sat Jan 11 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KERNEL_H
#define KERNEL_H


#include "particle.h"


/** Smoothing kernel for SPH.
 *
 *  Base class for all kernels.
 *
 *  @author Moritz Franosch
 */
class Kernel {

protected:
  
  int d; // spacial dimension of this kernel
  double h; // smoothing length
  double c_d; // alpha_d/h^d, ensures that int W(r,r') dV = 1.

public: 

  /** Constructs a kernel for a dimension d and smoothing length h.
   *
   *  The parameter alpha_d is a scaling factor dependent on the
   *  dimension.
   */
  Kernel(const int d, const double h, const double alpha_d);

  /** Virtual destructor.
   */
  virtual ~Kernel();

  /** Returns the smoothing length (radius) of the kernel.
   */
  double get_h() const;

  /** Returns the dimension of the kernel.
   */
  int get_d() const;

  /** Returns W(r) = alpha_d/h^d * w(r/h).
   */
  double operator()(const double r) const {
    return c_d * w(r/h);
  }

  /** Returns dW(r)/dr.
   */
  double dW(const double r) const;

  /** Returns grad r1 W(|r1-r2|).
   */
  const Particle::vector_t grad_r(const Particle::vector_t& r1, 
				  const Particle::vector_t& r2) const;

  /** Returns grad r1 W(|r1-r2|).
   *
   *  Precondition: r12=|r1-r2|.
   */
  const Particle::vector_t grad_r(const Particle::vector_t& r1, 
				  const Particle::vector_t& r2,
				  const double& r12) const;

  /** Writes grad r1 W(|r1-r2|) into vector g and returns a reference
   *  to it.
   *
   *  Precondition: r12=|r1-r2|.
   */
  const Particle::vector_t& grad_r(const Particle::vector_t& r1, 
				   const Particle::vector_t& r2,
				   const double& r12,
				   Particle::vector_t& g) const;

  /** Returns the form function w(x).
   *
   *  The precondition int alpha_d w(|r|) dV = 1 must hold.
   */
  virtual double w(const double x) const=0;

  /** Returns the derivative dw(x)/dx of the form function w.
   */
  virtual double dw(const double x) const=0;

};


/** Returns x_i for 1<=i<=3, NaN else.
 */
double select(const int i, double x1, double x2, double x3);


#endif // KERNEL_H
