/***************************************************************************
                          particle.cpp  -  description
                             -------------------
    begin                : Fri Jan 3 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "blitz/tinyvec-et.h"

#include "askosmath/askosmath.h"
#include "askosio/string_convert.h"

#include "errorcheck.h"
#include "particle.h"


using namespace std;
using namespace blitz;


string Particle::string_vector(const Particle::vector_t& v) {
  string r="(";
  for (int i=0; i<v.length(); ++i) {
    if (i!=0) {
      r += ", ";
    }
    r += string_convert::to_string(v[i]);
  }
  r += ")";
  return r;
}


string Particle::list_vector(const Particle::vector_t& v) {
  string r="";
  for (int i=0; i<v.length(); ++i) {
    if (i!=0) {
      r += " ";
    }
    r += string_convert::to_string(v[i]);
  }
  return r;
}


Particle::Particle(const vector_t& p) :
  m(0),
  rho(0),
  drho(0),
  position(p),
  velocity(0),
  acceleration(0),
  last_rho(0),
  last_position(p),
  last_velocity(0) {
}


Particle::Particle() :
  m(numeric_limits<double>::quiet_NaN()),
  rho(numeric_limits<double>::quiet_NaN()),
  drho(numeric_limits<double>::quiet_NaN()),
  position(numeric_limits<double>::quiet_NaN()),
  velocity(numeric_limits<double>::quiet_NaN()),
  acceleration(numeric_limits<double>::quiet_NaN()),
  last_rho(numeric_limits<double>::quiet_NaN()),
  last_position(numeric_limits<double>::quiet_NaN()),
  last_velocity(numeric_limits<double>::quiet_NaN()) {
  //  I(false);
}


void Particle::set_position(const vector_t& p) {
  position=p;
}


void Particle::set_velocity(const vector_t& v) {
  velocity=v;
}


void Particle::set_acceleration(const vector_t& a) {
  acceleration=a;
}


void Particle::set_mass(const double m_ini) {
  m=m_ini;
}


void Particle::set_density(const double rho_ini) {
  rho=rho_ini;
  last_rho=rho;
}


void Particle::set_drho(const double drho_ini) {
  drho=drho_ini;
}


void Particle::update_Euler(const double dt) {
  rho += dt*drho;
  I(rho>=0);
  position += dt*velocity;
  velocity += dt*acceleration;
}


double Particle::update_modified_midpoint(const int step, const double dt) {
  REQUIRE(step>=0 && step<=2);
  double delta_t=0;
  switch (step) {
  case 0:
    last_rho = rho;
    rho += dt*drho;
    last_position = position;
    position += dt*velocity;
    last_velocity = velocity;
    velocity += dt*acceleration;
    delta_t=dt;
    break;
  case 1:
    last_rho += 2*dt*drho;
    swap(rho,last_rho);
    last_position += 2*dt*velocity;
    swap(position,last_position);
    last_velocity += 2*dt*acceleration;
    swap(velocity,last_velocity);
    delta_t=dt;
    break;
  case 2:
    rho = 0.5*(rho+last_rho+dt*drho);
    I(rho>=0);
    position = 0.5*(position+last_position+dt*velocity);
    velocity = 0.5*(velocity+last_velocity+dt*acceleration);
    break;
  default:
    I(false);
  }
  return delta_t;

  // leap-frog integration:
  // last_z = last_z + dt * dz(z)
  // swap(last_z,z)

  /*
  last_rho += 2*dt*drho;
  swap(last_rho,rho);

  last_position += 2*dt*velocity;
  swap(last_position,position);

  last_velocity += 2*dt*acceleration;
  swap(last_velocity,velocity);
  */
}


double euclidian_distance(const Particle::vector_t& r1, const Particle::vector_t& r2) {
  REQUIRE(r1.length()==r2.length());
  return sqrt( sqr_euclidian_distance(r1,r2) );
}


double euclidian_norm(const Particle::vector_t& r) {
  double s=0;
  for (int i=0; i<r.length(); ++i) {
    s += sqr(r[i]);
  }
  return sqrt(s);
}
