/***************************************************************************
                          particle.h  -  description
                             -------------------
    begin                : Fri Jan 3 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef PARTICLE_H
#define PARTICLE_H

#include "blitz/tinyvec.h"

#include "askosmath/askosmath.h"

#include "errorcheck.h"


/** A particle in the sense of SPH.
 *
 *  @author Moritz Franosch
 */
class Particle {

public:

  typedef blitz::TinyVector<double,3> vector_t;

private:

  double m; // mass
  //  double V; // volume
  double rho; // density
  double drho; // derivative drho/dt

  vector_t position;
  vector_t velocity;
  vector_t acceleration;

  double last_rho; // rho_(m-1) for leap-frog integration
  vector_t last_position; // position_(m-1) for leap-frog integration
  vector_t last_velocity; // velocity_(m-1) for leap-frog integration

public:

  /** Converts the vector v to a string (v1, ...).
   */
  static std::string string_vector(const Particle::vector_t& v);

  /** Converts the vector v to a space-separated list v1 v2 ....
   */
  static std::string list_vector(const Particle::vector_t& v);

  /** Constructs a particle at potition p.
   */
  Particle(const vector_t& p);

  /** Constructs an invalid particle.
   */
  Particle();

  const vector_t& get_position() const {
    return position;
  }

  const vector_t& get_velocity() const {
    return velocity;
  }

  const vector_t& get_acceleration() const {
    return acceleration;
  }

  double get_mass() const {
    return m;
  }

  double get_density() const {
    return rho;
  }

  void set_position(const vector_t& p);

  void set_velocity(const vector_t& v);

  void set_acceleration(const vector_t& a);

  void set_mass(const double m);

  void set_density(const double rho);

  void set_drho(const double drho);

  /** Updates the position and the velocity of the particle accoding
   *  to its acceleration in one Euler integration step (dt).
   */
  void update_Euler(const double dt);

  /** Updates the position and the velocity of the particle according
   *  to its acceleration by the modified midpoint method, step m.
   *
   *  The advance in time is returned.
   */
  double update_modified_midpoint(const int m, const double dt);

};


/** Returns |r1-r2|^2.
 */
template<typename number, int N>
double sqr_euclidian_distance(const blitz::TinyVector<number,N>& r1, 
			      const blitz::TinyVector<number,N>& r2) {
  REQUIRE( r1.length()==N );
  REQUIRE( r2.length()==N );
  double s=0;
  for (int i=0; i<N; ++i) {
    s += sqr(r1[i]-r2[i]);
  }
  return s;
}


template<>
inline double sqr_euclidian_distance<double,3>(const blitz::TinyVector<double,3>& r1, 
					       const blitz::TinyVector<double,3>& r2) {
  REQUIRE( r1.length()==3 );
  REQUIRE( r2.length()==3 );
  return sqr(r1[0]-r2[0])+sqr(r1[1]-r2[1])+sqr(r1[2]-r2[2]);
}


template<>
inline double sqr_euclidian_distance<float,3>(const blitz::TinyVector<float,3>& r1, 
					      const blitz::TinyVector<float,3>& r2) {
  REQUIRE( r1.length()==3 );
  REQUIRE( r2.length()==3 );
  return sqr(r1[0]-r2[0])+sqr(r1[1]-r2[1])+sqr(r1[2]-r2[2]);
}


/** Returns |r1-r2|.
 */
double euclidian_distance(const Particle::vector_t& r1, const Particle::vector_t& r2);


/** Returns the euclidian norm |r| = sqrt( sum_i r_i^2 ).
 */
double euclidian_norm(const Particle::vector_t& r);


#endif // PARTICLE_H
