/***************************************************************************
                          rax_parser.cpp  -  description
                             -------------------
    begin                : Sat Feb 1 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>

#include "xercesc/parsers/SAXParser.hpp"
#include "xercesc/util/XMLString.hpp"
#include "xercesc/util/PlatformUtils.hpp"

#include "askosio/log.h"
#include "askosio/string_convert.h"

#include "errorcheck.h"
#include "rax_parser.h"


using namespace std;


Pointer<RAX_Handler> RAX_Handler_Pointer::default_handler(new RAX_Handler);


RAX_Handler_Pointer::RAX_Handler_Pointer()
  /*handler_ptr(&default_handler)*/ {
  *this = default_handler;
}


RAX_Handler_Pointer::RAX_Handler_Pointer(const Pointer<RAX_Handler> handler) :
  Pointer<RAX_Handler>(handler) {
}

/*
RAX_Handler* RAX_Handler_Pointer::operator->() {
  return handler_ptr;
}
*/

const string RAX_Parser::xmlch_to_string(const XMLCh* const name) {
  ENSURE(name!=0);
  char* cstr = XMLString::transcode(name);
  string s(cstr);
  delete[] cstr; // not very efficient
  return s;
}


void RAX_Parser::startElement(const XMLCh* const name,
			      AttributeList& attributes) {
  ENSURE( !handlers.empty() );
  LOG(log);

  if ( !element_stack.empty() ) {
    // handle collected text
    text( element_stack.top() );
  }

  string element=xmlch_to_string(name);

  log.debug() << "<" << element << ">" << endl;

  // push new element to stack of elements
  element_stack.push(element);

  // Call RAX_Handler::start() for this element.
  //
  // If handler 'RAX_Handler handlers.top()[element]' has not existed
  // before, the default virtual method RAX_Handler::start() just
  // calls start_element().
  handlers.top()[element]->start(*this,element,attributes);
}


void RAX_Parser::endElement(const XMLCh* const name) {
  ENSURE( !handlers.empty() );
  ENSURE( !element_stack.empty() );
  LOG(log);

  string element=xmlch_to_string(name);

  log.debug() << "</" << element << ">" << endl;

  // handle collected text
  text(element);

  // delete last element from stack of elements
  I( element_stack.top()==element );
  element_stack.pop();

  // Call RAX_Handler::end() for this element.
  //
  // If handler 'RAX_Handler handlers.top()[element]' has not existed
  // before, the default virtual method RAX_Handler::end() just
  // calls end_element().
  handlers.top()[element]->end(*this,element);
}


void RAX_Parser::characters(const XMLCh* const chars, const unsigned int) {
  // add characters to collected_text
  collected_text += xmlch_to_string(chars);
}


void RAX_Parser::text(const std::string& element) {
  // Ignore leading and trailing whitespace.
  collected_text=string_convert::without_leading_trailing_whitespace(collected_text);
  // Call RAX_Handler::text() for this element if text has been collected.
  if ( collected_text!="" ) {
    // If handler 'RAX_Handler handlers.top()[element]' has not existed
    // before, the default virtual method RAX_Handler::text() just
    // calls text().
    handlers.top()[element]->text(*this,element,collected_text);
    // delete collected text for next chunk
    collected_text="";
  }
}


void RAX_Parser::warning(const SAXParseException& exception) {
  LOG(log);
  log.warn() << xmlch_to_string(exception.getMessage())
	     << " at line: " << exception.getLineNumber()
	     << endl;
}


void RAX_Parser::error(const SAXParseException& exception) {
  LOG(log);
  log.error() << xmlch_to_string(exception.getMessage())
	      << " at line: " << exception.getLineNumber()
	      << endl;
}


void RAX_Parser::fatalError(const SAXParseException& exception) {
  LOG(log);
  log.error() << "Fatal XML Error: " << xmlch_to_string(exception.getMessage())
	      << " at line: " << exception.getLineNumber()
	      << endl;
  throw exception; // _must_ be thrown in case of a fatal error
}


RAX_Parser::RAX_Parser() {
  // initialize the stack with an empty set of handlers
  handlers.push( map<string,RAX_Handler_Pointer>() );
}


void RAX_Parser::start_element(const string& name, AttributeList&) {
  LOG(log);
  log.debug() << "<" << name << ">" << endl;
}


void RAX_Parser::end_element(const string& name) {
  LOG(log);
  log.debug() << "</" << name << ">" << endl;
}


void RAX_Parser::text(const std::string& name, const string& t) {
  LOG(log);
  log.debug() << name << "->" << "'" << t << "'" << endl;
}


/*
class Error_Handler : public HandlerBase {

public:

  virtual void error(const SAXParseException& exception) {
    cerr << "Error_Handler::error" << endl;
  }

};
*/


bool RAX_Parser::parse(const string& filename) {
  LOG(log);

  try {
    XMLPlatformUtils::Initialize();
  } catch (const XMLException& e) {
    log.error() << "Error during initialization: "
		<< xmlch_to_string(e.getMessage()) << endl;
    return false;
  }

  SAXParser parser;
  //  parser.setDoValidation(true);
  //  parser.setValidationScheme(SAXParser::Val_Always);
  /*
  parser.setValidationConstraintFatal(false);
  parser.setExitOnFirstFatalError(false);
  */
  //  parser.setDoNamespaces(true);

  parser.setDocumentHandler(this);
  parser.setErrorHandler(this);

  try {
    parser.parse( filename.c_str() );
  } catch (const XMLException& e) {
    log.error() << "Exception message is: "
		<< xmlch_to_string(e.getMessage()) << endl;
    return false;
  } catch (const SAXParseException& e) {
    log.error() << "Exception message is: "
		<< xmlch_to_string(e.getMessage()) << endl;
    return false;
  }
  catch (...) {
    log.error() << "Unexpected Exception." << endl;
    return false;
  }

  return true;
}


void RAX_Parser::push_handlers() {
  // push empty set of handlers onto stack
  handlers.push( map<string,RAX_Handler_Pointer>() );
}


void RAX_Parser::pop_handlers() {
  handlers.pop();
}
  

void RAX_Parser::set_handler(const std::string& element_name, 
			     Pointer<RAX_Handler> handler) {
  ENSURE( !handlers.empty() );
  handlers.top()[element_name]=handler;
}


RAX_Parser::~RAX_Parser() {
}
