/***************************************************************************
                          rax_parser.h  -  description
                             -------------------
    begin                : Sat Feb 1 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef RAX_PARSER_H
#define RAX_PARSER_H

#include <stack>
#include <map>

#include "xercesc/sax/HandlerBase.hpp"

#include "askosdata/pointer.h"

#include "rax_handler.h"


/** "Pointer" to a RAX_Handler.
 *
 *  @author Moritz Franosch
 */
class RAX_Handler_Pointer : public Pointer<RAX_Handler> {

private:

  static Pointer<RAX_Handler> default_handler;
  // default handler that just calls RAX_Parser::start_element(),
  // RAX_Parser::end_element(), RAX_Parser::text(), respectively.

public:

  /** Constructs the RAX_Handler_Pointer with the default handler that
   *  just calls RAX_Parser::start_element(),
   *  RAX_Parser::end_element(), RAX_Parser::text(), respectively.
   */
  RAX_Handler_Pointer();

  /** Constructs the RAX_Handler_Pointer with this handler.
   */
  RAX_Handler_Pointer(const Pointer<RAX_Handler> handler);
  
  /** Dereferencation.
   */
  //  RAX_Handler* operator->();

};


/** Recursive API for XML.
 *
 *  @author Moritz Franosch
 */
class RAX_Parser : private HandlerBase {

private:

  // stack of maps between element names and their handlers
  std::stack< std::map<std::string,RAX_Handler_Pointer> > handlers;

  std::string collected_text; // text characters collected between element tags

  // stack of elements is pushed and poped according to the nested
  // structure of XML-elements
  std::stack<std::string> element_stack;

  /** Called by SAXParser when a tag starts.
   */
  virtual void startElement(const XMLCh* const name, AttributeList& attributes);

  /** Called by SAXParser when a tag ends.
   */
  virtual void endElement(const XMLCh* const name);

  /** Called by SAXParser when character data has been read.
   */
  virtual void characters(const XMLCh* const chars, const unsigned int length);

  /** Called by SAXParser in case of a warning.
   *
   *  Prints the warning to stderr.
   */
  virtual void warning(const SAXParseException& exception);

  /** Called by SAXParser in case of an error.
   *
   *  Prints the error to stderr.
   */
  virtual void error(const SAXParseException& exception);

  /** Called by SAXParser in case of a fatal error.
   *
   *  Prints the fatal error to stderr.
   */
  virtual void fatalError(const SAXParseException& exception);

  /** Calls RAX_Handler::text() for the current element.
   */
  void text(const std::string& name);

public: 

  /** Converts a XMLCh* to a std::string.
   */
  static const std::string xmlch_to_string(const XMLCh* const name);

  /** Constructs a RAX Parser.
   */
  RAX_Parser();

  /** Called whenever a start-tag is encountered.
   *
   *  By default, do nothing. Application writers may override this
   *  method in a subclass to take specific actions at the start of
   *  each element (such as allocating a new tree node or writing
   *  output to a file).
   *
   *  @parameter name The name of the end-tag.
   *
   *  @parameter attributes The specified or defaulted attributes.
   */
  virtual void start_element(const std::string& name, AttributeList& attributes);

  /** Called whenever an end-tag is encountered.
   *
   *  By default, do nothing. Application writers may override this
   *  method in a subclass to take specific actions at the end of each
   *  element (such as finalising a tree node or writing output to a
   *  file).
   *
   *  @parameter name The name of the end-tag.
   */
  virtual void end_element(const std::string& name);

  /** Called whenever a start- or an end-tag is encountered.
   *
   *  By default, do nothing. Application writers may override this
   *  method in a subclass to take specific actions for each chunk of
   *  character data.
   *
   *  @parameter name The name of the XML-tag in which context the
   *  character data was collected.
   *
   *  @parameter t The character data.
   */
  virtual void text(const std::string& name, const std::string& t);

  /** Parses XML-File filename.
   *
   *  Returns true if successful.
   */
  bool parse(const std::string& filename);

  /** Pushes the current handlers on the stack.
   */
  void push_handlers();

  /** Pops the last pushed handles from the stack.
   */
  void pop_handlers();
  
  /** Sets *handler as a handler for element with name element_name.
   */
  void set_handler(const std::string& element_name, Pointer<RAX_Handler> handler);

  /** Virtual destructor.
   */
  virtual ~RAX_Parser();

};


#endif // RAX_PARSER_H
