/***************************************************************************
                          rax_parser_sph.cpp  -  description
                             -------------------
    begin                : Sat Feb 1 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "xercesc/sax/AttributeList.hpp"

#include "blitz/tinyvec-et.h"

#include "askosio/log.h"

#include "errorcheck.h"
#include "rax_parser_sph.h"
#include "sphere_space.h"
#include "sphere.h"
#include "box.h"
#include "sph.h"
#include "voxelizer.h"
#include "triangulator.h"


using namespace std;


RAX_Parser_SPH::RAX_Parser_SPH() :
  RAX_Parser(),
  dt_display( numeric_limits<double>::quiet_NaN() ),
  nspheres(10),
  first_simulate(true),
  initial_geometry(),
  nparticles_row(100),
  d_max(0) {
  // Set the handler for the (outermost) <input> tag.
  // This handler sets all other handlers recursively.
  Pointer<RAX_Handler> input_handler(new Input_Handler);
  set_handler("input",input_handler);
}


bool RAX_Parser_SPH::parse(const std::string& filename) {
  return RAX_Parser::parse(filename);
}


void RAX_Parser_SPH::construct_sph(const int dimensions,
				   const double xdim, const double ydim, 
				   const double zdim,
				   const double h, const double rho) {
  double velocity_of_sound=10;
  double dt=0.001;
  sph_ptr = Pointer<SPH>(new SPH(dimensions,xdim,ydim,zdim,h,
				       velocity_of_sound,dt));
  sph_ptr->set_density(rho);
}


void RAX_Parser_SPH::set_gravity(const Particle::vector_t& g) {
  REQUIRE(sph_ptr!=0);
  sph_ptr->set_gravity(g);
}


void RAX_Parser_SPH::set_nu(const double nu) {
  REQUIRE(sph_ptr!=0);
  sph_ptr->set_viscosity(nu);
}


void RAX_Parser_SPH::set_c(const double c) {
  REQUIRE(sph_ptr!=0);
  sph_ptr->set_velocity_of_sound(c);
}


void RAX_Parser_SPH::set_dt(const double dt) {
  REQUIRE(sph_ptr!=0);
  sph_ptr->set_dt(dt);
}


bool RAX_Parser_SPH::set_kernel(const string& name) {
  REQUIRE(sph_ptr!=0);
  return sph_ptr->set_kernel(name);
}


void RAX_Parser_SPH::set_dt_display(const double dt_display_ini) {
  dt_display = dt_display_ini;
}


void RAX_Parser_SPH::set_nspheres(const int nspheres_ini) {
  nspheres=nspheres_ini;
}


void RAX_Parser_SPH::set_nparticles_row(const int nparticles_row_ini) {
  nparticles_row=nparticles_row_ini;
}


void RAX_Parser_SPH::set_dmax(const double d_max_ini) {
  d_max=d_max_ini;
}


void RAX_Parser_SPH::simulate(const double duration) {
  REQUIRE(sph_ptr!=0);
  LOG(log);
  if ( first_simulate ) {

    if ( isnan(dt_display) ) {
      // dt_display not set
      dt_display = sph_ptr->get_dt(); // set display timestep to simulation timestep
    }

    /*
    Particle::vector_t middle(0);
    middle[0]=0.5;
    //  middle[1]=1.1*radius;
    middle[1]=0.5;
    middle[2]=0;
    double radius=0.4;

    int nparticles=3600;
    double nparticles_row=round(sqrt(double(nparticles)));
    double d=sph_ptr->get_space().get_length_x()/nparticles_row;

    // insert particles into a sphere
    for (int nx=0; nx<nparticles_row; ++nx) {
      for (int ny=0; ny<nparticles_row; ++ny) {
	Particle::vector_t p;
	p[0]=d/2+nx*d;
	p[1]=d/2+ny*d;
	p[2]=0;
	if ( euclidian_norm(p-middle)<radius ) {
	  sph_ptr->insert_particle(p);
	}
      }
    }
    
    // insert particles into a wave
    for (int nx=0; nx<nparticles_row; ++nx) {
      for (int ny=0; ny<nparticles_row; ++ny) {
	Particle::vector_t p;
	p[0]=d/2+nx*d;
	p[1]=d/2+ny*d;
	p[2]=0;
	if ( p[1] < 0.5+0.125*cos(2*pi*p[0]) ) {
	  sph_ptr->insert_particle(p);
	}
      }
    }
    */

    // insert particles
    I(nparticles_row>=0); // fixme
    double d=sph_ptr->get_space().get_length_x()/(nparticles_row+1);
    double offset=d/2;
    double boundary_width = sph_ptr->get_boundary_width();
    if ( !isnan(boundary_width) ) {
      // boundary_width has been set
      // respectfully place all particles out of boundary
      d = (sph_ptr->get_space().get_length_x()-2*boundary_width)/nparticles_row;
      offset = boundary_width/2;
    }
    switch ( sph_ptr->get_dimensions() ) {
    case 2:
      for (int nx=0; nx<=nparticles_row; ++nx) {
	for (int ny=0; ny<=nparticles_row; ++ny) {
	  Particle::vector_t p;
	  p[0]=offset+nx*d;
	  p[1]=offset+ny*d;
	  p[2]=0;
	  if ( initial_geometry(p) ) {
	    // position is contained in initial geometry of particles
	    sph_ptr->insert_particle(p);
	  }
	}
      }
      break;
    case 3:
      for (int nx=0; nx<=nparticles_row; ++nx) {
	for (int ny=0; ny<=nparticles_row; ++ny) {
	  for (int nz=0; nz<=nparticles_row; ++nz) {
	    Particle::vector_t p;
	    p[0]=offset+nx*d;
	    p[1]=offset+ny*d;
	    p[2]=offset+nz*d;
	    if ( initial_geometry(p) ) {
	      // position is contained in initial geometry of particles
	      sph_ptr->insert_particle(p);
	    }
	  }
	}
      }
      break;
    default:
      log.error() << sph_ptr->get_dimensions() << " dimensions not supported." << endl;
      return;
    }

    sph_ptr->start();
    first_simulate=false;
  }
  // generate XML output
  cout << "<simulation>" << endl;
  double next_display_time = sph_ptr->get_time();
  double stop_time = sph_ptr->get_time()+duration;
  while ( sph_ptr->get_time() < stop_time ) {
    if ( sph_ptr->get_time() >= next_display_time ) {
      next_display_time += dt_display;
      cout << *sph_ptr; // put out sph as XML

      /*
      // put out spheres as XML
      Sphere_Space sphere_space(get_sph(),nspheres);
      cout << sphere_space;
      */

      /*
      // put out voxels as XML
      Voxelizer voxelizer(get_sph(),get_sph().get_h()/10); // fixme
      cout << voxelizer;
      */

      // put out triangluated surface as XML
      Triangulator triangulator(get_sph(),d_max);
      cout << triangulator;

    }
    sph_ptr->update();
  }
  cout << "</simulation>" << flush << endl;
}


SPH& RAX_Parser_SPH::get_sph() {
  REQUIRE(sph_ptr!=0);
  return *sph_ptr;
}


void RAX_Parser_SPH::add_sphere(const Particle::vector_t& p, const double r) {
  Pointer<Geometry> sphere(new Sphere(p,r));
  initial_geometry.add(sphere);
}


void RAX_Parser_SPH::add_box(const Particle::vector_t& p1, const Particle::vector_t& p2) {
  Pointer<Geometry> box(new Box(p1,p2));
  initial_geometry.add(box);
}


void RAX_Parser_SPH::start_element(const std::string& name, AttributeList&) {
  LOG(log);
  log.error() << "Tag <" << name << "> not valid in this context." << endl;
  I(false);
}


void RAX_Parser_SPH::end_element(const std::string& name) {
  LOG(log);
  log.error() << "Tag </" << name << "> not valid in this context." << endl;
  I(false);
}


void RAX_Parser_SPH::text(const std::string& name, const std::string&) {
  LOG(log);
  log.error() << "Text in tag <" << name << "> not valid in this context." << endl;
  I(false);
}


Particle::vector_t RAX_Handler_SPH::vector_value(const std::string& s) const {
  Particle::vector_t r(0);
  istringstream input(s);
  double d=0;
  int i=0;
  while ( input >> d ) {
    if ( i>=r.length() ) {
      I(false); // fixme
    }
    r[i]=d;
    ++i;
  }
  if ( i!=r.length() ) {
    I(false);
  }
  if ( !input.eof() ) {
    I(false);
  }
  return r;
}


void Input_Handler::start(RAX_Parser& parser, 
			  const std::string&, AttributeList&) {
  Pointer<RAX_Handler> parameters_handler(new Parameters_Handler);
  parser.set_handler("parameters",parameters_handler);

  Pointer<RAX_Handler> simulate_handler(new Simulate_Handler);
  parser.set_handler("simulate",simulate_handler);

  Pointer<RAX_Handler> initial_conditions_handler(new Initial_Conditions_Handler);
  parser.set_handler("initial_conditions",initial_conditions_handler);

  cout << "<output>" << endl;
}


void Input_Handler::end(RAX_Parser&, const std::string&) {
  cout << "</output>" << endl;
}


void Input_Handler::text(RAX_Parser&,
			 const std::string& name, const std::string&) {
  LOG(log);
  //fixme  log.error() << "Text in tag <" << name << "> not valid in this context." << endl;
  //  I(false);
}


Parameters_Handler::Parameters_Handler() :
  xdim(1), 
  ydim(1), 
  zdim(1), 
  h(0.1),
  rho(1000),
  dimensions(3),
  d_max(0) {
}


void Parameters_Handler::start(RAX_Parser& parser, 
			       const std::string& name, AttributeList& attributes) {
  REQUIRE(name=="parameters");
  LOG(log);
  RAX_Parser_SPH* parser_sph=dynamic_cast<RAX_Parser_SPH*>(&parser);
  I(parser_sph!=0);
  // For each attribute ...
  for (int i=0; i<attributes.getLength(); ++i) {
    string attname=RAX_Parser::xmlch_to_string(attributes.getName(i));
    string value=RAX_Parser::xmlch_to_string(attributes.getValue(i));
    // ... store the parameter.
    if (attname=="xdim") {
      xdim=double_value(value);
    } else if (attname=="ydim") {
      ydim=double_value(value);
    } else if (attname=="zdim") {
      zdim=double_value(value);
    } else if (attname=="h") {
      h=double_value(value);
    } else if (attname=="rho") {
      rho=double_value(value);
    } else if (attname=="dimensions") {
      dimensions=int_value(value);
    } else if (attname=="nparticles_row") {
      parser_sph->set_nparticles_row(int_value(value));
    } else if (attname=="dmax") {
      d_max=double_value(value);
    } else {
      log.error() << "Unknown attribute of <" << name << "> '" << attname << "'." 
		  << endl;
    }
  }
}


void Parameters_Handler::end(RAX_Parser& parser, const std::string& name) {
  REQUIRE(name=="parameters");
  RAX_Parser_SPH* parser_sph=dynamic_cast<RAX_Parser_SPH*>(&parser);
  I(parser_sph!=0);
  if (d_max<=0) {
    d_max=h;
  }
  parser_sph->set_dmax(d_max);
  parser_sph->construct_sph(dimensions,xdim,ydim,zdim,h,rho);
}


void Parameters_Handler::text(RAX_Parser&, 
			      const std::string& name, const std::string&) {
  LOG(log);
  log.error() << "Text in tag <" << name << "> not valid in this context." << endl;
  I(false);
}


void Initial_Conditions_Handler::start(RAX_Parser& parser, 
				       const std::string&, AttributeList&) {
  parser.push_handlers();
  Pointer<RAX_Handler> sphere_handler(new Sphere_Handler);
  parser.set_handler("sphere",sphere_handler);
  Pointer<RAX_Handler> box_handler(new Box_Handler);
  parser.set_handler("box",box_handler);

  // must include myself to handle </initial_conditions>
  Pointer<RAX_Handler> initial_conditions_handler(new Initial_Conditions_Handler);
  parser.set_handler("initial_conditions",initial_conditions_handler);
}


void Initial_Conditions_Handler::end(RAX_Parser& parser, const std::string&) {
  parser.pop_handlers();
}


void Sphere_Handler::start(RAX_Parser& parser, 
			   const std::string& name, AttributeList& attributes) {
  REQUIRE(name=="sphere");
  LOG(log);
  RAX_Parser_SPH* parser_sph=dynamic_cast<RAX_Parser_SPH*>(&parser);
  I(parser_sph!=0);
  double radius = numeric_limits<double>::quiet_NaN();
  Particle::vector_t position(numeric_limits<double>::quiet_NaN());
  // For each attribute ...
  for (int i=0; i<attributes.getLength(); ++i) {
    string attname=RAX_Parser::xmlch_to_string(attributes.getName(i));
    string value=RAX_Parser::xmlch_to_string(attributes.getValue(i));
    // ... store the parameter.
    if (attname=="radius") {
      radius=double_value(value);
    } else if (attname=="position") {
      position=vector_value(value);
    } else {
      log.error() << "Unknown attribute of <" << name << "> '" << attname << "'." 
		  << endl;
    }
  }
  parser_sph->add_sphere(position,radius);
}


void Sphere_Handler::end(RAX_Parser&, const std::string&) {
}


void Box_Handler::start(RAX_Parser& parser, 
			const std::string& name, AttributeList& attributes) {
  REQUIRE(name=="box");
  LOG(log);
  RAX_Parser_SPH* parser_sph=dynamic_cast<RAX_Parser_SPH*>(&parser);
  I(parser_sph!=0);
  Particle::vector_t p1(numeric_limits<double>::quiet_NaN());
  Particle::vector_t p2(numeric_limits<double>::quiet_NaN());
  // For each attribute ...
  for (int i=0; i<attributes.getLength(); ++i) {
    string attname=RAX_Parser::xmlch_to_string(attributes.getName(i));
    string value=RAX_Parser::xmlch_to_string(attributes.getValue(i));
    // ... store the parameter.
    if (attname=="edge1") {
      p1=vector_value(value);
    } else if (attname=="edge2") {
      p2=vector_value(value);
    } else {
      log.error() << "Unknown attribute of <" << name << "> '" << attname << "'." 
		  << endl;
    }
  }
  parser_sph->add_box(p1,p2);
}


void Box_Handler::end(RAX_Parser&, const std::string&) {
}


Simulate_Handler::Simulate_Handler() :
  duration(0) {
}


void Simulate_Handler::start(RAX_Parser& parser, 
			     const std::string& name, AttributeList& attributes) {
  REQUIRE(name=="simulate");
  LOG(log);
  RAX_Parser_SPH* parser_sph=dynamic_cast<RAX_Parser_SPH*>(&parser);
  I(parser_sph!=0);
  // For each attribute ...
  for (int i=0; i<attributes.getLength(); ++i) {
    string attname=RAX_Parser::xmlch_to_string(attributes.getName(i));
    string value=RAX_Parser::xmlch_to_string(attributes.getValue(i));
    // ... set the appropriate parameter.
    if (attname=="duration") {
      duration=double_value(value);
    } else if (attname=="nspheres") {
      parser_sph->set_nspheres(int_value(value));
    } else if (attname=="gravity") {
      parser_sph->set_gravity(vector_value(value));
    } else if (attname=="nu") {
      parser_sph->set_nu(double_value(value));
    } else if (attname=="c") {
      parser_sph->set_c(double_value(value));
    } else if (attname=="dt") {
      parser_sph->set_dt(double_value(value));
    } else if (attname=="dt_display") {
      parser_sph->set_dt_display(double_value(value));
    } else if (attname=="boundary_width") {
      parser_sph->get_sph().set_boundary_width(double_value(value));
    } else if (attname=="boundary_repulsive_force") {
      parser_sph->get_sph().set_boundary_repulsive_force(double_value(value));
    } else if (attname=="kernel") {
      if ( !parser_sph->set_kernel(value) ) {
	// name of kernel is not known
	log.error() << "Unknown kernel '" << value << "'." << endl;
      }
    } else {
      log.error() << "Unknown attribute of <" << name << "> '" << attname << "'." 
		  << endl;
    }
  }
}


void Simulate_Handler::end(RAX_Parser& parser, const std::string& name) {
  REQUIRE(name=="simulate");
  RAX_Parser_SPH* parser_sph=dynamic_cast<RAX_Parser_SPH*>(&parser);
  I(parser_sph!=0);
  parser_sph->simulate(duration);
}


void Simulate_Handler::text(RAX_Parser&, 
			    const std::string& name, const std::string&) {
  LOG(log);
  log.error() << "Text in tag <" << name << "> not valid in this context." << endl;
  I(false);
}
