/***************************************************************************
                          rax_parser_sph.h  -  description
                             -------------------
    begin                : Sat Feb 1 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef RAX_PARSER_SPH_H
#define RAX_PARSER_SPH_H

#include "askosdata/pointer.h"

#include "rax_parser.h"
#include "particle.h"
#include "compound.h"


class SPH;


/** RAX Handler for SPH.
 */
class RAX_Handler_SPH : public RAX_Handler {

public:

  /** Converts the string s into a vector.
   *
   *  fixme: error handling
   */
  virtual Particle::vector_t vector_value(const std::string& s) const;

};


/** Class that handles <input>.
 */
class Input_Handler : public RAX_Handler_SPH {

public:

  /** Called when <input> starts.
   *
   *  Installs the handlers for input.
   *
   *  Prints <output>.
   */
  virtual void start(RAX_Parser& parser, 
		     const std::string& name, AttributeList& attributes);

  /** Called when <input> ends.
   *
   *  Prints </output>.
   */
  virtual void end(RAX_Parser& parser, const std::string& name);

  /** Called when character data is available.
   *  (should never be called).
   */
  virtual void text(RAX_Parser& parser, 
		    const std::string& name, const std::string& t);

};


/** Class that handles <parameters>.
 */
class Parameters_Handler : public RAX_Handler_SPH {

private:

  double xdim, ydim, zdim, h, rho; // stored parameters
  int dimensions; // stored parameter
  double d_max; // stored parameter

public:

  /** Constructor.
   */
  Parameters_Handler();

  /** Called when <parameters> starts.
   *
   *  Stores the parameters.
   */
  virtual void start(RAX_Parser& parser, 
		     const std::string& name, AttributeList& attributes);

  /** Called when <parameters> ends.
   *
   *  Constructs the SPH_Space of class RAX_Parser_SPH with the stored
   *  parameters.
   */
  virtual void end(RAX_Parser& parser, const std::string& name);

  /** Called when character data is available.
   *  (should never be called).
   */
  virtual void text(RAX_Parser& parser, 
		    const std::string& name, const std::string& t);

};


/** Class that handles <initial_conditions>.
 */
class Initial_Conditions_Handler : public RAX_Handler_SPH {

public:

  /** Called when <initial_conditions> starts.
   *
   *  Installs handlers for initial conditions.
   */
  virtual void start(RAX_Parser& parser, 
		     const std::string& name, AttributeList& attributes);

  /** Called when <initial_conditions> ends.
   *
   *  Restores handlers.
   */
  virtual void end(RAX_Parser& parser, const std::string& name);

};


/** Class that handles <sphere> in <initial_conditions>.
 */
class Sphere_Handler : public RAX_Handler_SPH {

public:

  /** Called when <sphere> starts.
   */
  virtual void start(RAX_Parser& parser, 
		     const std::string& name, AttributeList& attributes);

  /** Called when <sphere> ends.
   *
   *  Does nothing.
   */
  virtual void end(RAX_Parser& parser, const std::string& name);

};


/** Class that handles <box> in <initial_conditions>.
 */
class Box_Handler : public RAX_Handler_SPH {

public:

  /** Called when <box> starts.
   */
  virtual void start(RAX_Parser& parser, 
		     const std::string& name, AttributeList& attributes);

  /** Called when <box> ends.
   *
   *  Does nothing.
   */
  virtual void end(RAX_Parser& parser, const std::string& name);

};


/** Class that handles <simulate>.
 */
class Simulate_Handler : public RAX_Handler_SPH {

private:

  double duration; // parameter duration

public:

  /** Constructor.
   */
  Simulate_Handler();

  /** Called when <simulate> starts.
   */
  virtual void start(RAX_Parser& parser, 
		     const std::string& name, AttributeList& attributes);

  /** Called when <simulate> ends.
   */
  virtual void end(RAX_Parser& parser, const std::string& name);

  /** Called when character data is available.
   *  (should never be called).
   */
  virtual void text(RAX_Parser& parser, 
		    const std::string& name, const std::string& t);

};


/** Parser for input files.
 *
 *  @author Moritz Franosch
 */
class RAX_Parser_SPH : public RAX_Parser {

private:

  Pointer<SPH> sph_ptr; // simulation is calculated here

  double dt_display; // time step for XML output

  int nspheres; // number of spheres that approximate the fluid

  bool first_simulate; // simulate() has not yet been called

  Compound initial_geometry; // initial conditions

  int nparticles_row; // number of particles in a row
  double d_max; // largest triangle used for triangulation

public:

  /** Constructor.
   */
  RAX_Parser_SPH();

  /** Parses the XML-file filename.
   *
   *  Executes the SPH-simulation.
   *
   *  Returns true if successful.
   */
  bool parse(const std::string& filename);

  /** Constructs an sph-space with the given parameters.
   */
  void construct_sph(const int dimensions,
		     const double xdim, const double ydim, const double zdim,
		     const double h, const double rho);

  /** Sets gravity.
   *
   *  @precondition construct_sph() was called.
   */
  void set_gravity(const Particle::vector_t& g);

  /** Sets viscosity.
   *
   *  @precondition construct_sph() was called.
   */
  void set_nu(const double nu);

  /** Sets velocity of sound.
   *
   *  @precondition construct_sph() was called.
   */
  void set_c(const double c);

  /** Sets timestep.
   *
   *  @precondition construct_sph() was called.
   */
  void set_dt(const double dt);

  /** Sets the kernel.
   *
   *  Returns true if successful.
   *
   *  @precondition construct_sph() was called.
   */
  bool set_kernel(const std::string& name);

  /** Sets display timestep.
   */
  void set_dt_display(const double dt_display);

  /** Sets the number of spheres to approximate the fluid.
   */
  void set_nspheres(const int nspheres);

  /** Sets the number of particles in a row.
   */
  void set_nparticles_row(const int nparticles_row);

  /** Sets the maximum length of the triangles when triangulating the
   *  surface.
   */
  void set_dmax(const double d_max_ini);

  /** Runs simulation for this duration.
   *
   *  @precondition construct_sph() was called.
   */
  void simulate(const double duration);

  /** Returns the class SPH that is used for the simulation.
   *
   *  @precondition construct_sph() was called.
   */
  SPH& get_sph();

  /** Adds a sphere with radius r at position p.
   */
  void add_sphere(const Particle::vector_t& p, const double r);

  /** Adds a box with one corner at p and the other at p2.
   */
  void add_box(const Particle::vector_t& p1, const Particle::vector_t& p2);

  /** Called whenever a start-tag is encountered and no handler is
   *  present (should never be called).
   */
  virtual void start_element(const std::string& name, AttributeList& attributes);

  /** Called whenever an end-tag is encountered and no handler is
   *  present (should never be called).
   */
  virtual void end_element(const std::string& name);

  /** Called whenever a start- or an end-tag is encountered and no
   *  handler is present (should never be called).
   */
  virtual void text(const std::string& name, const std::string& t);

};


#endif // RAX_PARSER_SPH_H
