/***************************************************************************
                          space.cpp  -  description
                             -------------------
    begin                : Fri Jan 3 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "askosmath/askosmath.h"
#include "askosio/log.h"

#include "errorcheck.h"
#include "space.h"


using namespace std;


int Space::nboxes(const double length, const double diameter) {
  REQUIRE(length>=0);
  REQUIRE(diameter>0);
  return static_cast<int>(floor(length/diameter))+1;
}


int Space::nbox(const double x) const {
  return static_cast<int>(floor(x/diameter));
}


int Space::get_n(const Particle::vector_t& p) const {
  REQUIRE( p.length()==3 );
  REQUIRE( in_space(p) );
  int n=get_n(nbox(p[0]),nbox(p[1]),nbox(p[2]));
  ENSURE(n>=0);
  ENSURE( n<particles.size()-1 );
  return n;
}


int Space::get_n(const Particle& particle) const {
  return get_n( particle.get_position() );
}


bool Space::in_space(const Particle::vector_t& p) const {
  REQUIRE( p.length()==3 );
  return p[0]>=0 && p[0]<=length_x &&
    p[1]>=0 && p[1]<=length_y &&
    p[2]>=0 && p[2]<=length_z;
}


bool Space::in_space(const Particle& particle) const {
  return in_space( particle.get_position() );
}


bool Space::in_box(const Particle& particle, 
		   const int nx, const int ny, const int nz) const {
  return in_space(particle) && get_n(particle)==get_n(nx,ny,nz);
}


bool Space::particles_in_their_box() const {
  for (int nz=0; nz<nboxes_z; ++nz) {
    for (int ny=0; ny<nboxes_y; ++ny) {
      for (int nx=0; nx<nboxes_x; ++nx) {

	// current box
	const vector<Particle>& box = particles[get_n(nx,ny,nz)];

	for ( vector<Particle>::const_iterator j=box.begin();
	      j!=box.end(); ++j ) {
	  const Particle& particle = *j; // current particle
	  if ( !in_box(particle,nx,ny,nz) ) {
	    return false;
	  }
	}

      }
    }
  }
  return true;
}


Space::Space(const double length_x_ini, const double length_y_ini, 
	     const double length_z_ini, 
	     const double diameter_ini) :
  length_x(length_x_ini),
  length_y(length_y_ini),
  length_z(length_z_ini),
  diameter(diameter_ini),
  diameter_sqr(sqr(diameter_ini)),
  nboxes_x(nboxes(length_x,diameter)),
  nboxes_y(nboxes(length_y,diameter)),
  nboxes_z(nboxes(length_z,diameter)),
  particles(nboxes_x*nboxes_y*nboxes_z+1), // last box is always empty
  nx_particle(0), ny_particle(0), nz_particle(0),
  particle_iterator(particles.front().begin()),
  current_position(0), // nullvector
  nx_neighbor(0), ny_neighbor(0), nz_neighbor(0),
  neighbor_iterator(particles.front().end()),
  neighbor_box_ptr(&particles.front()) {
  reset();
}


double Space::get_diameter() const {
  return diameter;
}


double Space::get_length_x() const {
  return length_x;
}


double Space::get_length_y() const {
  return length_y;
}


double Space::get_length_z() const {
  return length_z;
}


double Space::get_length(const int i) const {
  REQUIRE(0<=i && i<=3);
  switch (i) {
  case 0: return length_x;
  case 1: return length_y;
  case 2: return length_z;
  default: I(false);
  }
  I(false);
  return 0;
}


void Space::insert_particle(const Particle::vector_t& position) {
  REQUIRE( in_space(position) );
  int n=get_n(position);
  I(n>=0);
  I( n<particles.size()-1 );
  particles[n].push_back(Particle(position));
  ENSURE( euclidian_distance(particles[n].back().get_position(),position)==0 );
}


void Space::reset() {
  nx_particle=ny_particle=nz_particle=0;
  particle_iterator=particles.front().begin();
  if ( find_particle() ) {
    reset_neighborhood();
  }
}


bool Space::next() {
  REQUIRE( !at_end() );
  REQUIRE( particle_iterator!=get_box().end() );
  ++particle_iterator;
  return find_particle();
}


bool Space::find_particle() {
  while ( particle_iterator==get_box().end() ) {
    // we are at the end of the current box, so move one box forward
    ++nx_particle;
    if ( nx_particle >= nboxes_x ) {
      nx_particle=0;
      ++ny_particle;
      if ( ny_particle >= nboxes_y ) {
	ny_particle=0;
	++nz_particle;
	if ( nz_particle >= nboxes_z ) {
	  nz_particle=0;
	  particle_iterator=get_box().end();
	  I(nz_particle==0);
	  I(ny_particle==0);
	  I(nz_particle==0);
	  // no more boxes, no particle found
	  ENSURE( at_end() );
	  return false;
	}
      }
    }
    particle_iterator=get_box().begin();
  }
  // we have found a particle
  ENSURE( !at_end() );
  return true;
}


bool Space::at_end() const {
  return particle_iterator==get_box().end();
}

/*
Particle& Space::get_particle() {
  REQUIRE( !at_end() );
  return *particle_iterator;
}


const Particle& Space::get_particle() const {
  REQUIRE( !at_end() );
  return *particle_iterator;
}
*/


void Space::reset_neighborhood() {
  REQUIRE( !at_end() );
  double r_sqr;
  reset_neighborhood(r_sqr);
}


void Space::reset_neighborhood(double& r_sqr) {
  REQUIRE( !at_end() );
  reset_neighborhood(get_particle().get_position(),r_sqr);
  ENSURE( at_end_of_neighborhood() ||
	  r_sqr==sqr_euclidian_distance(get_particle().get_position(),
					get_neighbor_particle().get_position()) );
}


void Space::reset_neighborhood(const Particle::vector_t& p, double& r_sqr) {
  // set current position
  current_position=p;
  I( p.length()==3 );

  // calculate box of current position
  nx_particle=nbox(p[0]);
  ny_particle=nbox(p[1]);
  nz_particle=nbox(p[2]);

  // calculate first neighboring box
  /*
  nx_neighbor=max(nx_particle-1,0);
  ny_neighbor=max(ny_particle-1,0);
  nz_neighbor=max(nz_particle-1,0);
  */
  nx_neighbor=nx_particle-1;
  ny_neighbor=ny_particle-1;
  nz_neighbor=nz_particle-1;

  // cache pointer to neighboring box
  neighbor_box_ptr=&particles[get_n(nx_neighbor,ny_neighbor,nz_neighbor)];

  // initialize neighbor iterator
  neighbor_iterator=get_neighbor_box().begin();

  // find the next neighbor
  find_neighbor(r_sqr);

  ENSURE( at_end_of_neighborhood() ||
	  r_sqr==sqr_euclidian_distance(p,get_neighbor_particle().get_position()) );
}


/*
bool Space::has_neighbor(const Particle::vector_t& p) {
  // fixme: O(N^2)
  for ( reset(); !at_end(); next() ) {
    if ( sqr_euclidian_distance(get_particle().get_position(),p) <=
	 diameter_sqr ) {
      return true;
    }
  }
  return false;
}
*/


bool Space::find_neighbor() {
  REQUIRE( !at_end() );
  double r;
  return find_neighbor(r);
}


bool Space::next_neighbor_box() {
  ++nx_neighbor;
  if ( nx_neighbor >= min(nboxes_x,nx_particle+2) ) {
    nx_neighbor=max(nx_particle-1,0);
    ++ny_neighbor;
    if ( ny_neighbor >= min(nboxes_y,ny_particle+2) ) {
      ny_neighbor=max(ny_particle-1,0);
      ++nz_neighbor;
      if ( nz_neighbor >= min(nboxes_z,nz_particle+2) ) {
	nz_neighbor=max(nz_particle-1,0);
	neighbor_box_ptr=&particles[get_n(nx_neighbor,ny_neighbor,nz_neighbor)];
	return false;
      }
    }
  }
  neighbor_box_ptr=&particles[get_n(nx_neighbor,ny_neighbor,nz_neighbor)];
  return true;
}


bool Space::find_neighbor(double& r_sqr) {
  for (;;) {
    if ( neighbor_iterator==get_neighbor_box().end() ) {
      // we are at the end of the current box, so move one box
      // forward
      if ( !next_neighbor_box() ) {
	// no next neighboring box
	neighbor_iterator=get_neighbor_box().end();
	ENSURE( at_end_of_neighborhood() );
	return false;
      }
      neighbor_iterator=get_neighbor_box().begin();
      // neighbor_iterator points to be beginning of the next box
    } else if (  !is_neighbor(current_position, get_neighbor_particle().get_position(), 
			      r_sqr )  ) {
      // we are not at the end of the current box, so try next
      // particle in the current box
      ++neighbor_iterator;
    } else {
      ENSURE( !at_end_of_neighborhood() );
      ENSURE( is_neighbor(current_position,get_neighbor_particle().get_position()) );
      ENSURE( r_sqr==sqr_euclidian_distance(current_position,
					    get_neighbor_particle().get_position()) );
      return true;
    }
  }
}


void Space::reorder_particles() {
  LOG(log);
  for (int nz=0; nz<nboxes_z; ++nz) {
    for (int ny=0; ny<nboxes_y; ++ny) {
      for (int nx=0; nx<nboxes_x; ++nx) {

	// current box
	vector<Particle>& box = particles[get_n(nx,ny,nz)];

	int i=0;
	while ( i<box.size() ) {
	  while ( i<box.size() && !in_box(box[i],nx,ny,nz) ) {

	    if ( !in_space(box[i]) ) {
	      // current particle is even out of the space
	      // issue a warning
	      log.warn() << "Particle at position "
			 << Particle::string_vector(box[i].get_position()) 
			 << " out of space." << endl;
	    } else {
	      // current particle is out of its box
	      I( get_n(box[i]) != get_n(nx,ny,nz) );
	      // add the particle to the box it belongs to
	      particles[get_n(box[i])].push_back(box[i]);
	    }

	    // remove the particle from the current box
	    I( box.size()>0 );
	    box[i]=box[box.size()-1];
	    box.resize(box.size()-1);
	  }
	  ++i;
	}

      }
    }
  }
  ENSURE( particles_in_their_box() );
  reset();
}
