/***************************************************************************
                          space.h  -  description
                             -------------------
    begin                : Fri Jan 3 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SPACE_H
#define SPACE_H

#include <vector>

#include "particle.h"


/** Space to store particles.
 *
 *  Iterate though particles and their neighborhood (i.e. neighboring
 *  particles) in an efficient manner.
 *
 *  @author Moritz Franosch
 */
class Space {

private:

  const double length_x, length_y, length_z; // dimensions of the space

  const double diameter; // longest interacting distance

  const double diameter_sqr; // diameter^2

  const int nboxes_x, nboxes_y, nboxes_z; // number of boxes in each direction

  std::vector< std::vector<Particle> > particles; 
  // stores all particles in boxes
  // particles[nz*nboxes_y*nboxes_x + ny*nboxes_x + nx][n]
  // is the particle number n in box (nx,ny,nz),
  // 0 <= nx < nboxes_x, 0 <= ny < nboxes_y, 0 <= nz < nboxes_z

  // iterator through particles
  int nx_particle, ny_particle, nz_particle; // box of particle or of current position
  std::vector<Particle>::iterator particle_iterator; // particle in the box
  Particle::vector_t current_position; // position of current particle or current position

  // iterator through neighboring particles
  int nx_neighbor, ny_neighbor, nz_neighbor; // box
  std::vector<Particle>::iterator neighbor_iterator; // particles in the box
  std::vector<Particle>* neighbor_box_ptr; // point to the current neighboring box

  /** Returns the number of boxes of diameter to cover length.
   */
  static int nboxes(const double length, const double diameter);

  /** Returns the number of the box the coordinate x is in.
   */
  int nbox(const double x) const;

  /** Same as is_neighbor(), besides
   *
   *  @postcondition r_sqr is the square of the euclidian distance
   *  between the particle and its neighbor if the return value is
   *  true (undefined otherwise).
   */
  bool is_neighbor(const Particle::vector_t& p1, const Particle::vector_t& p2,
		   double& r_sqr) const {
    r_sqr = sqr_euclidian_distance(p1,p2);
    return r_sqr < diameter_sqr;
  }

  /** True, if position p1 is neighbor of p2, i.e. if the positons are less than
   *  diameter away.
   */
  bool is_neighbor(const Particle::vector_t& p1, const Particle::vector_t& p2) const {
    double r_sqr;
    return is_neighbor(p1,p2,r_sqr);
  }

  /** Returns true if particle_iterator already points to a particle.
   *
   *  Else, moves particle_iterator and
   *  (nx_particle,ny_particle,nz_particle) to the next particle.
   *
   *  Return true is this was possible, false otherwise.
   */
  bool find_particle();

  /** Returns true if neighbor_iterator already points to a neighbor.
   *
   *  Else, moves neighbor_iterator and
   *  (nx_neighbor,ny_neighbor,nz_neighbor) to the next neighbor.
   *
   *  Return true is this was possible, false otherwise.
   */
  bool find_neighbor();

  /** Same as find_neighbor(), besides
   *
   *  @postcondition r_sqr is the square of the euclidian distance
   *  between the particle and its neighbor if the return value is
   *  true (undefined otherwise).
   */
  bool find_neighbor(double& r_sqr);

  /** Returns the number of the box at position (nx,ny,nz).
   */
  int get_n(const int nx, const int ny, const int nz) const {
    /*
    REQUIRE(nx>=0);
    REQUIRE(ny>=0);
    REQUIRE(nz>=0);
    REQUIRE(nx<nboxes_x);
    REQUIRE(ny<nboxes_y);
    REQUIRE(nz<nboxes_z);
    */
    // If position is outside of the space ...
    if (nx<0 || ny<0 || nz<0 || 
	nx>=nboxes_x || ny>=nboxes_y || nz>=nboxes_z ) {
      // ..., then return the number of a box that is always empty.
      ENSURE( particles[particles.size()-1].empty() );
      return particles.size()-1;
    }
    int n=nz*nboxes_y*nboxes_x + ny*nboxes_x + nx;
    ENSURE(n>=0);
    ENSURE( n<particles.size() );
    return n;
  }

  /** Returns the number of the box the coordinate p belongs to.
   */
  int get_n(const Particle::vector_t& p) const;

  /** Returns the number of the box the particle belongs to, according
   *  to its coordinates.
   */
  int get_n(const Particle& particle) const;

  /** Returns the box of the current particle.
   */
  std::vector<Particle>& get_box() {
    return particles[get_n(nx_particle,ny_particle,nz_particle)];
  }

  /** Returns the box of the current particle.
   */
  const std::vector<Particle>& get_box() const {
    return particles[get_n(nx_particle,ny_particle,nz_particle)];
  }

  /** Returns a reference to the box of the current neighbor particle.
   */
  std::vector<Particle>& get_neighbor_box() {
    //    return particles[get_n(nx_neighbor,ny_neighbor,nz_neighbor)];
    return *neighbor_box_ptr;
  }

  /** Returns a reference to the box of the current neighbor particle.
   */
  const std::vector<Particle>& get_neighbor_box() const {
    //    return particles[get_n(nx_neighbor,ny_neighbor,nz_neighbor)];
    return *neighbor_box_ptr;
  }

  /** Sets (nx_neighbor,ny_neighbor,nz_neighbor) and neighbor_box_ptr
   *  to the next neighbor box.
   */
  bool next_neighbor_box();

  /** True if the coordinate p is inside the space.
   */
  bool in_space(const Particle::vector_t& p) const;

  /** True if the coordinates of the particle are inside the space.
   */
  bool in_space(const Particle& particle) const;

  /** True if the particle is in its box (nx,ny,nz).
   */
  bool in_box(const Particle& particle, 
	      const int nx, const int ny, const int nz) const;

  /** True if all particles are in the box they belong to.
   */
  bool particles_in_their_box() const;

protected:

  /** Returns a reference to current particle.
   */
  Particle& get_particle() {
    REQUIRE( !at_end() );
    return *particle_iterator;
  }

  /** Returns a reference to current particle in the neighborhood of
   *  the current particle.
   */
  Particle& get_neighbor_particle() {
    REQUIRE( !at_end_of_neighborhood() );
    return *neighbor_iterator;
  }

public: 

  /** Constructs a space of particles.
   *  
   *  The space has dimensions (length_x,length_y,length_z) and diameter of
   *  one cell is diameter.
   */
  Space(const double length_x, const double length_y, const double length_z, 
	const double diameter);

  /** Returns the diameter of a cell.
   */
  double get_diameter() const;

  /** Returns the length in x-dimension of the space.
   */
  double get_length_x() const;

  /** Returns the length in y-dimension of the space.
   */
  double get_length_y() const;

  /** Returns the length in z-dimension of the space.
   */
  double get_length_z() const;

  /** Returns the length in dimension i of the space.
   */
  double get_length(const int i) const;

  /** Insert particle at postion p.
   *
   *  Calls reset().
   *
   *  Invalidates all references to particles.
   */
  void insert_particle(const Particle::vector_t& p);

  /** Resets iterator through particles.
   */
  void reset();

  /** Iterator set to next particle.
   *
   *  Returns false if no next particle exists.
   */
  bool next();

  /** Last next() has returned false.
   */
  bool at_end() const;

  /** Returns a reference to current particle.
   */
  const Particle& get_particle() const {
    REQUIRE( !at_end() );
    return *particle_iterator;
  }

  /** Resets iterator through neighborhood of current particle.
   */  
  void reset_neighborhood();

  /** Same as reset_neighborhood(), besides
   *
   *  @postcondition r_sqr is the square of the euclidian distance
   *  between the particle and its neighbor if
   *  !at_end_of_neighborhood() (undefined otherwise).
   */
  void reset_neighborhood(double& r_sqr);

  /** Resets iterator through neighborhood of current position p.
   *
   *  Invalidates current particle.
   */  
  void reset_neighborhood(const Particle::vector_t& p, double& r_sqr);

  /** True if position p has a neighboring particle.
   *
   *  Invalidates iterator through particles.
   */
  //  bool has_neighbor(const Particle::vector_t& p);

  /** Iterator of neighborhood set to next particle.
   *
   *  Returns false if no next particle exists.
   */
  bool next_neighbor() {
    REQUIRE( !at_end_of_neighborhood() );
    ++neighbor_iterator;
    return find_neighbor();
  }

  /** Same as next_neighbor(), besides
   *
   *  @postcondition r_sqr is the square of the euclidian distance
   *  between the particle and its neighbor if the return value is
   *  true (undefined otherwise).
   */
  bool next_neighbor(double& r_sqr) {
    REQUIRE( !at_end_of_neighborhood() );
    ++neighbor_iterator;
    return find_neighbor(r_sqr);
  }

  /** Last next_neighborhood() has returned false.
   */
  bool at_end_of_neighborhood() const {
    return neighbor_iterator==get_neighbor_box().end();
  }

  /** Returns a reference to current particle in the neighborhood of
   *  the current particle.
   */
  const Particle& get_neighbor_particle() const {
    REQUIRE( !at_end_of_neighborhood() );
    return *neighbor_iterator;
  }

  /** Reorders the particles.
   *
   *  Must be called when particles' positions have been changed.
   *
   *  Resets all iterators (see reset()).
   */
  void reorder_particles();

};


#endif // SPACE_H
