/***************************************************************************
                          sph.cpp  -  description
                             -------------------
    begin                : Sat Jan 11 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "mylimits.h"

#include "blitz/tinyvec-et.h"

#include "askosio/log.h"
#include "askosio/string_convert.h"

#include "cusp_kernel.h"
#include "poly5_kernel.h"
#include "lucy_kernel.h"
#include "ipower.h"
#include "sph.h"


using namespace std;
using namespace blitz;


SPH::SPH(const int dimensions_ini,
	 const double xdim, const double ydim, const double zdim, 
	 const double h, 
	 const double velocity_of_sound,
	 const double dt_ini) :
  dimensions(dimensions_ini),
  space(xdim,ydim,zdim,h),
  dt(dt_ini),
  W( Pointer<Kernel>(new Poly5_Kernel(dimensions,h)) ),
  state_equation(velocity_of_sound,1000),
  g(0),
  nu(1.004E-6),
  boundary_width(numeric_limits<double>::quiet_NaN()),
  boundary_repulsive_force(numeric_limits<double>::quiet_NaN()),
  t(0),
  start_called(false) {
  REQUIRE(dimensions>=1);
  REQUIRE(dimensions<=3);
  g[dimensions-1] = -9.81;
  // default gravity is in y-dimension for 2 dimensions, in
  // z-dimension for 3 dimensions
  ENSURE( isnan(boundary_width) );
  ENSURE( isnan(boundary_repulsive_force) );
}


const SPH_Space& SPH::get_space() const {
  return space;
}


void SPH::set_dt(const double dt_ini) {
  dt=dt_ini;
}


void SPH::set_density(const double rho) {
  state_equation.set_rho_0(rho);
}


void SPH::set_viscosity(const double nu_ini) {
  nu=nu_ini;
}


void SPH::set_gravity(const Particle::vector_t& g_ini) {
  g=g_ini;
}


void SPH::set_velocity_of_sound(const double c) {
  state_equation.set_velocity_of_sound(c);
}


void SPH::set_boundary_width(const double b) {
  boundary_width=b;
}


void SPH::set_boundary_repulsive_force(const double f) {
  boundary_repulsive_force=f;
}


bool SPH::set_kernel(const string& name) {
  REQUIRE(W!=0);
  Pointer<Kernel> W_new;
  int d=W->get_d();
  double h=W->get_h();
  string lower_name = string_convert::tolower(name);
  if (lower_name=="poly5") {
    W_new = Pointer<Kernel>( new Poly5_Kernel(d,h) );
  } else if (lower_name=="cusp") {
    W_new = Pointer<Kernel>( new Cusp_Kernel(d,h) );
  } else if (lower_name=="lucy") {
    W_new = Pointer<Kernel>( new Lucy_Kernel(d,h) );
  } else {
    return false;
  }
  W=W_new;
  ENSURE(W!=0);
  return true;
}


void SPH::reset() {
  space.reset();
}


bool SPH::next() {
  return space.next();
}


bool SPH::at_end() const {
  return space.at_end();
}


const Particle& SPH::get_particle() const {
  return space.get_particle();
}


void SPH::insert_particle(const Particle::vector_t& p) {
  REQUIRE( !start_called );
  space.insert_particle(p);
}


void SPH::start() {
  REQUIRE( !start_called );
  LOG(log);
  start_called=true;

  /*
  // set initial particle density to rho_0
  for (reset(); !at_end(); next()) {
    space.set_particle_density(rho_0);
  }
  */

  // set initial particle mass
  /*
  int nparticles=0;
  double rho_mean = 0; // mean density
  for ( reset(); !at_end(); next() ) {
    double s=0; // sum_j W_ij
    double r_ij_sqr=0; // square of euclidian distance
    for ( space.reset_neighborhood(r_ij_sqr); !space.at_end_of_neighborhood(); 
	  space.next_neighbor(r_ij_sqr) ) {
      I( abs(r_ij_sqr-sqr_euclidian_distance(space.get_particle_position(),
					     space.get_neighbor_position())) <
	 numeric_limits<double>::epsilon() );
      s += W( sqrt(r_ij_sqr) ); // W_ij
    }
    rho_mean += s; // volume V_i = 1/(sum_j W_ij) = 1/s
    ++nparticles;
  }
  rho_mean /= nparticles;
  
  double rho_0 = state_equation.get_rho_0();
  double m_0 = rho_0/rho_mean; // mean mass to achieve mean density rho_0
  */

  // set particle masses to achieve low presure energy
  // m_0 = rho_0/N * V_tot
  // V_tot = sum_i V_i
  int N=0; // number of particles
  double V_tot=0; // whole volume
  double r_0=0; // mean particle distance
  for ( reset(); !at_end(); next() ) {
    double s=0; // sum_j W_ij
    double r_ij_sqr=0; // square of euclidian distance
    double r_min=W->get_h(); // nearest neighbor distance
    for ( space.reset_neighborhood(r_ij_sqr); !space.at_end_of_neighborhood(); 
	  space.next_neighbor(r_ij_sqr) ) {
      I( abs(r_ij_sqr-sqr_euclidian_distance(space.get_particle_position(),
					     space.get_neighbor_position())) <
	 numeric_limits<double>::epsilon() );
      double r_ij = sqrt(r_ij_sqr);
      s += (*W)(r_ij); // W_ij
      if (r_ij>0) {
	r_min = min(r_min,r_ij);
      }
    }
    V_tot += 1/s; // volume V_i = 1/(sum_j W_ij) = 1/s
    I(r_min>0);
    I( r_min<=W->get_h() );
    r_0 += r_min;
    ++N;
  }
  double rho_0 = state_equation.get_rho_0();
  double m_0 = rho_0*V_tot/N;
  r_0 /= N;

  for ( reset(); !at_end(); next() ) {
    space.set_particle_mass(m_0);
    space.set_particle_density(rho_0); // fixme
  }

  // calculate densitiy for each particle
  // rho_i = sum_j m_j W_ij
  /*
  for ( reset(); !at_end(); next() ) {
    double rho_i=0; // density
    double r_ij_sqr=0; // square of euclidian distance
    for ( space.reset_neighborhood(r_ij_sqr); !space.at_end_of_neighborhood(); 
	  space.next_neighbor(r_ij_sqr) ) {
      I( abs(r_ij_sqr-sqr_euclidian_distance(space.get_particle_position(),
					     space.get_neighbor_position())) <
	 numeric_limits<double>::epsilon() );
      double m_j = space.get_neighbor_mass();
      rho_i += m_j * W( sqrt(r_ij_sqr) ); // W_ij
    }
    space.set_particle_density(rho_i);
  }
  */

  // If boundary_width is not set ...
  if ( isnan(boundary_width) ) {

    // ... set to half of mean particle distance.
    boundary_width = r_0/2;

    // If this is zero (only one particle?) ...
    if (boundary_width==0) {
      // set to smoothing length h.
      boundary_width = W->get_h();
    }

    // inform user
    log.info() << "Boundary width set to " << boundary_width << " m." << endl;
  }

  // If boundary_repulsive_force is not set ...
  if ( isnan(boundary_repulsive_force) ) {
    // set to expected gravity force.
    boundary_repulsive_force=0;
    for (int i=0; i<Particle::vector_t::numElements; ++i) {
      // With mass M = rho_0*length/2*r_0^(d-1), M * g is the force.
      boundary_repulsive_force = 
	max( boundary_repulsive_force, 
	     abs( rho_0 * 0.5*space.get_length(i) * pow(r_0,dimensions-1) * g[i] ) );
    }
    // inform user
    log.info() << "Boundary repulsive force set to " 
	       << boundary_repulsive_force << " N." << endl;
  }

  ENSURE(start_called);
}


double SPH::get_time() const {
  return t;
}


double SPH::get_dt() const {
  return dt;
}


double SPH::get_h() const {
  return W->get_h();
}


int SPH::get_dimensions() const {
  return W->get_d();
}


double SPH::get_boundary_width() const {
  return boundary_width;
}


double SPH::get_gravity_energy() {
  double E_g=0;
  for ( reset(); !at_end(); next() ) {
    // E_g = - m g * h
    E_g += - get_particle().get_mass() * dot(g,get_particle().get_position());
  }
  return E_g;
}


double SPH::get_pressure_energy() {
  REQUIRE(start_called);
  double E_p=0;
  for ( reset(); !at_end(); next() ) {

    double rho = get_particle().get_density();
    double m = get_particle().get_mass();

    /*
    // V = m / rho
    double V = get_particle().get_mass()/rho;

    // E_p = 1/2 p V
    E_p += 0.5 * state_equation.get_pressure_from_density(rho) * V;
    */

    E_p += state_equation.get_energy(m,rho);

  }
  return E_p;
}


double SPH::get_kinetic_energy() {
  double E_kin=0;
  for ( reset(); !at_end(); next() ) {
    // E_kin = 1/2 m v^2
    E_kin += 0.5 * get_particle().get_mass() * 
      sqr(euclidian_norm(get_particle().get_velocity()));
  }
  return E_kin;
}


double SPH::get_max_acceleration() {
  double a_max=0;
  for ( reset(); !at_end(); next() ) {
    a_max = max(a_max,euclidian_norm(get_particle().get_acceleration()));
  }
  return a_max;
}


double SPH::get_density_deviation() {
  REQUIRE(start_called);
  double r=0;
  int nparticles=0;
  for ( reset(); !at_end(); next() ) {
    r += sqr(get_particle().get_density()-state_equation.get_rho_0());
    ++nparticles;
  }
  r = sqrt(r/nparticles)/state_equation.get_rho_0();
  ENSURE(r>=0);
  return r;
}


void SPH::update() {
  REQUIRE(start_called);

  // for each particle
  for ( reset(); !at_end(); next() ) {

    // set acceleration to gravity acceleration
    Particle::vector_t dv(g);

    // for acceleration due to pressure gradient
    Particle::vector_t grad_p_i(0); // grad p_i = sum_j V_j (p_j+p_i) grad_i W_ij
    double rho_i = space.get_particle_density();
    double p_i = state_equation.get_pressure_from_density(rho_i);

    // for density variation drho_i/dt
    double div_v_i=0; // div v_i = sum_j V_j (v_j-v_i) * grad_i W_ij

    double r_ij_sqr=numeric_limits<double>::quiet_NaN(); // square of euclidian distance
    double r_ij=numeric_limits<double>::quiet_NaN(); // euclidian distance

    Particle::vector_t grad_i__W_ij(0); // gradient of kernel

    double m_j=numeric_limits<double>::quiet_NaN(); // mass of the neighbor particle
    double rho_j=numeric_limits<double>::quiet_NaN(); // density at the neighbor particle
    double V_j=numeric_limits<double>::quiet_NaN(); // volume of the neighbor particle
    double p_j=numeric_limits<double>::quiet_NaN(); // pressure at the neighbor particle

    for ( space.reset_neighborhood(r_ij_sqr); !space.at_end_of_neighborhood(); 
	  space.next_neighbor(r_ij_sqr) ) {
      // r_ij_sqr contains square of euclidian distance
      I( abs(r_ij_sqr-sqr_euclidian_distance(space.get_particle_position(),
					     space.get_neighbor_position())) <
	 numeric_limits<double>::epsilon() );

      r_ij = sqrt(r_ij_sqr);

      m_j = space.get_neighbor_mass();
      rho_j = space.get_neighbor_density();
      V_j = m_j/rho_j;

      W->grad_r( space.get_particle_position(),
		 space.get_neighbor_position(),
		 r_ij,
		 grad_i__W_ij );

      // for density variation drho_i/dt
      div_v_i += V_j * dot( space.get_neighbor_velocity() - space.get_particle_velocity(),
			    grad_i__W_ij );
      // V_j (v_j-v_i) * grad_i__W_ij;

      // for acceleration due to pressure gradient
      p_j = state_equation.get_pressure_from_density(rho_j);
      grad_p_i += V_j * (p_j+p_i) * grad_i__W_ij;
      // '+' because conservation of momentum, '-' for constant pressure? fixme?
      // conservation of energy not possible

      // for acceleration due to viscosity
      if ( r_ij>0 ) {
	dv += V_j * nu * (1+rho_j/rho_i) * W->dW(r_ij)/r_ij * 
	  ( space.get_particle_velocity() - space.get_neighbor_velocity() );
	// V_j nu (rho_i+rho_j)/rho_i * dW(r_ij)/r_ij * (v_i-v_j) =
	// m_j (mu_i+mu_j)/(rho_i*rho_j) * dW(r_ij)/r_ij * (v_i-v_j)
      }

    }

    // for density variation drho_i/dt
    double drho_i = - space.get_particle_density() * div_v_i; // - rho_i (div v_i)
    // Why not an additonal + grad rho * v, because drho = div(rho v) ?
    space.set_particle_drho(drho_i);

    // for acceleration due to pressure gradient
    dv += - (1/rho_i) * grad_p_i;

    // for acceleration due to being near the boundary
    for (int k=0; k<Particle::vector_t::numElements; ++k) {
      double r=0; // distance from boundary
      if ( space.get_particle_position()[k] < boundary_width ) {
	r=space.get_particle_position()[k];
      }
      if ( space.get_particle_position()[k] > space.get_length(k)-boundary_width ) {
	r=space.get_particle_position()[k]-space.get_length(k);
      }

      if ( r!=0 ) {
	// distance r from boundary is smaller than boundary_width

	// boundary is sticky
	dv += - nu / sqr(r) * space.get_particle_velocity();

	// boundary is repulsive
	double m_i = space.get_particle_mass();
	dv[k] += boundary_repulsive_force/m_i * sign(r) * 
	  ( pow<4>(boundary_width/r)-pow<2>(boundary_width/r) );
      }

    }

    space.set_particle_acceleration(dv);

  }

  // integrate particle positions and update simulated time
  t += space.update_particles(dt);

}


bool SPH::is_inner_point(const Particle::vector_t& p) {
  return get_implicit_surface(p)>0;
}


double SPH::get_implicit_surface(const Particle::vector_t& p) {
  return 2.0*get_density(p)/state_equation.get_rho_0()-1;
  //  return p[0]+p[1]-1.05;
  /*
  Particle::vector_t m(0.5,0.5,0.5);
  return 0.3-euclidian_distance(m,p);
  */
}


double SPH::get_density(const Particle::vector_t& p) {
  double rho_i=0; // density
  double r_ij_sqr=0; // square of euclidian distance
  for ( space.reset_neighborhood(p,r_ij_sqr); !space.at_end_of_neighborhood(); 
	space.next_neighbor(r_ij_sqr) ) {
    I( abs(r_ij_sqr-sqr_euclidian_distance(p,space.get_neighbor_position())) <
       numeric_limits<double>::epsilon() );
    double m_j = space.get_neighbor_mass();
    rho_i += m_j * (*W)( sqrt(r_ij_sqr) ); // W_ij
  }
  return rho_i;
}


ostream& operator<<(ostream& o, SPH& sph) {
  double gravity_energy=sph.get_gravity_energy();
  double pressure_energy=sph.get_pressure_energy();
  double kinetic_energy=sph.get_kinetic_energy();

  // critical value due to acceleration
  // should be less than 0.25
  double c_a = sqrt(sph.get_max_acceleration()/sph.get_h())*sph.get_dt();

  o << "<space" << endl
    << "  time=\"" << sph.get_time() << "\"" << endl
    << "  energy=\"" << gravity_energy+pressure_energy+kinetic_energy << "\"" << endl
    << "  gravity_energy=\"" << gravity_energy << "\"" << endl
    << "  pressure_energy=\"" << pressure_energy << "\"" << endl
    << "  kinetic_energy=\"" << kinetic_energy << "\"" << endl
    << "  crit_acceleration=\"" << c_a << "\"" << endl
    << "  density_deviation=\"" << sph.get_density_deviation() << "\"" << endl
    << "  dimensions=\"" << sph.get_dimensions() << "\"" << endl
    << ">" << endl;
  o << "  <particles>" << endl;
  // for each particle
  for ( sph.reset(); !sph.at_end(); sph.next() ) {
    const Particle& particle = sph.get_particle();
    o << "    <particle>" << endl;
    o << "      <position> " 
      << Particle::list_vector(particle.get_position()) << " </position>" << endl;
    o << "      <velosity> " 
      << Particle::list_vector(particle.get_velocity()) << " </velosity>" << endl;
    o << "      <acceleration> " 
      << Particle::list_vector(particle.get_acceleration())
      << " </acceleration>" << endl;
    o << "      <mass> " << particle.get_mass() << " </mass>" << endl;
    o << "      <density> " << particle.get_density() << " </density>" << endl;
    o << "    </particle>" << endl;
  }
  o << "  </particles>" << endl;
  o << "</space>" << endl;
  return o;
}
