/***************************************************************************
                          sph.h  -  description
                             -------------------
    begin                : Sat Jan 11 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SPH_H
#define SPH_H


#include <ostream>

#include "askosdata/pointer.h"

#include "sph_space.h"
#include "particle.h"
#include "kernel.h"
#include "state_equation.h"


/** Smoothed Particle Hydrodynamics.
 *
 *  @author Moritz Franosch
 */
class SPH {

private:

  int dimensions; // number of dimensions
  SPH_Space space; // particle space
  double dt; // timestep
  Pointer<Kernel> W; // kernel
  State_Equation state_equation; // equation of state
  Particle::vector_t g; // gravity acceleration
  double nu; // viscosity
  double boundary_width; // width of repulsive boundary
  double boundary_repulsive_force; // constant for repulsive force near boundary
  double t; // simulated time
  bool start_called; // true if start() has been called, false otherwise

public:

  /** Constructs a SPH space.
   */
  SPH(const int dimensions,
      const double xdim, const double ydim, const double zdim, 
      const double h, 
      const double velocity_of_sound,
      const double dt);

  /** Returns a reference to the SPH_Space used.
   */
  const SPH_Space& get_space() const;

  /** Sets timestep.
   */
  void set_dt(const double dt);

  /** Sets the initial density of the fluid.
   */
  void set_density(const double rho);

  /** Sets the viscosity of the fluid.
   */
  void set_viscosity(const double nu);

  /** Sets the gravity acceleration.
   */
  void set_gravity(const Particle::vector_t& g);

  /** Sets the velocity of sound.
   */
  void set_velocity_of_sound(const double c);

  /** Sets width of the boundary.
   */
  void set_boundary_width(const double b);

  /** Sets repulsive force of the boundary.
   */
  void set_boundary_repulsive_force(const double f);

  /** Sets the kernel.
   *
   *  Returns true if successful.
   */
  bool set_kernel(const std::string& name);

  /** Resets iterator through particles.
   */
  void reset();

  /** Iterator set to next particle.
   *
   *  Returns false if no next particle exists.
   */
  bool next();

  /** Last next() has returned false.
   */
  bool at_end() const;

  /** Returns a reference to current particle.
   */
  const Particle& get_particle() const;

  /** Insert particle at postion p.
   *
   *  Calls reset().
   *
   *  Invalidates all references to particles.
   */
  void insert_particle(const Particle::vector_t& p);

  /** Must be called after all particles are inserted and before the
   *  first update().
   */
  void start();

  /** Returns current simulated time in s.
   */
  double get_time() const;

  /** Returns current timestep in s.
   */
  double get_dt() const;

  /** Returns the radius of the kernel (smoothing length) in m.
   */
  double get_h() const;

  /** Returns the number of dimensions.
   */
  int get_dimensions() const;

  /** Returns width of the boundary.
   */
  double get_boundary_width() const;

  /** Returns current gravity energy in Joule.
   */
  double get_gravity_energy();

  /** Returns current pressure energy in Joule.
   */
  double get_pressure_energy();

  /** Returns current kinetic energy in Joule.
   */
  double get_kinetic_energy();

  /** Returns the maximum acceleration of the particles in m/s^2.
   */
  double get_max_acceleration();

  /** Returns the deviation sqrt( <(rho_i-rho_0)^2> )/rho_0 of the densities.
   */
  double get_density_deviation();

  /** Updates positions and velocities of all particles according to
   *  SPH.
   */
  void update();

  /** True iff position p is inside the fluid, i.e. iff
   *  get_implicit_surface()>0.
   *
   *  Invalidates iterator through particles.
   */
  bool is_inner_point(const Particle::vector_t& p);

  /** A differentiable function that returns a value >0 iff position p
   *  is inside the fluid and <0 otherwise.
   */
  double get_implicit_surface(const Particle::vector_t& p);

  /** Returns the density at position p.
   *
   *  Invalidates iterator through particles.
   */
  double get_density(const Particle::vector_t& p);

};


std::ostream& operator<<(std::ostream& o, SPH& sph);


#endif // SPH_H
