/***************************************************************************
                          sph_space.h  -  description
                             -------------------
    begin                : Thu Jan 16 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SPH_SPACE_H
#define SPH_SPACE_H

#include "space.h"


/** Particle space with access functions for SPH.
 *
 *  @author Moritz Franosch
 */
class SPH_Space : public Space {

private:

  int step; // step of integration method

  /** Ensures the boundary condition.
   *
   *  Sets velocity to zero (fixme).
   */
  void ensure_boundary_condition(Particle& particle);

public: 

  /** Constructs a space of particles.
   *  
   *  The space has dimensions (length_x,length_y,length_z) and diameter of
   *  one cell is diameter.
   */
  SPH_Space(const double length_x, const double length_y, const double length_z, 
	    const double diameter);

  /** Returns the position of the current particle.
   */
  const Particle::vector_t& get_particle_position() const {
    return get_particle().get_position();
  }

  /** Returns the position of the current neighbor particle.
   */
  const Particle::vector_t& get_neighbor_position() const {
    return get_neighbor_particle().get_position();
  }

  /** Returns the velocity of the current particle.
   */
  const Particle::vector_t& get_particle_velocity() const {
    return get_particle().get_velocity();
  }

  /** Returns the velocity of the current neighbor particle.
   */
  const Particle::vector_t& get_neighbor_velocity() const {
    return get_neighbor_particle().get_velocity();
  }

  /** Returns the mass of the current particle.
   */
  double get_particle_mass() const {
    return get_particle().get_mass();
  }

  /** Returns the mass of the current neighbor particle.
   */
  double get_neighbor_mass() const {
    return get_neighbor_particle().get_mass();
  }

  /** Sets the mass of the current particle to m.
   */
  void set_particle_mass(const double m);

  /** Returns the density of the current particle.
   */
  double get_particle_density() const {
    return get_particle().get_density();
  }

  /** Sets the density of the current particle to rho.
   */
  void set_particle_density(const double rho);

  /** Returns the density of the current neighbor particle.
   */
  double get_neighbor_density() const {
    return get_neighbor_particle().get_density();
  }

  /** Set the drho/dt of the current particle.
   */
  void set_particle_drho(const double& drho);

  /** Set the acceleration of the current particle.
   */
  void set_particle_acceleration(const Particle::vector_t& a);

  /** Velocity and positons of particles is updated according to the
   *  current accelerations.
   *
   *  Ensures boundary conditions are met.
   *
   *  The advance in time is returned.
   *
   *  Also reorganizes particle layout.
   *
   *  Invalidates all references to particles.
   */
  double update_particles(const double dt);

};


#endif // SPH_SPACE_H
