/***************************************************************************
                          triangulator.cpp  -  description
                             -------------------
    begin                : Sat Mar 29 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "mylimits.h"

#include "blitz/tinyvec-et.h"

#include "askosio/log.h"

#include "triangulator.h"
#include "triangle.h"
#include "smooth_triangle.h"


using namespace std;


void Triangulator::recurse_or_add(SPH& sph,
				  const Particle::vector_t& p1,
				  const Particle::vector_t& p2,
				  const Particle::vector_t& p3,
				  const Particle::vector_t& p4,
				  const double a,
				  const double d_min) {
  LOG(log);
  log.debug() << "a=" << a << endl;
  log.debug() << "p1=" << p1 << endl;
  log.debug() << "p2=" << p2 << endl;
  log.debug() << "p3=" << p3 << endl;
  log.debug() << "p4=" << p4 << endl;
  // If the sides of the tetrahedron Tetrahedron(p1,p2,p3,p4) with
  // length a ...
  /*  cerr << "|p1-p2|=" << euclidian_norm(p1-p2) << endl;
  cerr << "a=" << a << endl;
  */
  REQUIRE( abs(euclidian_norm(p1-p2)-a) < 10*numeric_limits<double>::epsilon() );
  REQUIRE( abs(euclidian_norm(p1-p3)-a) < 10*numeric_limits<double>::epsilon() );
  REQUIRE( abs(euclidian_norm(p1-p4)-a) < 10*numeric_limits<double>::epsilon() );
  REQUIRE( abs(euclidian_norm(p2-p3)-a) < 10*numeric_limits<double>::epsilon() );
  REQUIRE( abs(euclidian_norm(p2-p4)-a) < 10*numeric_limits<double>::epsilon() );
  REQUIRE( abs(euclidian_norm(p3-p4)-a) < 10*numeric_limits<double>::epsilon() );
  // ... are longer than d_min, ...
  if (a>d_min) {
    // ... then recurse into the five smaller tetrahedra.
    recurse_or_add(sph,
		   p1,
		   p1+(p2-p1)/2,p1+(p3-p1)/2,p1+(p4-p1)/2,
		   a/2,d_min);
    recurse_or_add(sph,
		   p2,
		   p2+(p3-p2)/2,p2+(p4-p2)/2,p2+(p1-p2)/2,
		   a/2,d_min);
    recurse_or_add(sph,
		   p3,
		   p3+(p4-p3)/2,p3+(p1-p3)/2,p3+(p2-p3)/2,
		   a/2,d_min);
    recurse_or_add(sph,
		   p4,
		   p4+(p1-p4)/2,p4+(p2-p4)/2,p4+(p3-p4)/2,
		   a/2,d_min);
    recurse_or_add(sph,
		   p4,
		   p4+(p1-p4)/2,p4+(p2-p4)/2,p4+(p3-p4)/2,
		   a/2,d_min);
  } else {
    consider_tetrahedron(sph,p1,p2,p3,p4);
  }
}


void Triangulator::consider_tetrahedron(SPH& sph,
					const Particle::vector_t& corner,
					const Particle::vector_t& p1,
					const Particle::vector_t& p2,
					const Particle::vector_t& p3,
					const Particle::vector_t& p4,
					int type,
					double d_max) {
  REQUIRE(type==0 || type==1);
  if (type==0) {
    // do not turn the tetrahedron
    consider_tetrahedron(sph,corner+p1,corner+p2,corner+p3,corner+p4);
  } else {
    // turn the tetrahedron 90 around the z-axis and move it into
    // position
    Particle::vector_t q1, q2, q3, q4;
    turn_and_move(p1,d_max,q1);
    turn_and_move(p2,d_max,q2);
    turn_and_move(p3,d_max,q3);
    turn_and_move(p4,d_max,q4);
    consider_tetrahedron(sph,corner+q1,corner+q2,corner+q3,corner+q4);
  }
}


void Triangulator::turn_and_move(const Particle::vector_t& p,
				 const double d_max,
				 Particle::vector_t& q) {
  REQUIRE(Particle::vector_t::numElements==3);
  q[2] = p[2]; // z-coordinate remains
  q[0] = -p[1]+d_max; // x'=-y+d_max;
  q[1] = p[0]; // y'=x;
}


void Triangulator::consider_tetrahedron(SPH& sph,
					const Particle::vector_t& p1,
					const Particle::vector_t& p2,
					const Particle::vector_t& p3,
					const Particle::vector_t& p4) {
  LOG(log);
  // If the tetrahedron ...
  bool is_inside1 = sph.is_inner_point(p1);
  bool is_inside2 = sph.is_inner_point(p2);
  bool is_inside3 = sph.is_inner_point(p3);
  bool is_inside4 = sph.is_inner_point(p4);
  int ninside = is_inside1+is_inside2+is_inside3+is_inside4;
  if (
      // ... totally belongs to the volume ...
      ninside==4
      // ... or ...
      ||
      // ... totally does not belong to it,
      ninside==0
      ) {
    // ... then do nothing.
    return;
  } else {
    // The tetrahedron only belongs partly to the volume.
    if (ninside==1) {
      // Exactly one corner belongs to the volume.
      if (is_inside1) {
	add_triangle_from_tetrahedron(sph,p1,p2,p3,p4);
      } else if (is_inside2) {
	add_triangle_from_tetrahedron(sph,p2,p3,p4,p1);
      } else if (is_inside3) {
	add_triangle_from_tetrahedron(sph,p3,p4,p1,p2);
      } else if (is_inside4) {
	add_triangle_from_tetrahedron(sph,p4,p1,p2,p3);
      }
    } else if (ninside==3) {
      // Exactly one corner does not belong to the volume.
      if (!is_inside1) {
	add_triangle_from_tetrahedron(sph,p1,p2,p3,p4);
      } else if (!is_inside2) {
	add_triangle_from_tetrahedron(sph,p2,p3,p4,p1);
      } else if (!is_inside3) {
	add_triangle_from_tetrahedron(sph,p3,p4,p1,p2);
      } else if (!is_inside4) {
	add_triangle_from_tetrahedron(sph,p4,p1,p2,p3);
      }
    } else {
      I(ninside==2);
      // Exactly two corners belong to the volume.
      if (is_inside1) {
	if (is_inside2) {
	  add_triangles_from_tetrahedron(sph,p1,p2,p3,p4);
	} else if (is_inside3) {
	  add_triangles_from_tetrahedron(sph,p1,p3,p4,p2);
	} else {
	  I(is_inside4);
	  add_triangles_from_tetrahedron(sph,p1,p4,p2,p3);
	}
      } else if (is_inside2) {
	if (is_inside3) {
	  add_triangles_from_tetrahedron(sph,p2,p3,p1,p4);
	} else {
	  I(is_inside4);
	  add_triangles_from_tetrahedron(sph,p2,p4,p3,p1);
	}
      } else {
	I(is_inside3 && is_inside4);
	add_triangles_from_tetrahedron(sph,p3,p4,p1,p2);
      }
    }
  }
}


void Triangulator::add_triangle_from_tetrahedron(SPH& sph,
						 const Particle::vector_t& p1,
						 const Particle::vector_t& p2,
						 const Particle::vector_t& p3,
						 const Particle::vector_t& p4) {
  REQUIRE( ( sph.is_inner_point(p1) 
	     && !sph.is_inner_point(p2) && !sph.is_inner_point(p3) && !sph.is_inner_point(p4) 
	     ) ||
	   ( !sph.is_inner_point(p1) 
	     && sph.is_inner_point(p2) && sph.is_inner_point(p3) && sph.is_inner_point(p4) 
	     )
	   );
  Particle::vector_t p1_2;
  get_interpolation(sph,p1,p2,p1_2);
  Particle::vector_t p1_3;
  get_interpolation(sph,p1,p3,p1_3);
  Particle::vector_t p1_4;
  get_interpolation(sph,p1,p4,p1_4);

  add_triangle(sph,p1_2,p1_4,p1_3);
  // the normal vector of the triangle points to the outside
}


void Triangulator::add_triangles_from_tetrahedron(SPH& sph,
						  const Particle::vector_t& p1,
						  const Particle::vector_t& p2,
						  const Particle::vector_t& p3,
						  const Particle::vector_t& p4) {
  LOG(log);
  REQUIRE( sph.is_inner_point(p1) );
  REQUIRE( sph.is_inner_point(p2) );
  REQUIRE( !sph.is_inner_point(p3) );
  REQUIRE( !sph.is_inner_point(p4) );
  REQUIRE( is_right_handed(p1,p2,p3,p4) );

  Particle::vector_t p1_3;
  get_interpolation(sph,p1,p3,p1_3);
  Particle::vector_t p1_4;
  get_interpolation(sph,p1,p4,p1_4);
  Particle::vector_t p2_4;
  get_interpolation(sph,p2,p4,p2_4);
  Particle::vector_t p2_3;
  get_interpolation(sph,p2,p3,p2_3);

  add_triangle(sph,p1_3,p2_4,p1_4);
  add_triangle(sph,p1_3,p2_3,p2_4);
  // the normal vector of the triangles point to the outside
}



void Triangulator::add_triangle(SPH& sph,
				const Particle::vector_t& p1,
				const Particle::vector_t& p2,
				const Particle::vector_t& p3) {
  Particle::vector_t n1, n2, n3; // the normals at each corner
  compute_normal(sph,p1,n1);
  compute_normal(sph,p2,n2);
  compute_normal(sph,p3,n3);
  Pointer<Geometry> triangle_ptr(new Smooth_Triangle(p1,p2,p3,n1,n2,n3) );
  triangles.add(triangle_ptr);
}


void Triangulator::compute_normal(SPH& sph,
				  const Particle::vector_t& p,
				  Particle::vector_t& n) {
  double f=sph.get_implicit_surface(p);
  for (int i=0; i<Particle::vector_t::numElements; ++i) {
    Particle::vector_t p2=p;

    // Numerical differentiation. We arbitrarily choose d_max/10 as
    // "small enough".
    p2[i] += d_max/10;
    n[i] = f-sph.get_implicit_surface(p2);

  }
  double norm_n=euclidian_norm(n);
  if (norm_n!=0) {
    n /= euclidian_norm(n);
  }
}


void Triangulator::get_interpolation(SPH& sph,
				     const Particle::vector_t& p0,
				     const Particle::vector_t& p1,
				     Particle::vector_t& p) {
  REQUIRE( ( sph.is_inner_point(p0) && !sph.is_inner_point(p1) ) ||
	   ( !sph.is_inner_point(p0) && sph.is_inner_point(p1) ) );
  double f_p0=sph.get_implicit_surface(p0);
  double f_p1=sph.get_implicit_surface(p1);
  I( ( f_p0<=0 && f_p1>=0 ) || ( f_p0>=0 && f_p1<=0 ) );
  I( abs(f_p0)+abs(f_p1)>0 );
  Particle::vector_t dp=p1-p0;
  I( euclidian_norm(dp)>0 );
  p = p0 + dp*(  ( abs(f_p0)/( abs(f_p0) + abs(f_p1) ) )  );
}


bool Triangulator::is_right_handed(const Particle::vector_t& p1,
				   const Particle::vector_t& p2,
				   const Particle::vector_t& p3,
				   const Particle::vector_t& p4) const {
  Particle::vector_t p2_3=p3-p2;
  Particle::vector_t p2_4=p4-p2;
  return dot(cross(p2_3,p2_4),p1-p2)>=0;
}


Triangulator::Triangulator(SPH& sph, const double d_max_ini) :
  triangles(),
  d_max(d_max_ini) {
  const SPH_Space& space=sph.get_space();
  const double border=2*sph.get_h()+2*d_max; // additional room
  // Decompose the regtangular space into cubes with ...
  int type_x=0;
  for (double x=-border; x<=space.get_length_x()+border; x+=d_max, type_x=(type_x+1) % 2) {
    int type_y=type_x;
    for (double y=-border; y<space.get_length_y()+border; y+=d_max, type_y=(type_y+1) % 2) {
      int type_z=type_y; // type of the tetrahedron (0 or 1), turned by 90 around the z-axis
      for (double z=-border; z<=space.get_length_z()+border; z+=d_max, type_z=(type_z+1) % 2) {
	// ... corners (x,y,z) and length of the sides d_max.
	// Decompose the cube into four tetraherdrons according to 
	// http://mathworld.wolfram.com/Tetrahedron.html.

	Particle::vector_t corner(x,y,z);

	// tetrahedron with top corner (d_max,0,0)
	{
	  Particle::vector_t p1(d_max,0,0);
	  Particle::vector_t p2(d_max,d_max,0);
	  Particle::vector_t p3(d_max,0,d_max);
	  Particle::vector_t p4(0,0,0);
	  consider_tetrahedron(sph,corner,p1,p2,p3,p4,type_z,d_max);
	}
	
	// tetrahedron with top corner (0,d_max,0)
	{
	  Particle::vector_t p1(0,d_max,0);
	  Particle::vector_t p2(0,0,0);
	  Particle::vector_t p3(0,d_max,d_max);
	  Particle::vector_t p4(d_max,d_max,0);
	  consider_tetrahedron(sph,corner,p1,p2,p3,p4,type_z,d_max);
	}
	
	// tetrahedron with top corner (0,0,d_max)
	{
	  Particle::vector_t p1(0,0,d_max);
	  Particle::vector_t p2(0,0,0);
	  Particle::vector_t p3(d_max,0,d_max);
	  Particle::vector_t p4(0,d_max,d_max);
	  consider_tetrahedron(sph,corner,p1,p2,p3,p4,type_z,d_max);
	}

	// tetrahedron with top corner (d_max,d_max,d_max)
	{
	  Particle::vector_t p1(d_max,d_max,d_max);
	  Particle::vector_t p2(d_max,0,d_max);
	  Particle::vector_t p3(d_max,d_max,0);
	  Particle::vector_t p4(0,d_max,d_max);
	  consider_tetrahedron(sph,corner,p1,p2,p3,p4,type_z,d_max);
	}
	
	// middle tetrahedron with side lengths sqrt(2)*d_max
	{
	  Particle::vector_t p1(0,0,0);
	  Particle::vector_t p2(d_max,0,d_max);
	  Particle::vector_t p3(0,d_max,d_max);
	  Particle::vector_t p4(d_max,d_max,0);
	  consider_tetrahedron(sph,corner,p1,p2,p3,p4,type_z,d_max);
	}

      }
    }
  }
  

  /*
  // calculate regular tetrahedron with top corner p1 and bottom
  // corners p2, p3, p4, such that the sph space is contained in it

  const SPH_Space& space=sph.get_space();

  // length of the sides of the incube
  double b = max(space.get_length_x(),max(space.get_length_y(),space.get_length_z()));

  // inradius
  double r = (sqrt(2.0)+1)/2 * b;
  
  // additional space
  r += d_min;

  // length of a side of the tetrahedron
  double a = 12*r/sqrt(6.0); // http://mathworld.wolfram.com/Tetrahedron.html

  // height of the tetrahedron
  double h = sqrt(6.0)/3*a; // http://mathworld.wolfram.com/Tetrahedron.html

  // vertices http://mathworld.wolfram.com/Tetrahedron.html
  double x = sqrt(3.0)/3 * a;
  double d = sqrt(3.0)/6 * a;
  Particle::vector_t p1(0,0,h);
  Particle::vector_t p2(x,0,0);
  Particle::vector_t p3(-d,a/2,0);
  Particle::vector_t p4(-d,-a/2,0);

  // mid point
  Particle::vector_t m(0,0,r);

  // move the vertices by dp such that the new midpoint is (d/2,d/2,d/2)
  Particle::vector_t dp=Particle::vector_t(b/2,b/2,b/2)-m;
  p1 += dp;
  p2 += dp;
  p3 += dp;
  p4 += dp;

  // the tetrahedron (p1,p2,p3,p4) now contains the sph space and in
  // addition leaves enough room for a length d_min
  recurse_or_add(sph,p1,p2,p3,p4,a,d_min);
  */
}


void Triangulator::reset() {
  triangles.reset();
}


bool Triangulator::next() {
  return triangles.next();
}


bool Triangulator::at_end() const {
  return triangles.at_end();
}


const Geometry& Triangulator::get_triangle() const {
  return triangles.get_object();
}


void Triangulator::to_xml(std::ostream& o) {
  triangles.to_xml(o);
}


std::ostream& operator<<(std::ostream& o, Triangulator& triangulator) {
  triangulator.to_xml(o);
  return o;
}
