/***************************************************************************
                          triangulator.h  -  description
                             -------------------
    begin                : Sat Mar 29 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TRIANGULATOR_H
#define TRIANGULATOR_H

#include "compound.h"
#include "sph.h"


/** Triangulates the surfaces of a sph space.
 *
 *  Uses an adoption of the marching tetrahedra algorithm.
 *
 *  @author Moritz Franosch
 */
class Triangulator {

public: 

private:

  Compound triangles; // all voxels that the voxelizer generates

  double d_max; // maximal side length of triangles

  /** If the sides of the tetrahedron Tetrahedron(p1,p2,p3,p4) with
   *  length a are longer than d_min, recurse into smaller tetrahedra.
   *
   *  Does not work: A large regular tetrahedron can not be decomposed
   *  into smaller regular tetrahrdra!
   *
   *  Else, if the tetrahedron totally belongs to the part of the sph
   *  space filled with particles, or totally does not belong to it,
   *  then do nothing.
   *
   *  Else, if the tetrahedron only belongs partly to the volume with
   *  particles, then add appropriate triangles.
   */
  void recurse_or_add(SPH& sph,
		      const Particle::vector_t& p1,
		      const Particle::vector_t& p2,
		      const Particle::vector_t& p3,
		      const Particle::vector_t& p4,
		      const double a,
		      const double d_min);

  /** Calculates the point p on line [p0,p1] such that the part [p0,p]
   *  is mostly inside and p lies approximately on the surface.
   */
  void get_interpolation(SPH& sph,
			 const Particle::vector_t& p0,
			 const Particle::vector_t& p1,
			 Particle::vector_t& p);

  /** (p1,p2,p3,p4) is a right-handed tetrahedron inside a cube with
   *  one corner at (0,0,0) and side length d_max. If type==1, turn
   *  the tetrahedron by 90 in its cube before adding appropriate
   *  triangles. The corner of the cube is 'corner'.
   */
  void consider_tetrahedron(SPH& sph,
			    const Particle::vector_t& corner,
			    const Particle::vector_t& p1,
			    const Particle::vector_t& p2,
			    const Particle::vector_t& p3,
			    const Particle::vector_t& p4,
			    int type,
			    double d_max);

  /** Turns p by 90 around the z-axis and moves the result d_max
   *  along the x-axis.
   */
  void turn_and_move(const Particle::vector_t& p,
		     const double d_max,
		     Particle::vector_t& q);

  /** Adds triangles according to the rules of marching tetrahedra,
   *  dependent on what corners of the tetrahedron (p1,p2,p3,p4) are
   *  inside and what corners are outside.
   *
   *  The triangle (p2,p3,p4) must be right-handed, i.e. its normal is
   *  pointing towards p1.
   */
  void consider_tetrahedron(SPH& sph,
			    const Particle::vector_t& p1,
			    const Particle::vector_t& p2,
			    const Particle::vector_t& p3,
			    const Particle::vector_t& p4);

  /** Adds one triangle. 
   *
   *  Position p1 is inside the volume, positions p2, p3 and p4 are
   *  outside.
   */
  void add_triangle_from_tetrahedron(SPH& sph,
				     const Particle::vector_t& p1,
				     const Particle::vector_t& p2,
				     const Particle::vector_t& p3,
				     const Particle::vector_t& p4);

  /** Adds two triangles.
   *
   *  Positions p1 and p2 are inside the volume, positions p3 and p4
   *  are outside.
   *
   *  The triangle (p2,p3,p4) is right-handed, i.e. its normal is
   *  pointing towards p1.
   */
  void add_triangles_from_tetrahedron(SPH& sph,
				      const Particle::vector_t& p1,
				      const Particle::vector_t& p2,
				      const Particle::vector_t& p3,
				      const Particle::vector_t& p4);


  /** Adds the triangle (p1,p2,p3).
   */
  void add_triangle(SPH& sph,
		    const Particle::vector_t& p1,
		    const Particle::vector_t& p2,
		    const Particle::vector_t& p3);


  /** Checks whether the triangle (p2,p3,p4) is right-handed, i.e. its
   *  normal is pointing towards p1.
   */
  bool is_right_handed(const Particle::vector_t& p1,
		       const Particle::vector_t& p2,
		       const Particle::vector_t& p3,
		       const Particle::vector_t& p4) const;

  /** Stores the normal of the surface in n, pointing to the outside.
   */
  void compute_normal(SPH& sph,
		      const Particle::vector_t& p,
		      Particle::vector_t& n);
  
public:

  /** Constructs triangulated surfaces from a sph space.
   *
   *  @parameter d_min
   *  Approximate length of the sides of the triangles.
   */
  Triangulator(SPH& sph, const double d_min);

  /** Resets iterator through triangles.
   */
  void reset();

  /** Iterator set to next triangle.
   *
   *  Returns false if no next triangle exists.
   */
  bool next();

  /** Last next() has returned false.
   */
  bool at_end() const;

  /** Returns a reference to the current triangle.
   */
  const Geometry& get_triangle() const;

  /** Output of triangulator as XML.
   */
  void to_xml(std::ostream& o);

};


/** Output of triangulator as XML.
 */
std::ostream& operator<<(std::ostream& o, Triangulator& triangulator);


#endif // TRIANGULATOR_H
