/***************************************************************************
                          voxelizer.cpp  -  description
                             -------------------
    begin                : Sun Mar 9 2003
    copyright            : (C) 2003 by Moritz Franosch
    email                : mail@Franosch.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "blitz/tinyvec-et.h"

#include "sph.h"
#include "voxelizer.h"


using namespace std;


Voxelizer::Voxelizer(SPH& sph, const double d_min) {
  Particle::vector_t dp(sph.get_h());
  for (double x=0; x<=sph.get_space().get_length_x(); x+=sph.get_h()) {
    for (double y=0; y<=sph.get_space().get_length_y(); y+=sph.get_h()) {
      for (double z=0; z<=sph.get_space().get_length_z(); z+=sph.get_h()) {
	Particle::vector_t p1(x,y,z);
	Particle::vector_t p2=p1+dp;

	// If the voxel belongs to the fluid, add it.
	//
	// If the voxel only belongs partly to the fluid, add smaller
	// voxels recursively.
	add_and_recurse(sph,p1,p2,d_min);

      }
    }
  }

}


void Voxelizer::add_and_recurse(SPH& sph,
				const Particle::vector_t& p1,
				const Particle::vector_t& p2, 
				const double d_min) {
  REQUIRE(p1.length()==3);
  REQUIRE(p2.length()==3);
  Particle::vector_t dp=p2-p1;
  // If all eight corners of the box ...
  int ninside = 
    sph.is_inner_point(p1) + 
    sph.is_inner_point(p2) +
    sph.is_inner_point(p1+Particle::vector_t(dp[0],0,0)) + 
    sph.is_inner_point(p1+Particle::vector_t(0,dp[1],0)) + 
    sph.is_inner_point(p1+Particle::vector_t(0,0,dp[2])) + 
    sph.is_inner_point(p2-Particle::vector_t(dp[0],0,0)) + 
    sph.is_inner_point(p2-Particle::vector_t(0,dp[1],0)) + 
    sph.is_inner_point(p2-Particle::vector_t(0,0,dp[2]));
  // are inside the fluid ...
  if (ninside==8) {
    // ... then add the corresponding voxel.
    voxels.add_voxel(p1,p2);
  } else if (ninside==0) {
    // If no corner of the box is inside the fluid then return.
    return;
  } else {
    // If some corners are inside and some are outside and
    // if largest structure ...
    double d = max(max(dp[0],dp[1]),dp[2]);

    /*
    cerr << "d=" << d << endl;
    cerr << "d_min=" << d_min << endl;
    */

    // ... is still larger than d_min, ...
    if ( d > d_min ) {
      // ... then recurse into the eight sub-boxes.

      dp /= 2;

      Particle::vector_t p0=p1;
      add_and_recurse(sph,p0,p0+dp,d_min);
      
      p0=p1+Particle::vector_t(dp[0],0,0);
      add_and_recurse(sph,p0,p0+dp,d_min);
      
      p0=p1+Particle::vector_t(0,dp[1],0);
      add_and_recurse(sph,p0,p0+dp,d_min);
      
      p0=p1+Particle::vector_t(dp[0],dp[1],0);
      add_and_recurse(sph,p0,p0+dp,d_min);
      

      p0=p1+Particle::vector_t(0,0,dp[2]);
      add_and_recurse(sph,p0,p0+dp,d_min);
      
      p0=p1+Particle::vector_t(dp[0],0,dp[2]);
      add_and_recurse(sph,p0,p0+dp,d_min);
      
      p0=p1+Particle::vector_t(0,dp[1],dp[2]);
      add_and_recurse(sph,p0,p0+dp,d_min);
      
      p0=p1+Particle::vector_t(dp[0],dp[1],dp[2]);
      add_and_recurse(sph,p0,p0+dp,d_min);

    }
    /*
    } else {

      // For each of the six surfaces add subvoxels.

      Particle::vector_t p0=p1;

      add_subvoxel(sph,
		   p0,
		   p0+Particle::vector_t(dp[0],0,0),
		   p0+Particle::vector_t(dp[0],dp[1],0),
		   p0+Particle::vector_t(0,dp[1],0),
		   Particle::vector_t(0,0,-dp[2])
		   );

      add_subvoxel(sph,
		   p0,
		   p0+Particle::vector_t(dp[0],0,0),
		   p0+Particle::vector_t(dp[0],0,dp[2]),
		   p0+Particle::vector_t(0,0,dp[2]),
		   Particle::vector_t(0,-dp[1],0)
		   );

      add_subvoxel(sph,
		   p0,
		   p0+Particle::vector_t(0,dp[1],0),
		   p0+Particle::vector_t(0,dp[1],dp[2]),
		   p0+Particle::vector_t(0,0,dp[2]),
		   Particle::vector_t(-dp[0],0,0)
		   );


      add_subvoxel(sph,
		   p1-Particle::vector_t(dp[0],dp[1],0),
		   p1-Particle::vector_t(dp[0],0,0),
		   p1,
		   p1-Particle::vector_t(0,dp[1],0),
		   Particle::vector_t(0,0,dp[2])
		   );

      add_subvoxel(sph,
		   p1-Particle::vector_t(dp[0],0,dp[2]),
		   p1-Particle::vector_t(dp[0],0,0),
		   p1,
		   p1-Particle::vector_t(0,0,dp[2]),
		   Particle::vector_t(0,dp[1],0)
		   );

      add_subvoxel(sph,
		   p1-Particle::vector_t(0,dp[1],dp[2]),
		   p1-Particle::vector_t(0,dp[1],0),
		   p1,
		   p1-Particle::vector_t(0,0,dp[2]),
		   Particle::vector_t(dp[0],0,0)
		   );

    }
    */

  }
}

/*
void Voxelizer::add_subvoxel(SPH& sph, 
			     const Particle::vector_t& p1,
			     const Particle::vector_t& p2,
			     const Particle::vector_t& p3,
			     const Particle::vector_t& p4,
			     const Particle::vector_t& pn) {

  // Determine the four largest heights at the corners of the
  // surface that are still inside.
  double h1 = get_largest_height(sph,p1,pn);
  double h2 = get_largest_height(sph,p2,pn);
  double h3 = get_largest_height(sph,p3,pn);
  double h4 = get_largest_height(sph,p4,pn);

  // Add the box-shaped voxel that fills up the surface to the
  // minimum of the computed heights.
  double h_min = min(min(min(h1,h2),h3),h4);
  Particle::vector_t pn_normed = pn/euclidian_norm(pn);
  voxels.add_voxel(p1,p3+h_min*pn_normed);
  
  // Add two pyramids with the edge at the point where the minimum
  // of the computed heights is.
  if (h_min==h1) {
    voxels.add_pyramid(p1, p2,p3,p3+h3*pn_normed,p2+h2*pn_normed);
    voxels.add_pyramid(p1, p4,p3,p3+h3*pn_normed,p4+h4*pn_normed);
  } else
  if (h_min==h2) {
    voxels.add_pyramid(p2, p3,p4,p4+h4*pn_normed,p3+h3*pn_normed);
    voxels.add_pyramid(p2, p1,p4,p4+h4*pn_normed,p1+h1*pn_normed);
  } else
  if (h_min==h3) {
    voxels.add_pyramid(p3, p4,p1,p1+h1*pn_normed,p4+h4*pn_normed);
    voxels.add_pyramid(p3, p2,p1,p1+h1*pn_normed,p2+h2*pn_normed);
  } else
  if (h_min==h4) {
    voxels.add_pyramid(p4, p1,p2,p2+h2*pn_normed,p1+h1*pn_normed);
    voxels.add_pyramid(p4, p3,p2,p2+h2*pn_normed,p3+h3*pn_normed);
  }

}
*/


/*
double Voxelizer::get_largest_height(SPH& sph, 
				     const Particle::vector_t& p0,
				     const Particle::vector_t& pn) {
  REQUIRE( sph.is_inner_point(p0) );
  REQUIRE( !sph.is_inner_point(p0+pn) );

  // Midpoint-of-interval-method.

  double h_max=euclidian_norm(pn);
  double h_min=0;
  Particle::vector_t pn_normed = pn/h_max;

  double accuracy=(h_max-h_min)/10;

  // While the upper and lower bound are still too far from each
  // other.
  while (h_max-h_min>accuracy) {
    // If the midpoint ...
    double h_mid=(h_max+h_min)/2;
    if ( sph.is_inner_point(p0+h_mid*pn_normed ) ) {
      // ... is inside then the new lower bound is the midpoint.
      h_min=h_mid;
    } else {
      // ... is outside then the new upper bound is the midpoint.
      h_max=h_mid;
    }
  }

  double h=h_min;
  ENSURE( sph.is_inner_point(p0+h*pn_normed) );
  return h;
}
*/


void Voxelizer::reset() {
  voxels.reset();
}


bool Voxelizer::next() {
  ENSURE( !at_end() );
  return voxels.next();
}


bool Voxelizer::at_end() const {
  return voxels.at_end();
}


const Geometry& Voxelizer::get_voxel() const {
  ENSURE( !at_end() );
  return voxels.get_object();
}


void Voxelizer::to_xml(std::ostream& o) {
  voxels.to_xml(o);
}


ostream& operator<<(ostream& o, Voxelizer& voxelizer) {
  voxelizer.to_xml(o);
  return o;
}
